#!/bin/sh
#######################################################################
# Copyright 2008-2010 VMware, Inc.  All rights reserved.
# -- VMware Confidential
#######################################################################
# Installer for offline esx4update on a VMvisor image
#######################################################################
# prep - Prepare the stage bootbank
#######################################################################
# install - Copy all components in the stage bank to bootbank
#######################################################################

DEBUG=${TRUE}

#
# Print debug messages to STDOUT
#

print_debug()
{
   if [ $DEBUG -eq ${TRUE} ];
   then
      echo "$@"
   fi
}

#
# Print error message to STDERR and terminate
#

panic()
{
   echo "PANIC: ${@}" >&2
   exit 1
}

#
# Verify file copy by comparing the digests of the source and destination
#

verify_copy()
{
   file1=$1
   file2=$2
   method=$3

   #
   # Get the md5 sums for the files passed
   #
   cmdline="openssl dgst -${method}"
   dgst_1=$(${cmdline} ${file1} 2>/dev/null | cut -d" " -f 2 )
   dgst_2=$(${cmdline} ${file2} 2>/dev/null | cut -d" " -f 2 )

   #
   # If the digest strings are empty, something went wrong
   #
   if [ -z "${dgst_1}" ];
   then
      panic "Unable to obtain digest for ${file1}"
   fi

   if [ -z "${dgst_2}" ];
   then
      panic "Unable to obtain digest for ${file2}"
   fi

   #
   # If the digests do not match, the copy failed
   #
   if [ "${dgst_1}" != "${dgst_2}" ];
   then
      panic "Copy of ${file1} to ${file2} failed on validation"
   fi
}

#
# Copy file and verify that the operation was successful
#

copy_and_verify()
{
   src=$1
   dest=$2

   #
   # If the source file is not there, no copy is possible
   #
   if [ ! -f "${src}" ];
   then
      panic "File to copy ${src} does not exist!"
   fi

   #
   # Copy operation
   #
   cp -f "${src}" "${dest}"

   #
   # Verify copy operation
   #
   verify_copy "${src}" "${dest}" md5
}

#
# Check if two files are different by comparing digests
#

is_modified()
{
   file1=$1
   file2=$2
   method=$3

   #
   # If the source file does not exist, no comparison is possible
   #
   if [ ! -f "${file1}" ];
   then
      panic "Source file ${file1} does not exist"
   fi

   #
   # If the destination file does not exist, the files are definitely
   # different. We should not fail here because we might be adding
   # a new file to the bootbank
   #
   if [ ! -f "${file2}" ];
   then
      return ${TRUE}
   fi

   #
   # Get the md5 sums for the files passed
   #
   cmdline="openssl dgst -${method}"
   dgst_1=$(${cmdline} ${file1} 2>/dev/null | cut -d" " -f 2 )
   dgst_2=$(${cmdline} ${file2} 2>/dev/null | cut -d" " -f 2 )

   #
   # If the digest strings are empty, something went wrong
   #
   if [ -z "${dgst_1}" ];
   then
      panic "Unable to obtain digest for ${file1}"
   fi

   if [ -z "${dgst_2}" ];
   then
      panic "Unable to obtain digest for ${file2}"
   fi

   if [ "${dgst_1}" != "${dgst_2}" ];
   then
      return ${TRUE}
   else
      return ${FALSE}
   fi
}

#
# Create and populate the stage bootbank
#

prepare_bootbankstage()
{
   #
   # Create stage bootbank
   #
   print_debug "Recreating a fresh ${stagebank}"
   rm -rf "${stagebank}"
   mkdir -p "${stagebank}"

   #
   # Check real bootbank and stage bootbank
   #
   if [ ! -d "${curbootbank}" ];
   then
      panic "Bootbank can not be found"
   fi

   if [ ! -d "${stagebank}" ];
   then
      panic "Stage bootbank can not be found"
   fi

   #
   # Check firmware state
   #
   print_debug "Checking the firmware state"
   curstate=$(sed -n 's/^bootstate=//p' $curbootbank/boot.cfg)
   if [ -z "${curstate}" ];
   then
      panic "The Visor image mounted must be a fresh image that has never been booted"
   fi

   if [ ${curstate} -ne 0 ];
   then
      panic "The Visor image mounted must be a fresh image that has never been booted"
   fi

   #
   # Copy all files from /bootbank to /bootbankstage/bank,
   # which is the destination to assemble all firmware components
   #
   for f in ${curbootbank}/* ;
   do
      copy_and_verify "${f}" "${stagebank}/$(basename ${f})"
      print_debug "Copied ${f} to ${stagebank}/$(basename ${f})"
   done
}

#
# Populate the image with the new tgz files
#

complete_install()
{
   cachedir=$(/usr/sbin/vib-env vibddi.cdcache)
   if [ "${cachedir}" != "none" -a ! -d "${cachedir}" ];
   then
      panic "CD cache directory can not be found"
   fi

   #
   # Check to see if boot.cfg is modified to see if this is a firmware vib.
   #
   if is_modified "${stagebank}/boot.cfg" "${curbootbank}/boot.cfg" md5 ;
   then
      print_debug "Detected firmware upgrade, preserving original boot state"

      #
      # Preserve the original boot state to make this look like
      # a generic image to the bootloader
      #
      sed -i -e "s:updated\=.*:updated\=1:" \
             -e "s:bootstate\=.*:bootstate\=0:" \
             "${stagebank}/boot.cfg"
   fi

   #
   # All VIBs are allowed, so copy all files whether changed or not.
   #
   for f in ${stagebank}/* ;
   do
      copy_and_verify "${f}" "${curbootbank}/$(basename ${f})"
      print_debug "Copied ${f} to ${curbootbank}/$(basename ${f})"
   done

   if [ "${cachedir}" != "none" ];
   then
      copy_and_verify "${stagebank}/oem.tgz" "${cachedir}/oem.tgz"
      print_debug "Copied ${stagebank}/oem.tgz to ${cachedir}/oem.tgz"

      if [ -f "${stagebank}/m.z" ];
      then
         copy_and_verify "${stagebank}/m.z" "${cachedir}/mod.tgz"
         print_debug "Copied ${stagebank}/m.z to ${cachedir}/mod.tgz"
      elif [ -f "${stagebank}/mod.tgz" ];
      then
         copy_and_verify "${stagebank}/mod.tgz" "${cachedir}/mod.tgz"
         print_debug "Copied ${stagebank}/mod.tgz to ${cachedir}/mod.tgz"
      elif [ -f "${stagebank}/extmod.tgz" ];
      then
         copy_and_verify "${stagebank}/extmod.tgz" "${cachedir}/mod.tgz"
         print_debug "Copied ${stagebank}/extmod.tgz to ${cachedir}/mod.tgz"
      fi
   fi
}

#
# Main
#

#
# Usage: bootbankstage-install.sh [-v] <prep|install>
#

cmd=''
for optarg in "$@"; do
   case ${optarg} in
      prep)
         cmd="prep"
         ;;
      install)
         cmd="install"
         ;;
      -v)
         verbose=1
         ;;
      *)
         ;;
   esac
done

#
# Get bootbank dirs from vib.conf; no default values are used in offline mode
#

if [ -f "/usr/sbin/vib-env" ];
then
   curbootbank=$(/usr/sbin/vib-env visor.bootbankdir)
   stagebank="$(/usr/sbin/vib-env visor.stagedir)/bank"
else
   panic "/usr/sbin/vib-env not found - it is required to run esxupdate in offline mode"
fi

if [ -z "${curbootbank}" -o "${stagebank}" == "/bank" ];
then
   panic "vib.conf from esxupdate.conf not found - it is required to run esxupdate in offline mode"
fi

if [ "${stagebank}" == "/altbootbank/bank" ];
then
   panic "Stage bootbank can not be located on /altbootbank"
fi

if [ -f "${curbootbank}/local.tgz" -o -f "${curbootbank}/state.tgz" ];
then
   panic "The Visor image mounted must be a fresh image that has never been booted"
fi

#
# Execute operation
#

case $cmd in
   prep)
      prepare_bootbankstage
      ;;
   install)
      complete_install
      ;;
   *)
      panic "$0 does not support command \"$cmd\""
esac
