#######################################################################
# Copyright 2008 VMware, Inc.  All rights reserved.
# -- VMware Confidential
#######################################################################
# utils.pl - helper routines for vibddi
#######################################################################

#
# 'Global' variables for lock file
#

$LOCKHANDLE;
$LOCKPATH = "/var/log/vibddi.log";

#
# Tracks all of the directories that are mount points
#

@MOUNT_POINTS = ();

#
# Maps return codes and error message for logging events
#

%ERR_CODES = ("Success" => { "RetCode" => 0, "ErrMsg" => "Operation completed successfully"},
              "Interrupted" => { "RetCode" => 0, "ErrMsg" => "Process interrupted"},
              "NotRoot" => { "RetCode" => 1, "ErrMsg" => "Not invoked as root user"},
              "BadArgs" => { "RetCode" => 2, "ErrMsg" => "Invalid arguments"},
              "NoLock" => { "RetCode" => 3, "ErrMsg" => "Unable to obtain lock file"},
              "BadImage" => { "RetCode" => 4, "ErrMsg" => "Invalid ESXi target"},
              "DepMissing" => { "RetCode" => 5, "ErrMsg" => "Dependency missing"},
              "NotFound" => { "RetCode" => 6, "ErrMsg" => "File not found"},
              "BadCopy" => { "RetCode" => 7, "ErrMsg" => "File copy failed"},
              "BadChmod" => { "RetCode" => 8, "ErrMsg" => "File chmod failed"},
              "BadPipeRead" => { "RetCode" => 9, "ErrMsg" => "Read pipe failed"},
              "BadFileRead" => { "RetCode" => 10, "ErrMsg" => "Read file failed"},
              "BadFileWrite" => { "RetCode" => 11, "ErrMsg" => "Write file failed"},
              "BadDir" => { "RetCode" => 12, "ErrMsg" => "Create directory failed"},
              "NoLoopDev" => { "RetCode" => 13, "ErrMsg" => "No free loop device found - is losetup available?"},
              "BadMount" => { "RetCode" => 14, "ErrMsg" => "Mount failed"},
              "BadFdisk" => { "RetCode" => 15, "ErrMsg" => "Bad data returned from fdisk command"},
              "BadConf" => { "RetCode" => 16, "ErrMsg" => "Data missing in configuration"},
              "BadOp" => { "RetCode" => 17, "ErrMsg" => "esxupdate operation failed"},
              "BadCheckOp" => { "RetCode" => 18, "ErrMsg" => "ddfsck operation failed"},
              "BadCd" => { "RetCode" => 19, "ErrMsg" => "ThinESX/Recovery CD customization failed"},
              "BadBootCfg" => { "RetCode" => 20, "ErrMsg" => "Invalid bootloader configuration"},
              "BadUmount" => { "RetCode" => 21, "ErrMsg" => "Umount failed"},
              "BadCmd" => { "RetCode" => 22, "ErrMsg" => "Executing external command failed"});

#
# Maps return codes to error messages for esx4update operations
#

%ESX4UPDATE_ERR_MSGS = (0 => "Operation completed successfully",
                        1 => "Not invoked as root user",
                        2 => "Invalid arguments",
                        3 => "Another instance of esxupdate is currently running",
                        4 => "Failed to download metadata",
                        5 => "The format of the metadata is invalid",
                        7 => "Failed to download VIB",
                        8 => "The format of the VIB is invalid",
                        9 => "Error reading or writing VIB data",
                        10 => "Unable to create, write or read a file as expected",
                        11 => "The format of the database is invalid",
                        12 => "Unable to read or write to database",
                        13 => "No matching bulletin or VIB was found in the metadata",
                        14 => "There was an error resolving dependencies",
                        15 => "The package manager transaction failed",
                        16 => "Failed to parse configuration files",
                        17 => "There was an error setting up ESXi installation destination",
                        18 => "The host is required to be in maintenance mode for this update, but it could not be put in this state",
                        19 => "A post-transaction script failed with a nonzero exit code",
                        20 => "VIB signature is missing",
                        21 => "Unsupported VIB signature version",
                        22 => "PKCS7 certificate is too short or corrupted",
                        23 => "PKCS7 certificate is not valid (expired, etc) or the public key is missing",
                        24 => "VIB digest does not match the digest in the PKCS7 certificate",
                        25 => "Additional VIBs are required to satisfy dependencies",
                        26 => "Failed to download offline bundle ZIP",
                        27 => "Invalid bundle ZIP archive, or missing metadata.zip inside",
                        28 => "The requested command is not supported on the platform",
                        29 => "The VIB is not supported on the platform",
                        80 => "System reboot required",
                        81 => "Host Agent reboot required",
                        82 => "Host was not updated, no changes required");

#
# Maps return codes to error messages for ddfsck operations
#

%DDFSCK_ERR_MSGS = (0 => "Operation completed successfully",
                    1 => "Invalid arguments",
                    2 => "File not found",
                    3 => "File I/O failed",
                    4 => "Out of memory",
                    5 => "Invalid filesystem type",
                    6 => "Invalid disk signature",
                    7 => "Invalid character detected in filename",
                    8 => "Duplicate start cluster detected",
                    9 => "Invalid cluster value detected",
                    10 => "Duplicate cluster detected",
                    11 => "Cluster chain length does not match filesize");


#######################################################################
#
# SignalHandler --
#
#      Handle all signals for vibenv/vibddi
#
# Results
#      None
#
# Side effects:
#      The application will terminate here
#
#######################################################################

sub SignalHandler
{
   LogEventAndDie($scriptName, "Interrupted");
}


#######################################################################
#
# Lock --
#
#      Grabs the lock file/starts the logger
#
# Results
#      Returns 0 if the lock is obtained, otherwise 1
#
# Side effects:
#      Another instance of vibenv/vibddi can not run concurrently
#      If the log is greater than 1 MB it will be rotated
#
#######################################################################

sub Lock
{
   if ((-s $LOCKPATH) > 1024000) {
      rename($LOCKPATH, "$LOCKPATH." . time()) || return 1;
   }
   open($LOCKHANDLE, ">>$LOCKPATH") || return 1;
   flock($LOCKHANDLE, 2 | 4) || return 1;
   return 0;
}


#######################################################################
#
# Unlock --
#
#      Releases the lock file/stops the logger
#
# Results
#      None
#
# Side effects:
#      None
#
#######################################################################

sub Unlock
{
   flock($LOCKHANDLE, 8);
   close($LOCKHANDLE);
}


#######################################################################
#
# Timestamp --
#
#      Creates an ISO 8601 timestamp and returns it as a string
#
# Results
#      Timestamp string
#
# Side effects:
#      None
#
#######################################################################

sub Timestamp
{
   my($date, $time);
   my($sec,
      $min,
      $hour,
      $mday,
      $mon,
      $year,
      $wday,
      $yday,
      $isdst) = localtime(time);
   $year += 1900;
   $mon++;
   $date = sprintf("%4d-%2.2d-%2.2d", $year, $mon, $mday);
   $time = sprintf("%2.2d:%2.2d:%2.2d", $hour, $min, $sec);
   return join(' ', $date, $time);
}


#######################################################################
#
# LogEvent --
#
#      Prints a formatted message with timestamp to the log
#      and just the message to stdout
#
# Results
#      None
#
# Side effects:
#      None
#
#######################################################################

sub LogEvent
{
   my $file = shift;
   my $msg = shift;
   my $debug = shift;
   my $timeStamp = &Timestamp();
   print $LOCKHANDLE "[$timeStamp] [$file] $msg\n";
   if (!defined($debug)) {
      print STDOUT "$msg\n";
   }
}


#######################################################################
#
# LogEventAndDie --
#
#      Prints a formatted message with timestamp to stderr/log
#      and exits with passed return code
#
# Results
#      None
#
# Side effects:
#      Terminates the application
#
#######################################################################

sub LogEventAndDie
{
   my $file = shift;
   my $key = shift;
   my $msg = shift;
   my $timeStamp = &Timestamp();
   print STDERR $ERR_CODES{$key}{"ErrMsg"};
   print $LOCKHANDLE "[$timeStamp] [$file] " . $ERR_CODES{$key}{"ErrMsg"};
   if (defined $msg) {
      print STDERR ": $msg\n";
      print $LOCKHANDLE ": $msg\n";
   } else {
      print STDERR "\n";
      print $LOCKHANDLE "\n";
   }
   CleanUpMounts();
   Unlock();
   exit $ERR_CODES{$key}{"RetCode"};
}


#######################################################################
#
# IsVisor --
#
#      Inspects the last two bytes of the boot sector (offset 0x1FE) 
#      for the x86 disk signature (0xAA55)
#
# Results
#      Returns 0 if the path passed is a valid VMvisor
#      dd image, otherwise 1
#
# Side effects:
#      None
#
#######################################################################

sub IsVisor
{
   my $ddImage = shift;
   my $offset = 0x1FE;
   my $x86Sig = "0xAA55";
   my $readBuffer;

   if (! -f $ddImage) {
      return 1;
   }

   open(FILE, "<$ddImage") || return 1;
   binmode(FILE);
   seek(FILE, $offset, SEEK_SET);
   read(FILE, $readBuffer, 2);
   close(FILE);

   my @bytes = split(//, $readBuffer);
   my $byteString = sprintf("0x%02X%02X", ord($bytes[1]), ord($bytes[0]));
   if ($byteString eq $x86Sig) {
      return 0;
   } else {
      return 1;
   }
}


#######################################################################
#
# CreateDir --
#
#      Creates a directory
#
# Results
#      Returns 0 if the directory was created, otherwise 1
#
# Side effects:
#      None
#
#######################################################################

sub CreateDir
{
   my $dirPath = shift;
   if (length($dirPath) == 0) {
      return 1;
   }
   if (-e $dirPath || -d $dirPath) {
      return 1;
   }
   eval { mkpath("$dirPath", 0, 0755) };
   if (! -d $dirPath) {
      return 1;
   }
   return 0;
}


#######################################################################
#
# CreateFreshDir --
#
#      Deletes the directory, the creates a fresh one with
#      the same path
#
# Results
#      Returns 0 if the directory was created, otherwise 1
#
# Side effects:
#      None
#
#######################################################################

sub CreateFreshDir
{
   my $dirPath = shift;
   eval { rmtree($dirPath) };
   return CreateDir($dirPath);
}


#######################################################################
#
# EmptyFile --
#
#      Empties a file or creates a new one if it does not exist
#
# Results
#      Returns 0 if the file was emptied or created, otherwise 1
#
# Side effects:
#      None
#
#######################################################################

sub EmptyFile
{
   my $filePath = shift;
   open(FILE, ">$filePath") || return 1;
   close(FILE);
   return 0;
}


#######################################################################
#
# PipeReadSingle --
#
#      Executes a command and returns the single line output
#
# Results
#      Returns the command output, otherwise an empty string
#
# Side effects:
#      None
#
#######################################################################

sub PipeReadSingle
{
   my $cmd = shift;
   open(FILE, "$cmd |") || return "";
   my $line = <FILE>;
   chomp($line);
   close(FILE);
   return $line;
}


#######################################################################
#
# PipeReadMulti --
#
#      Executes a command and returns an array of output, and
#      the return code if necessary
#
# Results
#      Returns the command output array, otherwise an empty array
#
# Side effects:
#      None
#
#######################################################################

sub PipeReadMulti
{
   my $cmd = shift;
   my $rc = shift;
   my @lines = ();
   open(FILE, "$cmd |") || return @lines;
   foreach $line (<FILE>) {
      chomp($line);
      push(@lines, $line);
   }
   close(FILE);

   if (defined($rc)) {
      return (\@lines, $? >> 8);
   } else {
      return @lines;
   }
}


#######################################################################
#
# ReadEnvProp --
#
#      Reads a property string from esxupdate.conf using vib-env
#
# Results
#      Returns the property string
#
# Side effects:
#      None
#
#######################################################################

sub ReadEnvProp
{
   my $propKey = shift;
   my $propVal = PipeReadSingle("/usr/sbin/vib-env $propKey");
   chomp($propVal);
   return $propVal;
}


#######################################################################
#
# WriteEnvProp --
#
#      Writes a new value to vib.conf, then replaces esxupdate.conf
#      with the new vib.conf file so the value persists
#
# Results
#      Returns 0 if the value was written, otherwise 1
#
# Side effects:
#      Deletes the old vib.conf file if it exists
#
#######################################################################

sub WriteEnvProp
{
   my $propKey = shift;
   my $propVal = shift;
   my $confPath = "/etc/vmware/esxupdate/vib.conf";
   if (-e $confPath) {
      if (EmptyFile($confPath) != 0) {
         return 1;
      }
   }
   `/usr/sbin/vib-env $propKey $propVal &> /dev/null`;
   if (ReadEnvProp($propKey) ne $propVal) {
      return 1;
   }
   copy($confPath, "/etc/vmware/esxupdate/esxupdate.conf") || return 1;
   return 0;
}


#######################################################################
#
# StringTrim --
#
#      Trims leading and trailing whitespace from the string
#
# Results
#      Returns the trimmed string
#
# Side effects:
#      None
#
#######################################################################

sub StringTrim
{
	my $string = shift;
	$string =~ s/^\s+//;
	$string =~ s/\s+$//;
	return $string;
}


#######################################################################
#
# IsDigit --
#
#      Validates that an input string is a digit value
#
# Results
#      Returns 0 if input is a digit value, otherwise 1
#
# Side effects:
#      None
#
#######################################################################

sub IsDigit
{
   my $input = shift;
   if ($input =~ /^\d+$/) {
      return 0;
   }
   return 1;
}


#######################################################################
#
# FindLoopDevice --
#
#      Finds the first free loop device
#
# Results
#      Returns the path to the loop device, otherwise 1
#
# Side effects:
#      None
#
#######################################################################

sub FindLoopDevice
{
   my $losetupCmd = "/sbin/losetup -f";
   my $loopDev = PipeReadSingle($losetupCmd);
   if (length($loopDev) == 0) {
      return 1;
   }
   return $loopDev;
}


#######################################################################
#
# MountTarget --
#
#      Mounts the target and pushes the mount point onto
#      the global MOUNT_POINTS array
#
# Results
#      Returns 0 if the mount operation was successful, otherwise 1
#
# Side effects:
#      None
#
#######################################################################

sub MountTarget
{
   my $target = shift;
   my $mountPoint = shift;
   my $extra = shift;
   my $isReadOnly = shift;

   my $mountCmd = "/bin/mount";
   if ($isReadOnly == 1) {
      $mountCmd .= " -r";
   }
   if ($extra eq "iso") {
      $mountCmd .= " -t iso9660 -o loop $target";
   } elsif (-b $target) {
      $mountCmd .= " $target$extra";
   } else {
      $mountCmd .= " -o loop -o offset=$extra $target";
   }
   $mountCmd .= " $mountPoint &> /dev/null";

   `$mountCmd`;
   my $rc = $? >> 8;
   if ($rc == 0) {
      push(@MOUNT_POINTS, $mountPoint);
   }
   if ((! -e "$mountPoint/boot.cfg") &&
       (! -e "$mountPoint/vmkernel.gz") &&
       (! -e "$mountPoint/vmk.gz") &&
       (! -e "$mountPoint/k.gz") && 
       (! -d "$mountPoint/packages")) {
      return 1;
   }
   return 0;
}


#######################################################################
#
# CleanUpMounts --
#
#      Unmount and delete mount points
#
# Results
#      Returns 0 if all mount points are unmounted
#      and deleted, otherwise 1
#
# Side effects:
#      All directories contained in the MOUNT_POINTS
#      array will be deleted
#
#######################################################################

sub CleanUpMounts
{
   my ($numRetries, $unmounted);
   while (@MOUNT_POINTS) {
      my $mountPoint = $MOUNT_POINTS[$#MOUNT_POINTS];
      $numRetries = 3;
      $unmounted = 0;
      while (($numRetries > 0) && ($unmounted == 0)) {
         `/bin/umount $mountPoint  &> /dev/null`;
         my $rc = $? >> 8;
         if ($rc == 0) {
            $unmounted = 1;
         } else {
            sleep(1);
            $numRetries--;
         }
      }
      if ($unmounted == 0) {
         return 1;
      }
      pop(@MOUNT_POINTS);
      eval { rmtree($mountPoint) };
   }
   return 0;
}


#######################################################################
#
# CopyDirFirstLevel --
#
#      Copies the first level contents of a directory to some
#      destination, ignores subdirectories and hidden files
#
# Results
#      Returns 0 if the copy operation was successful, otherwise 1
#
# Side effects:
#      None
#
#######################################################################

sub CopyDirFirstLevel
{
   my $dirPath = shift;
   my $destPath = shift;
   opendir(DIR, $dirPath) || return 1;
   my @dirList = readdir(DIR);
   closedir(DIR);
   foreach $file (@dirList) {
      if (($file !~ m/^\./) && (! -d "$dirPath/$file")) {
         copy("$dirPath/$file", "$destPath/$file") || return 1;
      }
   }
   return 0;
}


#######################################################################
#
# ListFilesInDir --
#
#      Returns all of the files and directories contained
#      in a directory, ignores hidden files
#
# Results
#      Returns the directory list array, otherwise an empty array
#
# Side effects:
#      None
#
#######################################################################

sub ListFilesInDir
{
   my $dirPath = shift;
   my @dirList = ();
   opendir(DIR, $dirPath) || return @dirList;
   my @fileList = readdir(DIR);
   closedir(DIR);
   foreach $name (@fileList) {
      if ($name !~ m/^\./) {
         push(@dirList, $name);
      }
   }
   return @dirList;
}


#######################################################################
#
# ComputeMD5 --
#
#      Gets the MD5 checksum of a file using Digest::MD5
#
# Results
#      Returns the md5 checksum as a string, or empty string if there is an error.
#
# Side effects:
#      None
#
#######################################################################

sub ComputeMD5
{
   my $fileName = shift;
   open(INPUTFILE, $fileName) or return "";
   my $md5Obj = Digest::MD5->new;
   $md5Obj->addfile(*INPUTFILE);
   my $md5Str = $md5Obj->hexdigest;
   close(INPUTFILE);
   return $md5Str;
}


#######################################################################
#
# GenerateMd5File --
#
#      Equivalent to calling "md5sum" and redirecting output to
#      a file, but this allows you to specify a different filename
#      in the output
#
# Results
#      Returns 0 if the file is generated, otherwise 1
#
# Side effects:
#      None
#
########################################################################

sub GenerateMd5File
{
   my $file = shift;
   my $fileName = shift;
   my $md5File = shift;

   my $md5Contents = ComputeMD5($file) . "  $fileName\n";
   open(NEWFILE, ">$md5File") || return 1;
   print NEWFILE $md5Contents;
   close(NEWFILE);

   return 0;
}


#######################################################################
#
# CleanupFiles --
#
#      Unlink a set of files stored in an array and return status message
#
# Results
#      If any files were passed in the array, returns the number of files
#      deleted along with a text description. If no files were passed,
#      returns empty string.
#
# Side effects:
#      Files in the input array are removed
#
#######################################################################

sub CleanupFiles
{
   my @fileList = @_;
   my $numFiles = $#fileList + 1;
   my $numDeletedFiles;
   if ($numFiles > 0) {
      $numDeletedFiles = unlink(@fileList);
      if ($numFiles == $numDeletedFiles) {
         return "; $numFiles files had been at least partially written before" .
            " the failure--all of those have now been removed";
      }
      elsif ($numDeletedFiles > 0) {
         return "; $numFiles files had been at least partially written before" .
            "the failure--however, only $numDeletedFiles files could be removed";
      }
      else {
         return "; $numFiles files had been at least partially written before" .
            " the failure--however, none of those files could be removed";
      }
   }
   else {
      return "";
   }
}


#######################################################################
#
# GenerateThinUUIDFile --
#
#      Generates a file that contains a proper UUID
#      for ThinESX
#
# Results
#      Returns 0 if the file is generated, otherwise 1
#
# Side effects:
#      None
#
#######################################################################

sub GenerateThinUUIDFile
{
   my $uuidFile = shift;
   my $partitionNum = shift;

   my $uuidFormat = "VMWARE FAT16    %s";
   my $md5 = Digest::MD5->new;
   $md5->add(time(), "$partitionNum");
   my $digest = $md5->digest;
   my $uuid =  sprintf("%s%s",
                       chr(0x7 << 4 | ord($digest) & 0xF),
                       substr($digest, 1));
   my $output = sprintf($uuidFormat, $uuid);

   open (FILE, ">$uuidFile") || return 1;
   print FILE $output;
   close(FILE);
   return 0;
}
1;
