#######################################################################
# Copyright (C) 2009 VMWare, Inc.
# All Rights Reserved
########################################################################
#
# Bulletin.py
#

__all__ = ['Bulletin', 'BulletinCollection']

import datetime
import logging
import time

try:
   from lxml import etree
except ImportError:
   try:
      import xml.etree.cElementTree as etree
   except ImportError:
      import cElementTree as etree

import Descriptor
import Vib
import VIBCollection
import errors
import version
import util

log = logging.getLogger(__name__)

class Bulletin(VIBCollection.VIBCollection):
   """A Bulletin defines a set of Vib packages for a patch, update,
      or ESX extension.

      Class Variables:
         * XML_DEST_META      - A constant to signify XML output destined for
                                vmware.xml (metadata.zip).
         * XML_DEST_DB        - A constant to signify XML output destined for
                                bulletins.zip (in the esxudpate database).
       Attributes:
           * id             - A string specifying the unique bulletin ID.
           * description    - The (multi-line) bulletin description text.
           * category       - A string specifying the bulletin's category.
           * severity       - A string specifying the bulletin's severity.
           * urgency        - A string specifying the bulletin's "urgency."
           * kburl          - A URL to a knowledgebase article related to the
                              bulletin.
           * contact        - Contact information for the bulletin's publisher.
           * summary        - The abbreviated (single-line) bulletin summary
                              text.
           * releasetype    - A string specifying the release type.
           * releasedate    - An integer or float value giving the bulletin's
                              release date/time. May be None if release date is
                              unknown.
           * releasedatestr - A string value which may represent an unparseable
                              date from bulletin meta-data.
           * installdate    - An integer or float value giving the time at which
                              the bulletin became compliant on the host.  May be
                              None if unknown or if bulletin is not compliant.
           * vendor         - A string specifying the vendor/publisher.
           * version        - A Version object giving the version number of the
                              Bulletin.
           * vibids         - A set of VIB IDs with no corresponding VIB
                              object. These would generally be from a <vibList>
                              XML element, and are meant to be referenced later
                              in order to properly establish keys in the
                              Bulletin with proper VIB objects as values.
           * compliant      - True if host is compliant with the bulletin; False
                              if not. (A bulletin is compliant if all of its
                              VIBs which match the host's software and hardware
                              platforms are either installed on the host, or
                              obsoleted by something installed on the host.)
           * persist        - A boolean specifying whether to persist this
                              Bulletin in the esxupdate database or not.

       Properties:
           * swplatforms     - The union of all swplatforms attributes of each
                               VIB in the bulletin.
           * hwplatforms     - The union of all hwplatforms attributes of each
                               VIB in the bulletin.
           * meetssystemreqs - An instance of VIBCollection containing the
                               subset of VIBs which have a meetssystemreqs
                               property that evaluates to True.
   """

   XML_DEST_META  = 'meta'
   XML_DEST_DB    = 'db'

   def __init__(self, id, vibs=(), **kwargs):
      """Class constructor.
            Parameters:
               * id     - A string giving the unique ID of the Bulletin.
               * vibs   - A sequence containing VIB objects belonging to the
                          bulletin.
               * kwargs - A list of keyword arguments used to populate the
                          object's attributes.
            Returns: A new Bulletin instance.
      """
      assert id is not None, "id parameter cannot be None"
      self._id = id
      VIBCollection.VIBCollection.__init__(self)
      for v in vibs:
         self[v.id] = v
      self.installed = set()
      self.obsoleted = set()
      self.conflicts = set()
      self.newer = set()

      self.description    = kwargs.pop('description', '')
      self.category       = kwargs.pop('category', '')
      self.severity       = kwargs.pop('severity', '')
      self.urgency        = kwargs.pop('urgency', '')
      self.kburl          = kwargs.pop('kburl', '')
      self.contact        = kwargs.pop('contact', '')
      self.summary        = kwargs.pop('summary', '')
      self.vendor         = kwargs.pop('vendor', '')
      self.releasetype    = kwargs.pop('releasetype', '')
      self.installdate    = kwargs.pop('installdate', None)
      self.releasedate    = kwargs.pop('releasedate', None)
      self.releasedatestr = kwargs.pop('releasedatestr', '')
      self.version        = kwargs.pop('version', version.Version())
      self.compliant      = kwargs.pop('compliant', False)
      self.vibids         = kwargs.pop('vibids', set())
      self.persist        = kwargs.pop('persist', True)

      if kwargs:
         badkws = ', '.join("'%s'" % kw for kw in kwargs)
         raise TypeError("Unrecognized keyword argument(s): %s." % badkws)

      self._scanned = False
      self._meta = None
      # See comment in fromxml().
      self._legacyformat = False

   __str__  = lambda self: '<"%s">' % (self._id)
   __repr__ = __str__
   __hash__ = lambda self: hash(self._id)

   id = property(lambda self: self._id)

   # needed for id in set(Bulletin) test
   def __eq__(self, other):
      return self._id == str(other)

   def __add__(self, other):
      """Merge this bulletin with another to form a new object consisting
         of the attributes and VIB list from the newer bulletin.  Note that if
         a VIB object exists in both self and other, they will be added
         together to create a new VIB object in the returned Bulletin instance.
            Parameters:
               * other - another Bulletin instance.
            Returns: A new Bulletin instance.
            Raises:
               * RuntimeError        - If attempting to add bulletins with
                                       different IDs, or attempting to add an
                                       object that is not a Bulletin object.
               * MetadataFormatError - If attempting to add two bulletins with
                                       the same ID and release date, but which
                                       are not identical.
      """
      if not isinstance(other, self.__class__):
         msg = "Operation not supported for type %s." % other.__class__.__name__
         raise RuntimeError(msg)

      if self.id != other.id:
         raise RuntimeError("Cannot merge bulletins with different IDs.")

      metaattrs = ('description', 'summary', 'vendor', 'version', 'category',
                   'severity', 'urgency', 'kburl', 'contact', 'releasetype',
                   'releasedate', 'releasedatestr')

      samevibs = set(self) == set(other)

      if self.releasedate > other.releasedate:
         newer = self
         older = other
      elif self.releasedate < other.releasedate:
         newer = other
         older = self
      elif self._legacyformat and not other._legacyformat:
         # See comment in fromxml().
         newer = other
         older = self
      elif other._legacyformat and not self._legacyformat:
         newer = self
         older = other
      else:
         for attr in metaattrs:
            if getattr(self, attr) != getattr(other, attr):
               msg = ("Duplicate definitions of bulletin %s with unequal "
                      "attributes." % self.id)
               raise errors.MetadataFormatError(None, msg)
         if not samevibs:
            msg = ("Duplicate definitions of bulletin %s with unequal VIB "
                   "lists." % self.id)
            raise errors.MetadataFormatError(None, msg)
         # Doesn't really matter which one is 'newer' at this point...
         newer = self
         older = other

      rc = Bulletin(self.id)

      for attr in metaattrs:
         setattr(rc, attr, getattr(newer, attr))

      rc._meta = newer._meta

      # If the two have the same set of VIBs, then compliant, installdate,
      # and persist attributes for one are valid for both.  Otherwise, just
      # use the values from 'newer', since that is where the VIB list will
      # come from.
      if samevibs:
         rc.compliant = newer.compliant or older.compliant
         rc.installdate = max(newer.installdate, older.installdate)
         rc.persist = newer.persist or older.persist
      else:
         rc.compliant = newer.compliant
         rc.installdate = newer.installdate
         rc.persist = newer.persist

      # The VIB list comes from newer, but if the same VIB IDs are in older,
      # then we will merge the two VIB instances together.
      for v in newer.values():
         rc[v.id] = v
         other = older.get(v.id)
         if other:
            rc[v.id] += other

      return rc

   def setscanned(self, value):
      if not value:
         self.installed = set()
         self.obsoleted = set()
         self.conflicts = set()
         self.newer = set()
      self._scanned = value

   scanned = property(lambda self: self._scanned, setscanned)

   def scan(self):
      """Fill installed, obsoleted, newer, conflicts."""
      self.scanned = False
      for vib in self.meetssystemreqs.values():
         if vib.state.INSTALLED in vib.state:
            self.installed.add(vib)
         elif vib.state.RETIRED in vib.state:
            self.obsoleted.add(vib)
         else:
            self.newer.add(vib)
         if vib.conflictswithhost:
            self.conflicts.add(vib)
      compliant = self.ishostcompliant()
      if compliant and not self.compliant:
         # We are compliant now, but we weren't before. Update timestamp.
         # This has (at least for now) the side effect that this bulletin's
         # meta-data will be persisted to disk if it is in a VibDB object.
         self.installdate = time.time()
      self.compliant = compliant
      self.scanned = True

   def isfulfilledbyvibs(self, vibs, considerhost = True):
      """Determine if this bulletin fulfilled by a set of VIBs.
         A bulletin is considered fulfilled by the VIBs if each VIB
         contained within it meets any of the following criteria:
            1.  It is obsoleted by one of the VIBs passed in as a parameter to
                this method.
            2.  It is (directly) one of the VIBs passed in as a parameter to
                this method.
            3.  It is obsoleted by the host, and considerhost is True.
         Parameters:
            * vibs         - A set of vibs to consider when evaluating whether
                             a bulletin is fulfilled or not.
            * considerhost - See criteria #3, above.
         Returns: True if the bulletin is filfilled, False otherwise
      """
      fulfilled = set()
      ourvibs = set(self.keys())

      # It is obsoleted by one of the VIBs passed in as a parameter
      for vib in vibs:
         fulfilled |= ourvibs & vib.oldervibs

      # It is (directly) one of the VIBs passed in as a parameter
      fulfilled |= ourvibs & set(vibs)

      # It is obsoleted by or installed on the host, and considerhost is True.
      if considerhost:
         fulfilled |= set(self.obsoleted) | set(self.installed)

      return fulfilled == ourvibs

   def ishostcompliant(self):
      compliant_vibs = set(v for v in self.meetssystemreqs.values()
                           if not v.state.UNINSTALLED in v.state)
      return self.isfulfilledbyvibs(compliant_vibs, True)

   def toxml(self, target=XML_DEST_DB):
      """Expresses Bulletin attributes as an ElementTree element.
            Parameters:
               * target - One of XML_DEST_META or XML_DEST_DB, specifying the
                          intended target of the output. The target controls
                          which attributes are expressed in the ElementTree
                          element.
            Returns: An object conforming to the ElementTree element API.
      """
      if self._meta is not None:
         meta = self._meta
      else:
         meta = Descriptor.Bulletin()

      meta.writeFullVibs = (target == self.XML_DEST_META)

      meta.id          = self._id
      meta.description = self.description
      meta.summary     = self.summary
      meta.vendor      = self.vendor
      meta.releaseType = self.releasetype
      meta.category    = self.category
      meta.severity    = self.severity
      meta.urgency     = self.urgency
      meta.kbUrl       = self.kburl
      meta.contact     = self.contact
      if self.installdate:
         if target == self.XML_DEST_DB:
            meta.installDate = str(self.installdate)
         else:
            dt = datetime.datetime.utcfromtimestamp(self.installdate)
            meta.installDate = dt.isoformat()
      if self.releasedate:
         dt = datetime.datetime.utcfromtimestamp(self.releasedate)
         meta.releaseDate = dt.isoformat()
      elif self.releasedatestr:
         meta.releaseDate = self.releasedatestr

      if target == self.XML_DEST_DB:
         meta.compliant = self.compliant

      node = meta.ToXml()
      viblist = node.find('vibList')
      if target == self.XML_DEST_DB:
         vibids = set(self.vibids)
         vibids.update(self.keys())
         for vibid in vibids:
            etree.SubElement(viblist, 'vibID').text = vibid
      elif target == self.XML_DEST_META:
         for vib in self.values():
            viblist.append(vib.toxml(target=vib.XML_DEST_META))

      return node

   @classmethod
   def fromxml(cls, xml, **kwargs):
      """Creates a Bulletin instance from XML.
            Parameters:
               * xml    - Must be either an instance of ElementTree, or a
                          string of XML-formatted data.
               * kwargs - Initialize constructor arguments from keywords.
                          Primarily useful to provide default or required
                          arguments when XML data is from a template.
            Returns: A new Bulletin object.
            Raises:
               * MetadataFormatError - A string was passed, but could not be
                                       parsed as valid XML data; or the parsed
                                       XML did not include required tags or
                                       attributes, or was otherwise not in the
                                       expected format.
               * VibFormatError      - One or more elements in <vibList> could
                                       not be properly parsed.
      """
      if etree.iselement(xml):
         node = xml
      else:
         try:
            node = etree.fromstring(xml)
         except Exception, e:
            msg = "Error parsing XML data: %s." % e
            raise errors.MetadataFormatError(None, msg)

      meta = Descriptor.Bulletin(node)

      if meta.vibIDs:
         kwargs['vibids'] = set()
         for vibid in meta.vibIDs:
            kwargs['vibids'].add(vibid)
      if meta.vibs:
         kwargs['vibs'] = set()
         for vib in meta.vibs:
            kwargs['vibs'].add(vib)

      kwargs['id']          = meta.id
      kwargs['description'] = meta.description
      kwargs['summary']     = meta.summary
      kwargs['vendor']      = meta.vendor
      kwargs['releasetype'] = meta.releaseType
      kwargs['compliant']   = meta.compliant
      kwargs['category']    = meta.category
      kwargs['severity']    = meta.severity
      kwargs['urgency']     = meta.urgency
      kwargs['kburl']       = meta.kbUrl
      kwargs['contact']     = meta.contact

      legacyformat = False

      if meta.installDate:
         try:
            kwargs['installdate'] = float(meta.installDate)
         except:
            # If meta.installDate is set, but it is not a float value, we
            # assume that the bulletin was written by a legacy version of
            # esxupdate or vima-update. This becomes important in the __add__
            # method. See PR 482994.
            legacyformat = True
            try:
               kwargs['installdate'] = util.isotofloat(meta.installDate)
            except:
               pass
      if meta.releaseDate:
         try:
            releasedate = util.isotofloat(meta.releaseDate)
            kwargs['releasedate'] = releasedate
            # If we're able to parse the <releaseDate> tag into a time stamp,
            # set the releasedatestr from that value. This way the
            # releasedatestr will compare the same for two objects, even if
            # the original <releaseDate> tag in the XML represented the same
            # time in two different formats.
            dt = datetime.datetime.utcfromtimestamp(releasedate)
            kwargs['releasedatestr'] = dt.isoformat()
         except:
            kwargs['releasedatestr'] = meta.releaseDate

      if 'id' not in kwargs:
         msg = "Could not parse valid bulletin ID from XML data."
         raise errors.MetadataFormatError(None, msg)

      bulletin = cls(**kwargs)
      bulletin._meta = meta
      bulletin._legacyformat = legacyformat
      return bulletin


class BulletinCollection(dict):
   """This class represents a collection of Bulletin objects and provides
      methods and properties for modifying the collection and determining
      the relationships between members.
   """
   def sync(self, other):
      """Updates the common keys in this object and other to point to the same
         Bulletin objects. It does not affect values that are not common
         between the two.
            Parameters:
               * other - Another VIBCollection instance.
      """
      for key in set(self.keys()) & set(other.keys()):
         if id(self[key]) != id(other[key]):
            b = self[key] + other[key]
            self[key] = b
            other[key] = b

   def __add__(self, other):
      """Merge this collection with another to form a new collection consisting
         of the union of Bulletins from both.
            Parameters:
               * other - another BulletinCollection instance.
            Returns: A new BulletinCollection instance.
      """
      new = BulletinCollection(self)
      new.update(self)
      for b in other.values():
         if b.id in new and id(b) != id(new[b.id]):
            new[b.id] += b
         else:
            new[b.id] = b
      return new

   @property
   def meetssystemreqs(self):
      return BulletinCollection((b.id, b) for b in self.values()
                                if b.meetssystemreqs)

   @property
   def failssystemreqs(self):
      return BulletinCollection((b.id, b) for b in self.values()
                                if not b.meetssystemreqs)

   @property
   def compliant(self):
      rc = BulletinCollection()
      for b in self.values():
         if b.compliant:
            rc[b.id] = b
      return rc

   @property
   def succinct(self):
      rc = BulletinCollection()
      for b in self.values():
         b.scan()
         if not b.compliant or not b.installed:
            continue
         add = True
         for other in rc.values():
            if other.installed < b.installed:
               del rc[other.id]
            elif other.installed > b.installed:
               add = False
               break
         if add:
            rc[b.id] = b
      return rc

   def fulfilledbyvibs(self, vibs, considerhost = True):
      """Return all of the Bulletin objects in this collection which are
         fulfilled by the given set of VIBs. A bulletin is considered fulfilled
         by the VIBs if each VIB contained within it meets any of the following
         criteria:
            1.  It is obsoleted by one of the VIBs passed in as a parameter to
                this method.
            2.  It is (directly) one of the VIBs passed in as a parameter to
                this method.
            3.  It is obsoleted by the host, and considerhost is True.
         Parameters:
            * vibs         - A set of vibs to consider when evaluating whether
                             a bulletin is fulfilled or not.
            * considerhost - See criteria #3, above.
         Returns: A BulletinCollection instance.
      """
      rc = BulletinCollection()
      for b in self.values():
         if b.isfulfilledbyvibs(vibs, considerhost):
            rc[b.id] = b
      return rc

   def hostcompliant(self):
      rc = BulletinCollection()
      for b in self.values():
         if b.ishostcompliant():
            rc[b.id] = b
      return rc
