#######################################################################
# Copyright (C) 2009 VMWare, Inc.
# All Rights Reserved
########################################################################
#
# Database.py
#

__all__ = ['Database']

import fcntl
import logging
import os
import util.zipfile

try:
   from lxml import etree
except ImportError:
   try:
      import xml.etree.cElementTree as etree
   except ImportError:
      import cElementTree as etree

import Bulletin
import Vib
import VIBScanner
import errors

log = logging.getLogger(__name__)

class Database(VIBScanner.VIBScanner):
   """Implements a databse to track VIB state and bulletins. Information stored
      in the database includes:
         - VIBs being currently installed.
         - VIBs previously installed.
         - bulletins previously installed.

      The database currently consists of two files in a directory:
         - <dbpath>/vibs.xml
         - <dbpath>/bulletins.zip
   """
   def __init__(self, dbpath='/etc/vmware/esxupdate', dbcreate=True):
      """Class constructor.
            Parameters:
               * dbpath   - The location of the esxupdate database.
               * dbcreate - If True, the database will be created if it does
                            not exist. Note that if database files exist in
                            the directory given by dbpath, but cannot be
                            properly read or parsed, an exception will be
                            raised regardless of this parameter's value.
            Returns: A new Database instance.
            Raises:
               * DatabaseIOError - If dbpath cannot be created; if dbpath does
                                   not exist and dbcreate is False; if dbpath
                                   exists but is not a directory; or if the
                                   database cannot be read from or written to
                                   persistent storage.
      """
      VIBScanner.VIBScanner.__init__(self, usepkgdb=True)

      self._vibdb = os.path.join(dbpath, 'vibs.xml')
      self._bulletindb = os.path.join(dbpath, 'bulletins.zip')
      self._serialfile = os.path.join(dbpath, 'serial')

      self._dbpath = dbpath

      self._serialfobj = None
      self._serialno = None

      if not os.path.exists(dbpath):
         if not dbcreate:
            msg = "Database directory '%s' does not exist." % dbpath
            raise errors.DatabaseIOError(dbpath, msg)
      elif not os.path.isdir(dbpath):
         msg = "Database path '%s' exists, but is not a directory." % dbpath
         raise errors.DatabaseIOError(dbpath, msg)

      # Allow VIBs in database to be scanned for obsolescence
      self.pkgscanner.AddVibCollectionChannel(self.vibs)

   dbpath   = property(lambda self: self._dbpath)

   def lock(self):
      """Locks the database. No reading or writing will be possible while
         the database remains locked.  If the database is currently locked,
         this method blocks until the lock can be obtained.
            Raises:
               * DatabaseIOError - If the serial file cannot be opened or
                                   locked.
      """
      try:
         # If we already have an open file descriptor for the serial file,
         # do not open another one. See PR 457302.
         if self._serialfobj:
            fd = self._serialfobj.fileno()
         else:
            fd = os.open(self._serialfile, os.O_RDWR|os.O_CREAT)
            self._serialfobj = os.fdopen(fd, 'rb+')
      except Exception, e:
         msg = "Error opening database serial file: %s." % e
         raise errors.DatabaseIOError(self._dbpath, msg)

      try:
         fcntl.flock(fd, fcntl.LOCK_EX)
      except Exception, e:
         msg = "Error locking database serial file: %s" % e
         raise errors.DatabaseIOError(self._dbpath, msg)

   def unlock(self):
      """Unlocks the database.
            Raises:
               * DatabaseIOError - If the file cannot be unlocked.
      """
      if self._serialfobj is None:
         return
      try:
         fcntl.flock(self._serialfobj, fcntl.LOCK_UN)
         self._serialfobj.close()
         self._serialfobj = None
      except Exception, e:
         msg = "Error unlocking database serial file: %s" % e
         raise errors.DatabaseIOError(self._dbpath, msg)

   def _getserialno(self):
      # You MUST call lock() before calling this method, otherwise _serialfobj
      # will not be a valid open file.
      try:
         self._serialfobj.seek(0, 0)
         # We don't actually expect the serial number to be 512 digits.  That's
         # just there to keep us from reading a whole bunch in the case that
         # this file contains something we don't expect.
         data = self._serialfobj.read(512).strip()
         if not data:
            return 0
         return int(data)
      except ValueError, e:
         msg = "Invalid content in database serial file."
         raise errors.DatabaseFormatError(msg)
      except Exception, e:
         msg = "Error reading serial file: %s." % e
         raise errors.DatabaseIOError(self._dbpath, msg)

   def _loadvibs(self):
      if not os.path.exists(self._vibdb):
         msg = "VIB database not loaded: '%s' does not exist." % self._vibdb
         log.debug(msg)
         return

      try:
         root = etree.parse(self._vibdb).getroot()
      except EnvironmentError, e:
         msg = "Unable to read VIB database: %s" % e
         raise errors.DatabaseIOError(self._dbpath, msg)
      except Exception, e:
         msg = "Unable to parse VIB database: %s" % e
         raise errors.DatabaseFormatError(self._dbpath, msg)

      for child in root.getchildren():
         if child.tag == 'vib':
            try:
               vib = Vib.VIB.fromxml(child, Vib.VIB.XML_DEST_DB)
               if vib.id in self.vibs:
                  self.vibs[vib.id] += vib
               else:
                  self.vibs[vib.id] = vib
            except Exception, e:
               msg = "Error parsing VIB from VIB database: %s" % e
               raise errors.DatabaseFormatError(self._dbpath, msg)

   def _loadbulletins(self):
      # zipfile has a bug, returning IOError if .zip size is less than 22
      # bytes, due to a seek(-22,2).
      if (not os.path.exists(self._bulletindb)
          or os.path.getsize(self._bulletindb) < 22):
         msg = ("Bulletin database not loaded: '%s' does not exist or is empty"
                % self._bulletindb)
         log.debug(msg)
         return

      try:
         z = util.zipfile.ZipFile(self._bulletindb, 'r')
      except Exception, e:
         msg = "Error reading bulletin database format: %s" % e
         raise errors.DatabaseFormatError(self._dbpath, msg)

      for name in z.namelist():
         try:
            root = etree.fromstring(z.read(name))
            bulletin = Bulletin.Bulletin.fromxml(root)
         except EnvironmentError, e:
            msg = "Error reading bulletin database: %s" % e
            raise errors.DatabaseIOError(self._dbpath, msg)
         except errors.MetadataFormatError, e:
            msg = "Could not parse bulletin from database: %s" % e.message
            raise errors.DatabaseFormatError(self._dbpath, msg)
         except Exception, e:
            msg = "Error reading bulletin database format: %s" % e
            raise errors.DatabaseFormatError(self._dbpath, msg)

         self.bulletins[bulletin.id] = bulletin
         for vibid in bulletin.vibids:
            if vibid in self.vibs:
               bulletin[vibid] = self.vibs[vibid]
            else:
               log.warn("Bulletin '%s' contains a reference to VIB '%s' which "
                        "was not found in VIB database." %
                        (bulletin.id, vibid))
         # You never know:
         self.vibs.sync(bulletin)

   def load(self, keeplock = False):
      """Populates the bulletins and vibs attributes of this object with
         information from the database. It is safe to call this method
         repeatedly to update this object when the database has changed.
            Parameters:
               * keeplock - If True, database remains locked after loading,
                            until the unlock method is called. Callers should
                            use this parameter if they intend to modify the
                            database. Note that the database will be unlocked
                            if an exception occurs, regardless of the value of
                            this parameter.
            Returns:  True if the database has been changed, False otherwise.
            Exceptions:
               * DatabaseFormatError - If vibs.xml or bulletins.zip cannot be
                                       parsed, or format is otherwise invalid.
               * DatabaseIOError     - If database lock cannot be obtained, or
                                       a file cannot be read.
      """
      self.lock()

      try:
         serialno = self._getserialno()
      except:
         self.unlock()
         raise

      updated = self._serialno is None or self._serialno != serialno

      if updated:
         try:
            self._loadvibs()
            self._loadbulletins()
            # Only assign self._serialno if we got this far without an
            # exception.
            self._serialno = serialno
         except:
            self.unlock()
            raise

      if not keeplock:
         self.unlock()

      return updated

   def _writenewbulletindb(self):
      newbulletindb = self._bulletindb + '.new'
      try:
         z = util.zipfile.ZipFile(newbulletindb, 'w')
      except Exception, e:
         msg = "Error creating temporary bulletin database: %s." % e
         raise errors.DatabaseIOError(self._dbpath, msg)

      try:
         for bulletin in self.bulletins.values():
            node = bulletin.toxml(bulletin.XML_DEST_DB)
            z.writestr(bulletin.id, etree.tostring(node))
         z.close()
      except Exception, e:
         msg = "Error writing to temporary bulletin database: %s" % e
         raise errors.DatabaseIOError(self._dbpath, msg)
      return newbulletindb

   def _writenewvibdb(self):
      newvibdb = self._vibdb + '.new'
      root = etree.Element('vib-index')
      tree = etree.ElementTree(element=root)
      for vib in self.vibs.values():
         root.append(vib.toxml(vib.XML_DEST_DB))
      try:
         tree.write(newvibdb)
      except Exception, e:
         msg = "Error writing temporary VIB database: %s" % e
         raise errors.DatabaseIOError(self._dbpath, msg)
      return newvibdb

   def _updateserialno(self, serialno):
      # You MUST call lock() before calling this method, otherwise _serialfobj
      # will not be a valid open file.
      try:
         self._serialfobj.truncate(0)
         self._serialfobj.seek(0, 0)
         self._serialfobj.write(str(serialno))
         self._serialno = serialno
      except Exception, e:
         msg = "Error updating database serial number: %s" % e
         raise errors.DatabaseIOError(self._dbpath, msg)

   def sync(self, other=None):
      #
      # Remove 'staged' state if VIB not extracted at packagelocation
      for vib in self.vibs.values():
         if vib.state.STAGED in vib.state:
            if not vib.packagelocation or \
               not os.path.exists(vib.packagelocation):
               log.debug('VIB %s staged state was reverted due to missing '
                         'package file expected at %s' % (vib.id,
                                                          vib.packagelocation))
               vib.state.discard(vib.state.STAGED)
               vib.remotelocation = ''
      # Let other pick up any state changes
      VIBScanner.VIBScanner.sync(self, other)

   def save(self, keeplock = False):
      """Writes the information currently in the object back to the database.
            Parameters:
               * keeplock - If True, database remains locked after writing,
                            until the unlock method is called. Callers should
                            use this parameter if they intend to make
                            additional changes to the database. Note that the
                            database will be unlocked if an exception occurs,
                            regardless of the value of this parameter.
            Raises:
               * DatabaseIOError     - On a failure to write to a file, or if
                                       database has changed since last read
                                       (serial number has been incremented).
      """
      self.lock()

      try:
         serialno = self._getserialno()
         if serialno != self._serialno:
            msg = ("Refusing to save database; database has been modified since "
                   "previous read.")
            raise errors.DatabaseIOError(self._dbpath, msg)
         newbulletindb = self._writenewbulletindb()
         newvibdb = self._writenewvibdb()
         serialno += 1
         self._updateserialno(serialno)
      except:
         self.unlock()
         raise

      try:
         os.rename(newbulletindb, self._bulletindb)
      except Exception, e:
         msg = "Error committing changes to bulletin database: %s" % e
         self.unlock()
         raise errors.DatabaseIOError(self._dbpath, msg)

      try:
         os.rename(newvibdb, self._vibdb)
      except Exception, e:
         msg = "Error committing changes to VIB database: %s" % e
         self.unlock()
         raise errors.DatabaseFormatError(self._dbpath, msg)

      if not keeplock:
         self.unlock()
