########################################################################
# Copyright (C) 2008 VMWare, Inc.
# All Rights Reserved
########################################################################
#
# Descriptor.py
#   XmlElement, Bulletin, DescriptorBase, DescriptorFile, VibMeta,
#   SystemReqs, Platforms
#

try:
   from lxml.etree import ElementTree, Element, iselement, XMLSyntaxError
   xmlerrclass = XMLSyntaxError
except ImportError:
   try:
      from xml.etree.cElementTree import ElementTree, Element, iselement
   except ImportError:
      from cElementTree import ElementTree, Element, iselement
   from xml.parsers.expat import ExpatError
   xmlerrclass = ExpatError

from datetime import datetime

from util.xmlUtils import *
from const import RELEASE_TYPES, INT_PRODUCTS, EXT_PRODUCTS

import Vib
import util
import re

#
# Global constants
#
DESCVER = "3.0"
NOTIFICATION_BULLETIN_TAGS=['recalledVibList', 'recalledBulletinList', 'contentBody', 'resolvedRecalls']


# -----------------------------------------------------------------------------
class BadBulletinError(Exception):
   def __init__(self, filename, msg):
      Exception.__init__(self, "Error parsing file %s: %s"
            % (filename, msg))

class BadDescriptorError(Exception):
   def __init__(self, filename, msg):
      Exception.__init__(self, "Error parsing file %s: %s"
            % (filename, msg))

class BadPlatformError(Exception):
   def __init__(self, filename, msg):
      if filename != "":
         Exception.__init__(self, "Error parsing file %s :%s" % (filename, msg))
      else:
         Exception.__init__(self, "%s" % msg)

# -----------------------------------------------------------------------------

class XmlElement(object):
   """ Abstract base class for xml parsing tasks."""

   def __init__(self):
      if self.__class__ is XmlElement:
         raise NotImplementedError

   def _reset(self):
      raise NotImplementedError

   def Reset(self):
      self._reset()

   def _fromXml(self, root):
      raise NotImplementedError

   def _toXml(self, root):
      raise NotImplementedError

   def FromXml(self, root):
      """ Populates instance members from root Element node."""
      assert iselement(root), "Expecting Element type but got %s" % (type(root))
      self._reset()
      self._fromXml(root)

   def ToXml(self):
      """ Return an Element node representing the instance."""
      return self._toXml()

   def Write(self, xmlfile):
      """ Write the instance to a XML file."""
      tree = ElementTree(element=self._toXml())
      IndentElementTree(tree.getroot())
      tree.write(xmlfile)

   @staticmethod
   def _findtext(root, path, default=''):
      """ Returns the stripped text from the first XML node found from
          Element root and XPath expression path.
          If the element cannot be found, or is found but there is no
          text, then default is returned.
          Gets around bug in lxml.etree which returns None if the element
          is fouund but there is no text.
      """
      txt = root.findtext(path)
      if txt:
         return txt.strip()
      else:
         return default

class Bulletin(XmlElement):
   """ Represents a bulletin in vmware.xml, or a standalone bulletin.xml.
       In one case, <vibList> contains <vibID>s, in the other, the actual
       <vib> metadata.  In the first case, self.vibIDs will be populated;
       in the second, both self.vibIDs and self.vibs will.
   """

   def __init__(self, node=None, writeFullVibs=False, bulletinFile=None):
      """ Construct a Bulletin instance from the XML element node or
          bulletin file if provided; otherwise a blank Bulletin.
       * writeFullVibs  - If true, write <vibList> with full <vib>
                          metadata, if self.vibs is not empty.
                          Otherwise, just write out vibIDs.
      """
      self.Reset()
      self.writeFullVibs = writeFullVibs

      if bulletinFile:
         self._filename = bulletinFile
         fp = open(bulletinFile, 'r')

         try:
            try:
               tree = ElementTree(file=fp)
            except xmlerrclass, e:
               raise BadBulletinError(self._filename, str(e))
         finally:
            fp.close()

         self._fromXml(tree.getroot())

      elif node is not None:
         self._fromXml(node)

   # needed for id in set(Bulletin) test
   def __hash__(self):
      if self.id:
         return self.id.__hash__()
      else:
         raise ValueError('Cannot hash Bulletin instance without id.')

   def __str__(self):
      return self.id

   def __repr__(self):
      return 'Bulletin<"%s">' % (self.id)

   # needed for id in set(Bulletin) test
   def __eq__(self, other):
      return str(self) == str(other)

   def SetInstallDate(self):
      """ Sets installDate to the current time in ISO8601 format.
          The time is in local timezone with a trailing +HH:MM offset.
      """
      self.installDate = GetCurrentDateTime().isoformat()

   def _reset(self):
      self.id = ''
      self.summary = ''
      self.description = ''
      self.vendor = ''
      self.contact = ''
      self.severity = ''
      self.category = ''
      self.urgency = ''
      self.vendorProduct = ''
      self.releaseType = RELEASE_TYPES.patch
      self.componentName = None
      self.platforms = Platforms()
      self.vibIDs = set()
      self.vibs = set()
      self.kbUrl = ''
      self.releaseDate = GetCurrentDateTime().isoformat()
      self.installDate = ''
      self.compliant = False
      self.notificationTags = set()

   def _fromXml(self, root):
      self.id = self._findtext(root, 'id')
      self.summary = self._findtext(root, 'summary')
      self.description = self._findtext(root, 'description')
      self.vendor = self._findtext(root, 'vendor')
      self.contact = self._findtext(root, 'contact')
      self.severity = self._findtext(root, 'severity').lower()
      self.category = self._findtext(root, 'category').lower()
      self.urgency = self._findtext(root, 'urgency').lower()
      self.vendorProduct = self._findtext(root, 'vendorProduct').lower()
      self.releaseType = self._findtext(root, 'releaseType', 'patch').lower()
      self.componentName = self._findtext(root, 'componentName')

      platforms = root.find('platforms')
      self.platforms = Platforms(platforms)
      vibs = root.find('vibList')
      if vibs is not None:
         for vibID in vibs.findall('vibID'):
            if vibID.text:
               self.vibIDs.add(vibID.text)
         for vibelem in vibs.findall('vib'):
            vib = Vib.VIB.fromxml(vibelem, Vib.VIB.XML_DEST_META)
            self.vibs.add(vib)
            self.vibIDs.add(vib.id)

      self.kbUrl =  self._findtext(root, 'kbUrl')
      releaseDate =  self._findtext(root, 'releaseDate')
      if releaseDate:
         self.releaseDate = releaseDate
      self.installDate =  self._findtext(root, 'installDate')
      self.compliant = StrToBool(self._findtext(root, 'compliant'))

      for tag in NOTIFICATION_BULLETIN_TAGS:
	 elem = root.find(tag)
	 if elem:
	    temp_elem = Element(tag)
	    temp_elem = elem
	    self.notificationTags.add(temp_elem)

   def _checkAttr(self, attr):
      if not getattr(self, attr):
         raise BadBulletinError(self._filename, '<%s> is required in bulletin'\
            % attr)

   def CheckSchema(self):

      self._checkAttr('id')
      self._checkAttr('summary')
      self._checkAttr('description')
      self._checkAttr('vendor')
      self._checkAttr('contact')

      self._checkAttr('releaseType')
      if self.releaseType not in RELEASE_TYPES:
         raise BadBulletinError(self._filename, \
                  "Unrecognized value '%s'.  releaseType must be one of %s." \
                  % (self.releaseType, list(RELEASE_TYPES)))

      if getattr(self, 'platforms'):
         try:
            self.platforms.CheckSchema()
         except BadPlatformError, e:
            raise BadBulletinError(self._filename, '%s' % e)

      #
      # vibIDs is not required for notification bulletins
      #
      if not self.releaseType == RELEASE_TYPES.notification and self.vibIDs == set([]):
         raise BadBulletinError(self._filename, \
                  '<vibList> is required in bulletin')

      self._checkAttr('kbUrl')
      self._checkAttr('releaseDate')
      try:
         util.isotofloat(self.releaseDate)
      except:
         msg = ("<releaseDate> must be specified as a calendar date and time "
                "of day in ISO 8601:2004 extended format:\n"
                "   YYYY-MM-DDThh:mm:ss\n"
                "   YYYY-MM-DDThh:mm:ssZ\n"
                "   YYYY-MM-DDThh:mm:ss[+-]hh:mm\n"
                "   YYYY-MM-DDThh:mm:ss[+-]hh\n"
                "Example: 2009-10-27T14:21:30+09:00")
         raise BadBulletinError(self._filename, msg)

   def _toXml(self):
      root = Element('bulletin')
      AppendElement(root, 'id', self.id)
      AppendElement(root, 'vendor', self.vendor)
      AppendElement(root, 'summary', self.summary)
      AppendElement(root, 'severity', self.severity)
      AppendElement(root, 'category', self.category)
      AppendElement(root, 'urgency', self.urgency)
      AppendElement(root, 'vendorProduct', self.vendorProduct)
      AppendElement(root, 'releaseType', self.releaseType)
      SetElement(root, 'componentName', self.componentName)
      AppendElement(root, 'description', self.description)
      AppendElement(root, 'kbUrl', self.kbUrl)
      AppendElement(root, 'contact', self.contact)
      AppendElement(root, 'releaseDate', self.releaseDate)
      SetElement(root, 'installDate', self.installDate)
      # We don't write this out if False, because we don't want it to show up
      # in bulletin meta-data generated by build scripts.  This is OK, since
      # the non-existence of this tag is assumed to mean 'false'.  Really,
      # though, this should be sub-classed for bulletins.xml.
      if self.compliant:
         AppendBoolElement(root, 'compliant', self.compliant)

      root.append(self.platforms.ToXml())

      vibs = Element('vibList')
      if self.writeFullVibs:
         for vib in self.vibs:
            vibs.append(vib.toxml(Vib.VIB.XML_DEST_META))
      else:
         for vibID in self.vibIDs:
            AppendElement(vibs, 'vibID', vibID)
      root.append(vibs)

      for tag in self.notificationTags:
	 root.append(tag)

      return root


# -----------------------------------------------------------------------------
class DescriptorBase(XmlElement):
   """Creates, parses, modifies, writes release descriptors."""
   #
   # I wonder if there is a cleaner API than the init for
   # 3.5 Descriptor, where we have to do tons of type-checking.
   #
   def __init__(self, root=None):
      """ Create a descriptor from a descriptor element node."""
      self.Reset()
      if root is not None:
         self._fromXml(root)

   def _reset(self):
      """ Resets all members of a descriptor to default values."""
      self.vibVersion = ''
      self.vibID = ''
      self.name = ''
      self.vendor = ''
      self.version = ''
      self.buildDate = ''
      self.vibType = ''
      self.systemReqs = SystemReqs()
      self.rebootRequired = True
      self.hostdRestart = True
      self.visorDestination = ''
      self.pkgfile = ''
      self.swTags = []
      self.groups = {}

   def _fromXml(self, root):
      """ Populates the descriptor from an Element node."""
      self.vibVersion = self._findtext(root, 'vibVersion')
      self.vibID = self._findtext(root, 'vibID')
      self.name = self._findtext(root, 'name')
      self.vendor = self._findtext(root, 'vendor')
      self.version = self._findtext(root, 'version')
      self.buildDate = self._findtext(root, 'buildDate')
      self.vibType = self._findtext(root, 'vibType')

      systemReqs = root.find('systemReqs')
      if systemReqs is not None:
         self.systemReqs = SystemReqs(systemReqs)

      self.rebootRequired = StrToBool(
            self._findtext(root, './postInstall/rebootRequired'))
      self.hostdRestart = StrToBool(
            self._findtext(root, './postInstall/hostdRestart'))
      self.visorDestination = self._findtext(root, 'visorDestination')
      self.pkgfile = self._findtext(root, 'pkgfile')

      tags = root.findall('./softwareTags/tag')
      for t in tags:
         self.swTags.append(t.text.strip())

      groupnodes = root.findall('./softwareTags/group')
      for g in groupnodes:
         val = g.get('value', '').strip()
         grouplist = self.groups.setdefault(g.text.strip(), [])
         if val:
            grouplist.append(val)

   def _toXml(self):
      root = Element('descriptor')

      AppendElement(root, 'vibVersion', self.vibVersion)
      AppendElement(root, 'vibID', self.vibID)
      AppendElement(root, 'name', self.name)
      AppendElement(root, 'vendor', self.vendor)
      AppendElement(root, 'version', self.version)
      AppendElement(root, 'buildDate', self.buildDate)
      AppendElement(root, 'vibType', self.vibType)
      root.append(self.systemReqs.ToXml())

      postInst = Element('postInstall')
      AppendBoolElement(postInst, 'rebootRequired', self.rebootRequired)
      AppendBoolElement(postInst, 'hostdRestart', self.hostdRestart)
      root.append(postInst)
      SetElement(root, 'visorDestination', self.visorDestination)
      SetElement(root, 'pkgfile', self.pkgfile)

      if self.swTags or self.groups:
         tags = Element('softwareTags')
         for t in self.swTags:
            AppendElement(tags, 'tag', t)
         for g in self.groups:
            if self.groups[g]:
               for val in self.groups[g]:
                  AppendElement(tags, 'group', g, value=val)
            else:
               AppendElement(tags, 'group', g)
         root.append(tags)

      return root

   def UpdateBuildDate(self, dt=None):
      """ Update buildDate to current or a given datetime."""
      if dt is None:
         dt = GetCurrentDateTime()
      else:
         assert isinstance(dt, datetime) , \
            "Expecting datetime type but got %s" % (type(dt))
      dt = dt.replace(microsecond=0)
      self.buildDate = dt.isoformat()

   def UpdateVibID(self):
      """
      Construct vibID from current data members.
      """
      if self.name == '' or self.version == '' or self.vibType == '':
         raise ValueError
      else:
         self.vibID = self.vibType + '_' + self.name + '_' + self.version

   def AddSoftwareTag(self, tag):
      """
      Append to the list of tags.
      """
      self.swTags.append(tag)

# -----------------------------------------------------------------------------
class DescriptorFile(DescriptorBase):
   """ Represents the descriptor file in a vib."""

   def __init__(self, desc_file=None):
      """ Creates a Descriptor instance, optionally by reading
      from a file or file object.
      """
      self.Reset()
      self._filename = ""

      if desc_file:
         if type(desc_file) is str:
            self._filename = desc_file
            fp = open(desc_file, 'r')
         else:
            self._filename = getattr(desc_file, 'name', None)
            fp = desc_file

         try:
            try:
               tree  = ElementTree(file=fp)
            except xmlerrclass, e:
               raise BadDescriptorError(self._filename, str(e))
         finally:
            fp.close()

         self._fromXml(tree.getroot())

   def _reset(self):
      DescriptorBase._reset(self)
      self.descVersion = DESCVER

   def _fromXml(self, root):
      DescriptorBase._fromXml(self, root)
      self.descVersion = root.get('version', DESCVER).strip()

   def _toXml(self):
      root = DescriptorBase._toXml(self)
      root.tag = 'descriptor'
      root.set('version', self.descVersion)
      return root


# -----------------------------------------------------------------------------
class VibMeta(DescriptorBase):
   """ Represents entry for a bundle in metadata.zip/vmware.xml.
   Most entries are the same as descriptor but a few more are reserved
   here.
   """

   def _reset(self):
      DescriptorBase._reset(self)
      self.sourceUrl = ''
      self.relativePath = ''
      self.checksums = []
      self.packedSize = 0

   def _fromXml(self, root):
      DescriptorBase._fromXml(self, root)
      self.sourceUrl = self._findtext(root, './vibFile/sourceUrl')
      self.relativePath = self._findtext(root, './vibFile/relativePath')

      try:
         self.packedSize = int(root.findtext('./vibFile/packedSize'))
      except TypeError, err:
         #
         # XXX Raise a error?
         #
         self.packedSize = 0

      for checksum in root.findall('./vibFile/checksum'):
         item = (self._findtext(checksum, 'checksumType'),
                 self._findtext(checksum, 'checksum'))
         self.checksums.append(item)

   def _toXml(self):
      root = DescriptorBase._toXml(self)
      vibFile = Element('vibFile')
      AppendElement(vibFile, 'sourceUrl', self.sourceUrl)
      AppendElement(vibFile, 'relativePath', self.relativePath)
      AppendElement(vibFile, 'packedSize', "%d" % (self.packedSize))

      for checksum in self.checksums:
         checkelem = Element('checksum')
         AppendElement(checkelem, 'checksumType', checksum[0])
         AppendElement(checkelem, 'checksum', checksum[1])
         vibFile.append(checkelem)

      root.append(vibFile)

      root.tag = 'vib'
      return root


# -----------------------------------------------------------------------------
class SystemReqs(XmlElement):
   """ Represents <systemReqs> descriptor tag."""

   def __init__(self, root=None):
      """ Creates an instance from a systemReqs Element node """
      self.Reset()
      if root is not None:
         self._fromXml(root)

   def _fromXml(self, root):
      assert iselement(root), "Expecting Element type but got %s" % (type(root))
      for platform in root.findall('swPlatform'):
         item = (platform.get('productLineID', '').strip(),
            platform.get('version', '').strip(),
            platform.get('locale', '').strip())
         self.swplatforms.append(item)
      for hwplatform in root.findall('hwPlatform'):
         item = (hwplatform.get('vendor', '').strip(),
                 hwplatform.get('model', '').strip())
         self.hwplatforms.append(item)

      self.maintenanceMode = StrToBool(root.findtext('maintenanceMode', 'true'))

      # To be backward compatible and conservative , if the element is false, no
      # need to check the property value
      self.mModeInstall = None
      mmNode = root.find('maintenanceMode')
      if self.maintenanceMode and mmNode is not None:
         installValue = mmNode.get('install', '')
         if installValue:
            self.mModeInstall = StrToBool(installValue)

   def ToXml(self, mmodeOverride=None):
      return self._toXml(mmodeOverride)

   def _toXml(self, mmodeOverride=None):
      root = Element('systemReqs')
      for platform in self.swplatforms:
         AppendElement(root, 'swPlatform', None, productLineID=platform[0],
               version=platform[1], locale=platform[2])
      for hwplatform in self.hwplatforms:
         AppendElement(root, 'hwPlatform', None, vendor=hwplatform[0],
                       model=hwplatform[1])

      if mmodeOverride is not None:
         mmodeOut = mmodeOverride
      else:
         mmodeOut = self.maintenanceMode

      if self.mModeInstall is None:
         AppendBoolElement(root, 'maintenanceMode', mmodeOut)
      else:
         AppendBoolElement(root, 'maintenanceMode', mmodeOut,
               install=str(self.mModeInstall))
      return root

   def MatchSWPlatform(self, product, version, locale=''):
      """ Do we support one of the <swPlatform>'s?
      Args:
         product - The productID (currently 'esx' or 'esxEmbedded').
                   Note:  Match is case-insensitive.
         version - The ESX version in x.y.z notation.
         locale  - The locale.  If not specified, it is ignored, and matches
                   will be found based on the productID and version.
      Returns:
         True if there exists a matching swPlatform in this metadata.
         False if there is no match.
      """
      matches = list()
      for swproduct, swversion, swlocale in self.swplatforms:
         # If product doesn't match, it's a no-go.
         if swproduct.strip().lower() != product.strip().lower():
            continue

         # Version matching is a bit more complex.
         swversionparts = [part.strip() for part in swversion.split('.')]
         versionparts = [part.strip() for part in version.split('.')]
         # 'x' in 'x.y.z' as a wildcard is not appropriate.
         if '*' in (swversionparts[0], versionparts[0]):
            continue
         # substitute wildcards.
         for i in range(len(swversionparts)):
            if swversionparts[i] == '*' and i < len(versionparts):
               swversionparts[i] = versionparts[i]
         # If swversionparts is x.y.z, and versionparts is only x.y, that's bad
         # unless z == '*':
         if [x for x in swversionparts[len(versionparts):] if x != '*']:
            continue
         # If we've gotten this far, it's ok to match the minimum length of
         # either version.  If swversion is x.y and platform version is x.y.z,
         # it's ok just to match x and y.  Likewise, if swversion is x.y.z and
         # version is x.y, then we know that z must be a '*'.
         minlen = min(len(swversionparts), len(versionparts))
         if swversionparts[:minlen] != versionparts[:minlen]:
            continue

         # Locale matching.
         swlocale = swlocale.strip().lower()
         locale = locale.strip().lower()
         # Substitute wildcards (assume '' is a wildcard).
         if not (swlocale in ('', '*') or locale == '') and swlocale != locale:
            continue

         # If we've made it through all of the conditionals, we have a match.
         matches.append((swproduct, swversion, swlocale))

      return matches

   def MatchHWPlatform(self, biosVendor='', biosModel=''):
      """
      Matches the list of <hwPlatform>'s against the vendor and model
      strings passed in, returning a set of problems if there is no match.
      If there are no hwPlatforms, no match is performed, set() is returned.
       * biosVendor - Vendor Name string from BIOS or '' to not match
       * biosModel  - Model string from BIOS
      """
      problems = set()
      if biosVendor:
         for vendor, model in self.hwplatforms:
            if vendor == biosVendor:
               #
               # An empty model string means "match any model"
               # An empty biosModel means do not match
               if not model or not biosModel:
                  return set()
               elif model == biosModel:
                  return set()
               else:
                  problems.add("This Vib supports hardware model '%s' "
                            "but BIOS reports model as '%s'" %
                            (model, biosModel))
            else:
               problems.add("This Vib supports hardware vendor '%s' "
                            "but BIOS reports vendor as '%s'" %
                            (vendor, biosVendor))
      return problems

   def _reset(self):
      self.swplatforms = []
      self.hwplatforms = []
      self.maintenanceMode = True
      self.mModeInstall = None

   def AddPlatform(self, product, version, locale=""):
      self.swplatforms.append([product, version, locale]);

   def AddHWPlatform(self, vendor, model=""):
      if not vendor:
         raise ValueError('Nonempty vendor string must be passed')
      self.hwplatforms.append((vendor, model))

   def CalculateMMode(self, intendOp):
      if self.maintenanceMode and self.mModeInstall is not None:
         if intendOp == 'install':
            return self.mModeInstall
         elif intendOp == 'update':
            return True
         else:
            raise ValueError('Unsupported operation :%s' % (intendOp))
      return self.maintenanceMode

class Platforms(XmlElement):
   """
   Document describing the target platforms of a metadata set.  This will
   be used to facilitate indexing the metadata for VUM.
   """

   def __init__(self, root=None, platformFile=None):
      self.Reset()

      if platformFile:
         self._filename = platformFile
         fp = open(platformFile, 'r')

         try:
            try:
               tree = ElementTree(file=fp)
            except xmlerrclass, e:
               raise BadPlatformError(self._filename, str(e))
         finally:
            fp.close()

         self._fromXml(tree.getroot())

      elif root is not None:
         self._fromXml(root)

   def _reset(self):
      self._filename = ""
      self.platforms = []

   def AddPlatform(self, product, version, locale=''):
      if product not in INT_PRODUCTS:
         raise BadPlatformError('', "Invalid Product: %s." % product)

      match = re.search('^\d+\.(\*|\d+(\.(\*|\d+))?)$', version)
      if match is None and version != 'e.x.p':

         msg = "Invalid version: %s. Version must be X.Y.Z where X, "   \
            "Y, and Z are numbers and X.Y is required.  The last "   \
            "specified field (Y or Z) may be '*' to indicate match " \
            "all versions." % version

         raise BadPlatformError('', msg)

      self.platforms.append((product, version, locale))

   def _toXml(self):
      root = Element('platforms')
      for p, v, l in self.platforms:
         AppendElement(root, 'softwarePlatform', None,
                       productLineID = p,
                       version = v,
                       locale = l)

      return root

   def _fromXml(self, root):
      for softwarePlatform in root.findall('softwarePlatform'):
         productLineID = softwarePlatform.get('productLineID', '').strip()
         version = softwarePlatform.get('version', '').strip()
         locale = softwarePlatform.get('locale', '').strip()

         item = (productLineID, version, locale)
         self.platforms.append(item)

   def CheckSchema(self):
      for p, v, l in self.platforms:
         if not p:
            raise BadPlatformError(self._filename, "productLineID attribute is"
                             " 'required in <softwarePlatform>")
         elif p not in INT_PRODUCTS:
            raise BadPlatformError(self._filename, "Invalid productLineID "
                  "attribute '%s' in <softwarePlatform>.  Product must be "
                  "one of %s" % (p, list(INT_PRODUCTS)))

         match = re.search('^\d+\.(\*|\d+(\.(\*|\d+))?)$', v)
         if not match and v != 'e.x.p':
            msg = "Invalid version for <softwarePlatform>: %s. Version " \
                  "must be X.Y.Z where X, Y, and Z are numbers and X.Y " \
                  "is required.  The last specified field (Y or Z) may " \
                  "be '*' to indicate match all versions." % v

            raise BadPlatformError(self._filename, msg )


#
# Run from the command line to test Platform code
#

if __name__ == "__main__":
   import sys
   sys.path.append("../pythonroot")

if __name__ == "__main__":

   p = Platforms()
   p.AddPlatform('esx', '4.0', '')
   p.AddPlatform('embeddedEsx', '4.0', '')
   p.AddPlatform('esx', '4.5', 'EN-US')
   p.Write('platforms.xml')

   fp = open('platforms.xml')
   tree = ElementTree(file=fp)
   p2 = Platforms()
   p2.FromXml(tree.getroot())
   p2.Write('platforms2.xml')
