#######################################################################
# Copyright (C) 2008 VMWare, Inc.
# All Rights Reserved
########################################################################
#
#  iface.py
#    Iface, CommandLineIface, HostdIface
#

import datetime
import logging
import sys
import textwrap
import time

try:
   from lxml import etree
except ImportError:
   try:
      import xml.etree.cElementTree as etree
   except ImportError:
      import cElementTree as etree

import errors
import util

# Constants
# Version of EsxupdateSpec that this output corresponds to
SCAN_RESPONSE_VER = "1.30"

log = logging.getLogger('Iface')


class Iface(object):
   """ Iface is a base class for esxupdate user interface (ie output).
       All the work is done by derived classes for CLI and host agent
       --HA modes.
   """
   def __init__(self, outfile=sys.stdout):
      """ outfile - File instance for printing output
      """
      self.outfile = outfile

   def AddError(self, errorObj):
      """ Add an error to the output. Whether it is printed right away
          or at the end of a command run depends on implementation.
      """
      return NotImplemented

   def AddScanResponse(self, meta, **kwargs):
      """ Add a scan response to the output.
      * meta - a VIBScanner instance after running Scan()
      """
      return NotImplemented

   def AddQueryResponse(self, bulletins):
      """ Add a bulletin query to the output """
      return NotImplemented

   def AddVibQueryResponse(self, vibs):
      """ Add a Vib query response to the output """
      return NotImplemented

   def ShowBulletinDetails(self, bulletin):
      """ Display the fields in a bulletin and the list of
          constituent VIBs.
      """
      return NotImplemented

   def AddCheckRespose(self, checkresult):
      """ Add a check response to the output """
      return NotImplemented

   def Flush(self):
      """ For Ifaces that only output at the end of a command, this
          method prints out all of the output. See derived class
          for exact behavior.
      """
      pass

   def GetProgress(self):
      '''Returns an object that supports the following methods:
            * start()    - Called at the beginning of download with no
                           arguments.
            * set()      - Takes two arguments, the first is the amount
                           completed and the second is the total.
            * show()     - Takes no arguments, and causes the progress to be
                           shown on screen.
            * setTopic() - Takes one argument, the name of the file being
                           retrieved.
            * stop()     - Takes no arguments, called at finish.
         The default implementation supports these methods, but does nothing.
         The smart.progress.Progress class and its derivatives support this
         API.
      '''
      class SilentProgress(object):
         def start(self): pass
         def set(self, current, total): pass
         def show(self): pass
         def setTopic(self, topic): pass
         def stop(self): pass

      return SilentProgress()

   def DependencyPrompt(self, deps):
      '''Prompt the user to accept additional dependencies.
         Arguments:
            * deps - A VIBCollection instance of the additional VIBs required.
         Returns: True or False
      '''
      return False

   def ShowMessage(self, msg):
      '''A generic method for showing a message to the user.
         Arguments:
            * message - The message to display.
         Returns:  None
      '''
      pass

   def ShowException(self, e):
      '''A method to format an exception for consumption by whomever might be
         running esxupdate.  'e' may be an instance of
         errors.EsxupdateException, in which case the errno, description and
         attrs attributes will be available.  Otherwise, an unexpected error
         is assumed, and the only assumption is that 'e' is an instance of
         the built-in Exception class.'''
      pass

   def _bulletincmp(self, a, b):
      if a.installdate and b.installdate:
         return cmp(a.installdate, b.installdate)
      elif a.installdate:
         return -1
      elif b.installdate:
         return 1
      else:
         return cmp(a.id, b.id)

class CommandLineIface(Iface):
   """ Outputs to stdout for CLI users """
   def __init__(self, *args):
      Iface.__init__(self, *args)
      self.width = 80

   def GetProgress(self):
      import smart.interface
      smartif = smart.interface.createInterface('text', None)
      return smartif.getProgress(None)

   def ShowMessage(self, msg, indent=0):
      formatted_text = textwrap.fill(msg, width = self.width,
                                     initial_indent = ' ' * indent,
                                     subsequent_indent = ' ' * indent)
      if not formatted_text.endswith('\n'):
         formatted_text += '\n'
      self.outfile.write(formatted_text)

   def ShowListItem(self, name, description, indent=0):
      secondindent = indent + len(name + ' - ')
      msg = '%s - %s' % (name, description)
      formatted_text = textwrap.fill(msg, width = self.width,
                                     initial_indent = ' ' * indent,
                                     subsequent_indent = ' ' * secondindent)
      if not formatted_text.endswith('\n'):
         formatted_text += '\n'
      self.outfile.write(formatted_text)

   def Prompt(self, msg, options):
      def match_option(x):
         x = x.strip().lower()
         matches = [o for o in options if o.lower().startswith(x)]
         if len(matches) == 1:
            return matches[0]
         return None
      message = '%s %s:' % (msg, '/'.join(options))
      self.ShowMessage(message)
      answer = raw_input()
      option = match_option(answer)
      while not option:
         self.ShowMessage("'%s' is not a valid option." % answer)
         self.ShowMessage(message)
         answer = raw_input()
         option = match_option(answer)
      return option

   def DependencyPrompt(self, deps):
      self.ShowMessage('Additional VIBs required to satisfy dependencies.')
      self.ShowMessage('The following additional VIBs will be added in '
                       'order to satisfy dependencies:')
      for dep in deps.values():
         self.ShowMessage(dep.id, indent = 3)

      return self.Prompt('Do you want to continue?', ('Yes', 'No')) == 'Yes'

   def ShowException(self, e):
      if isinstance(e, errors.DependencyError):
         self.ShowMessage('The following problems were encountered trying to '
                          'resolve dependencies:')
         for problem in e.problemset:
            self.ShowMessage(str(problem), 3)
      elif isinstance(e, errors.UnsatisfiedDependencies):
         self.ShowMessage(e.message)
      elif isinstance(e, errors.NormalExit):
         if e.description:
            self.ShowMessage(e.description)
         if 'message' in e.attrs:
            self.ShowMessage(e.message)
      elif isinstance(e, errors.EsxupdateError):
         self.ShowMessage('Encountered error %s:' % e.__class__.__name__)
         self.ShowMessage('The error data is:')
         attrs = tuple(e.attrs) + ('errno', 'description')
         maxlen = max([len(x) for x in attrs])
         for attr in attrs:
            self.ShowListItem(attr.title().ljust(maxlen), getattr(e, attr), 3)
      else:
         self.ShowMessage('Encountered error %s:' % e.__class__.__name__)
         self.ShowMessage('This is an unexpected error.  Please report it as '
                          'a bug.', 3)
         self.ShowListItem('Error Message', str(e), 3)

   def ShowTable(self, table, header=False, indent=' '*2, wrap=False):
      """ Prints out a text table, automatically adjusting the width of
          each column to be the longest string in that column.  If the entire
          line exceeds self.width, the line is wrapped at column boundaries.
       * table - a list of rows, where each row is a list of items in each
                 column. All the rows must have a uniform number of
                 columns. Strings are left justified.
       * header - if True, center header row strings
       * indent - string to prefix wrapped lines
      """
      columns = range(len(table[0]))
      widths = [max(len(row[col]) for row in table) for col in columns]
      for r, row in enumerate(table):
         pos = 0
         for col in columns:
            if header and r == 0:
               item = row[col].center(widths[col], '-')
            else:
               item = row[col].ljust(widths[col])

            if wrap and self.width - pos <= len(item):
               self.outfile.write('\n' + indent)
               pos = len(indent)

            self.outfile.write(item + ' ')
            pos += len(item) + 1

         self.outfile.write('\n')

   def AddScanResponse(self, meta, vibview=False, all=False, scanmode='scan'):
      """List bulletins or VIBs in metadata. They are shown from oldest to
         newest, sorted by release date. By default only bulletins with updates
         are shown.
            Parameters:
               * meta    - A Metadata instance after running Scan().
               * vibview - List VIBs instead of bulletins.
               * all     - List all VIBs or bulletins.
               * scanmode- scan or check.
      """
      if vibview:
         values = meta.vibs.values()
         values.sort(lambda a, b: cmp(a.releasedate, b.releasedate))

         table = list()
         for v in values:
            state = ','.join(v.state)
            ts = util.floattoiso(v.releasedate)
            table.append((v.id, state, ts, str(v.meetssystemreqs)))

         table.insert(0, ("Vib ID",
                          "Pkgstate",
                          "Build Date",
                          "PlatformMatch",
                          ))
         self.ShowTable(table, header=True)
      else:
         if all:
            self.ShowMessage('All of the bulletins are listed, '
                             'including compliant ones and ones not '
                             'applicable to this platform.')
         else:
            self.ShowMessage('Applicable bulletins with updates are listed.')

         values = meta.bulletins.values()
         values.sort(self._bulletincmp)

         table = list()
         for bul in values:
            if all or bul.newer:
               # Date: print YYYY-MM-DD, skip time
               # Summary: only supposed to be 40 chars long
               table.append((bul.id, bul.releasedatestr, bul.summary[:40]))

         table.insert(0, ("Bulletin ID",
                          "Date",
                          "Summary",
                      ))
         self.ShowTable(table, header=True)

   def AddCheckRespose(self, checkresult):
      self.ShowMessage('Esxupdate local cache states:')
      self.ShowMessage('Location: %s' % (checkresult['location']), indent = 1)
      self.ShowMessage('Available space: %s [MB]' % (str(checkresult['size'])),
            indent = 1)

   def AddQueryResponse(self, bulletins):
      """ Add a bulletin query to the output """
      values = bulletins.values()
      values.sort(self._bulletincmp)
      table = list()
      for bulletin in values:
         installdate = util.floattoiso(bulletin.installdate or 0)[:19]
         table.append((bulletin.id, installdate, bulletin.summary))
      table.insert(0, ("Bulletin ID", "Installed", "Summary"))
      self.ShowTable(table, header=True)

   def AddVibQueryResponse(self, vibs):
      """ Add a Vib query response to the output """
      values = vibs.values()
      values.sort(lambda a, b: cmp(a.installdate, b.installdate))

      table = list()
      for v in values:
         state = ','.join(v.state)
         if v.installdate is not None:
            ts = util.floattoiso(v.installdate)
         else:
            ts = ' - '
         table.append((v.id, state, ts))

      table.insert(0, ("VIB ID", "Package State", "Timestamp"))

      self.ShowTable(table, header=True)

   def ShowBulletinDetails(self, bulletin):
      """ Display the fields in a bulletin and the list of
          constituent VIBs.
      """
      if bulletin.installdate:
         installdate = util.floattoiso(bulletin.installdate)
      else:
         installdate = ''
      self.ShowListItem('ID          ', bulletin.id, 3)
      self.ShowListItem('Release Date', bulletin.releasedatestr, 3)
      self.ShowListItem('Vendor      ', bulletin.vendor, 3)
      self.ShowListItem('Summary     ', bulletin.summary, 3)
      self.ShowListItem('Severity    ', bulletin.severity, 3)
      self.ShowListItem('Urgency     ', bulletin.urgency, 3)
      self.ShowListItem('Category    ', bulletin.category, 3)
      self.ShowListItem('Install Date', installdate, 3)
      self.ShowListItem('Description ', bulletin.description, 3)
      self.ShowListItem('KB URL      ', bulletin.kburl, 3)
      self.ShowListItem('Contact     ', bulletin.contact, 3)
      self.ShowListItem('Compliant   ', str(bulletin.compliant), 3)
      if not bulletin.compliant:
         # show system requirements
         vib_attrs = ('RebootRequired ',
                      'HostdRestart   ',
                      'MaintenanceMode')
         flags = {}
         for vib_attr in vib_attrs:
            flags[vib_attr] = False
         for vib in bulletin.values():
            for vib_attr in vib_attrs:
               flags[vib_attr] |= getattr(vib, vib_attr.lower().strip())
         for vib_attr in vib_attrs:
            self.ShowListItem(vib_attr, flags[vib_attr], 3)

      self.ShowMessage('List of constituent VIBs:', 3)
      vibids = bulletin.vibids | set(bulletin.keys())
      for vibid in vibids:
         self.ShowMessage(vibid, 6)
      self.ShowMessage('')     # blank line

class HostdProgress(object):
   TOPIC_LENGTH = 40
   INTERVAL = 0.5    # progress update frequency, in second
   DELTA = 25        # % of a single task to increase overall progress

   def __init__(self):
      import config
      self.topic = ""
      self._overall = 0  # % of overall tasks
      self._value = 0    # % of current task
      self._threshold = 95
      self._laststage = 0
      self._lastupdate = time.time()
      self._showprogress = True
      if config.GetConf().platform.version == '4.0.0':
         self._showprogress = False

   def start(self):
      self._lastupdate = 0

   def set(self, current, total):
      if total <= 0:
         self._value = 100
      else:
         self._value = int(current * 100 / total)
      self._updateprogress(self._laststage, self._value)

   def show(self):
      '''Update progress through stderr'''
      if self._showprogress:
         now = time.time()
         if now - self._lastupdate > self.INTERVAL:
            msg = '<progress topic="%s" value="%d" overall="%d" />\n' % (
                  (self.topic[:self.TOPIC_LENGTH], self._value, self._overall))
            sys.stderr.write(msg)
            self._lastupdate = now

   def setTopic(self, topic):
      self.topic = topic

   def stop(self):
      self.set(1, 1)
      self._laststage = 0
      self._value = 0

   def _updateprogress(self, oldvalue, newvalue):
      if newvalue - oldvalue >= self.DELTA:
         t = (newvalue - oldvalue) / self.DELTA
         self._laststage += t * self.DELTA
         self._overall = min(self._threshold, self._overall + t)


class HostdIface(Iface):
   """ Formats XML output for --HA mode for consumption by all users
       of the --HA / hostd PatchManager interface.
   """
   def __init__(self, *args):
      import config
      Iface.__init__(self, *args)
      self._progress = None
      prologue = '<esxupdate-response>\n<version>%s</version>\n' % (
         SCAN_RESPONSE_VER)
      self.outfile.write(prologue)

      if config.GetConf().platform.version != '4.0.0':
         msg = '<progress topic="esxupdate" value="0" overall="0" />\n'
         sys.stderr.write(msg)

   def _writeXmlNode(self, node):
      util.xmlUtils.IndentElementTree(node)
      self.outfile.write(etree.tostring(node))

   @staticmethod
   def _requiresNode(requires):
      """ Returns ElementTree node for each Requires instance passed in """
      node = etree.Element('requires')
      util.xmlUtils.AppendElement(node, 'requirement', requires.name)
      if requires.providedby:
         prov = etree.Element('providedBy')
         for provvib in requires.providedby:
            util.xmlUtils.AppendElement(prov, 'vibID', str(provvib))
         node.append(prov)

      util.xmlUtils.AppendBoolElement(node, 'providedByHost',
                                      requires.providedbyhost)
      HostdIface._appendPackageList(node, requires.hostprovides, 'hostProvides')
      return node

   @staticmethod
   def _appendPackageList(root, pkgset, listtag):
      """
      Appends an Element list of <package name='' version=''> elements
      if pkgset is a non-empty set of (name, version) tuples
      """
      if pkgset:
         node = etree.Element(listtag)
         for n, v in pkgset:
            node.append(etree.Element('package', name=n, version=v))
         root.append(node)

   def AddScanResponse(self, meta, **kwargs):
      """Adds XML scan response to output.
            Parameters:
               * meta - a VIBScanner instance after running Scan(). Note that
                        only VIBs with scanned=True will be reported.
      """
      skipped = 0
      for vib in meta.vibs.values():
         if vib.scanned:
            vibnode = etree.Element('vib-scan-data')
            util.xmlUtils.AppendElement(vibnode, 'id', vib.id)
            # scan-results currently only provides for one state.
            if len(vib.state) == 1:
               state = iter(vib.state).next()
            elif vib.state.INSTALLED in vib.state:
               if vib.state.PENDING in vib.state:
                  state = vib.state.PENDING
               else:
                  state = vib.state.INSTALLED
            elif vib.state.RETIRED in vib.state:
               state = vib.state.RETIRED
            else:
               if vib.state.STAGED in vib.state:
                  state = vib.state.STAGED
               else:
                  state = vib.state.UNINSTALLED
            util.xmlUtils.AppendElement(vibnode, 'pkgstate', state)
            util.xmlUtils.AppendBoolElement(vibnode, 'meetsSystemReq',
                                            vib.meetssystemreqs)
            for failure in vib.systemreqfailures:
               util.xmlUtils.AppendElement(vibnode, 'systemReqFailure', failure)
            util.xmlUtils.AppendBoolElement(vibnode, 'pkgDepsMetByHost',
                                            vib.pkgdepsmetbyhost)

            for req in vib.requires:
               vibnode.append(self._requiresNode(req))

            confnode = etree.Element('conflicts')
            for confvib in vib.conflicts:
               util.xmlUtils.AppendElement(confnode, 'vibID', confvib)
            util.xmlUtils.AppendBoolElement(confnode, 'conflictsWithHost',
                                            vib.conflictswithhost)
            self._appendPackageList(confnode, vib.hostconflicts,
                                    'hostConflicts')
            vibnode.append(confnode)

            obsnode = etree.Element('obsoletes')
            for oldervib in vib.oldervibs:
               util.xmlUtils.AppendElement(obsnode, 'vibID', oldervib)
            util.xmlUtils.AppendBoolElement(obsnode, 'obsoletesHost',
                                            vib.obsoleteshost)
            self._appendPackageList(obsnode, vib.hostobsoleted,
                                    'hostObsoleted')
            util.xmlUtils.AppendBoolElement(obsnode, 'obsoletedByHost',
                                            vib.obsoletedbyhost)
            vibnode.append(obsnode)

            # Repeat systemReq and postInstall from vmware.xml <vib> 
            # definition, per agreement with VUM team schema.
            if vib._meta:
               if kwargs.get('scanmode', 'scan') == 'check':
                  # Didn't factor this in to current (new) VIB class design.
                  # Should fix this later.
                  vibnode.append(vib._meta.systemReqs.ToXml())
               else:
                  mmodeOverride = vib.requiresmaintenancemode
                  vibnode.append(vib._meta.systemReqs.ToXml(
                     mmodeOverride=mmodeOverride))

            postInst = etree.Element('postInstall')
            util.xmlUtils.AppendBoolElement(postInst, 'rebootRequired',
                                            vib.rebootrequired)
            util.xmlUtils.AppendBoolElement(postInst, 'hostdRestart',
                                            vib.hostdrestart)
            vibnode.append(postInst)

            self._writeXmlNode(vibnode)
         else:
            skipped += 1

      self._addBulletins(meta.bulletins)

      log.debug('Skipped %d unscanned VIBs in scan output' % (skipped))

   def _addBulletins(self, bulletins):
      values = bulletins.values()
      values.sort(self._bulletincmp)
      for b in values:
         bulnode = etree.Element('bulletin')
         util.xmlUtils.AppendElement(bulnode, 'id', b.id)
         util.xmlUtils.AppendElement(bulnode, 'summary', b.summary)
         if b.releasedate:
            releasedate = util.floattoiso(b.releasedate)
         else:
            releasedate = b.releasedatestr
         util.xmlUtils.AppendElement(bulnode, 'releaseDate', releasedate)
         installdate = util.floattoiso(b.installdate or 0)
         util.xmlUtils.AppendElement(bulnode, 'installDate', installdate)
         util.xmlUtils.AppendBoolElement(bulnode, 'matchesPlatform',
                                         b.meetssystemreqs)
         util.xmlUtils.AppendBoolElement(bulnode, 'newerVibs', b.newer)
         self._writeXmlNode(bulnode)

   def AddQueryResponse(self, bulletins):
      """ Add a bulletin query to the output """
      self._addBulletins(bulletins)

   def ShowBulletinDetails(self, bulletin):
      """ Display the fields in a bulletin and the list of
          constituent VIBs.
      """
      return NotImplemented

   def ShowException(self, e):
      error = etree.Element('error', errorClass=e.__class__.__name__)
      if isinstance(e, errors.EsxupdateError):
         util.xmlUtils.AppendElement(error, 'errorCode', str(e.errno))
         util.xmlUtils.AppendElement(error, 'errorDesc', e.description)

         # 'message' has important error details
         if 'message' in e.attrs:
            util.xmlUtils.AppendElement(error, 'errorDesc', e.message)

         if isinstance(e, errors.DependencyError):
            elem = etree.Element('problemset')
            error.append(elem)
            for problem in e.problemset:
               util.xmlUtils.AppendElement(elem, 'problem', str(problem))
         elif isinstance(e, errors.UnsatisfiedDependencies):
            elem = etree.Element('suggestedvibs')
            error.append(elem)
            for vib in e.suggestedvibs.values():
               util.xmlUtils.AppendElement(elem, 'vibID', vib.id)
         else:
            for attr in e.attrs:
               util.xmlUtils.AppendElement(error, attr, str(getattr(e, attr)))
      else:
         #
         # Unknown error, use 99
         util.xmlUtils.AppendElement(error, 'errorCode', '99')
         util.xmlUtils.AppendElement(error, 'errorDesc', str(e))

      self._writeXmlNode(error)

   def AddCheckRespose(self, checkresult):
      cacheNode = etree.Element('cache')
      util.xmlUtils.AppendElement(cacheNode, 'location',
                                  checkresult['location'])
      util.xmlUtils.AppendElement(cacheNode, 'size', str(checkresult['size']))
      self._writeXmlNode(cacheNode)

   def Flush(self):
      import config
      if config.GetConf().platform.version != '4.0.0':
         msg = '<progress topic="esxupdate" value="100" overall="100" />\n'
         sys.stderr.write(msg)
      self.outfile.write('</esxupdate-response>\n')


   def GetProgress(self):
      if self._progress is None:
         self._progress  =  HostdProgress()
      self._progress.stop()
      return self._progress
