#######################################################################
# Copyright (C) 2009 VMWare, Inc.
# All Rights Reserved
########################################################################
#
# Metadata.py
#

__all__ = ['Metadata']

import logging
import os

try:
   from lxml import etree
except ImportError:
   try:
      import xml.etree.cElementTree as etree
   except ImportError:
      import cElementTree as etree

import Bulletin
import Vib
import VIBCollection
import VIBScanner
import config
import downloader
import errors
import scan

log = logging.getLogger(__name__)

class Metadata(VIBScanner.VIBScanner):
   """This class provides methods for scanning against zipped meta-data
      (metadata.zip). It assumes that each metadata.zip can contain any
      combination of VIB types; however, 'rpm' VIB types are supported by
      meta-data in repomod format, 'deb' VIB types will have corresponding
      deb-dir meta-data, and 'cross' VIB types will have both.

      Attributes:
         * bulletins - An instance of BulletinCollection, containing the
                       bulletins parsed from meta-data.
         * vibs      - An instance of VIBCollection, containing the VIBs parsed
                       from meta-data (both those contained within bulletin
                       definitions, and any not enclosed in a bulletin).
   """

   def __init__(self, *args, **kwargs):
      """Class constructor.
            Parameters:
               * args   - One or more paths or URLs to zipped meta-data.
               * kwargs - Only one keyword argument is supported: usepkgdb. If
                          False, data from the local host's package database
                          will not be considered when calculating dependencies.
                          The default is True.
            Returns: A new Metadata instance.
      """
      usepkgdb = kwargs.pop('usepkgdb', True)
      if kwargs:
         msg = "Unrecognized keyword argument(s): '%s'" % ', '.join(kwargs)
         raise TypeError(msg)

      # The keys of this dictionary are the URLs, the values are the local
      # cache location for each URL. If a key has a None value, that means we
      # haven't loaded data from that URL yet.
      self._metalocations = dict((url, None) for url in args)
      VIBScanner.VIBScanner.__init__(self, usepkgdb=usepkgdb)

      # I want this to go away some day. -RS
      self._cachedir = config.GetConf().default.cachedir

   def _loadmeta(self, metaurl, progress):
      try:
         d = downloader.CachedZipDownloader(metaurl,
                                            cachedir=self._cachedir,
                                            prefix='metadata',
                                            destzipname='metadata.zip',
                                            reget = True,
                                            progress=progress)
         d.GetAndUnpack()
      except downloader.DownloaderError, e:
         raise errors.MetadataDownloadError(e.url, e.local, e.err)
      except downloader.UnzipError, e:
         raise errors.MetadataFormatError(e.zipfile, e.msg)

      realurl = d.url
      if realurl != metaurl:
         log.debug('Meta-data URL is redirected. Will construct relative '
                   'URLs using new URL: %s' % realurl)
         self._metalocations[realurl] = None
         del self._metalocations[metaurl]
      localdir = d.destdir

      vmwarexml = os.path.join(localdir, 'vmware.xml')
      if not os.path.exists(vmwarexml):
         raise errors.MetadataFormatError(localdir, 'vmware.xml missing')
      try:
         root = etree.parse(vmwarexml).getroot()
      except EnvironmentError, e:
         msg = 'Unable to load %s: %s' %(vmwarexml, e)
         raise errors.FileIOError(vmwarexml, msg)
      except Exception, e:
         msg = 'Unable to parse %s: %s' % (vmwarexml, e)
         raise errors.MetadataFormatError(vmwarexml, msg)

      for bulnode in root.findall('bulletin'):
         try:
            bul = Bulletin.Bulletin.fromxml(bulnode)
         except Exception, e:
            log.warn('Error parsing bulletin from %s: %s' % (vmwarexml, e))
            continue

         if bul in self.bulletins:
            self.bulletins[bul.id] += bul
            bul = self.bulletins[bul.id]
         else:
            self.bulletins[bul.id] = bul

         for vib in bul.values():
            iszip = metaurl.startswith('zip:')
            if vib.sourceurl and not iszip:
               vib.remotelocation = vib.sourceurl
            else:
               # If remotelocation is a zip: URL, we use '?' to separate the ZIP
               # file name from the archive member name. If remotelocation is a
               # URL, relative paths are calculated based on the last '/' in the
               # URL.
               prefix = realurl[:realurl.rfind(iszip and '?' or '/') + 1]
               if vib.relativepath:
                  vib.remotelocation = prefix + vib.relativepath
               else:
                  vib.remotelocation = prefix + vib.id + '.vib'

            if vib.id in self.vibs:
               new = self.vibs[vib.id] + vib
               self.vibs[new.id] = new
               for bulletin in self.bulletins.values():
                  if new.id in bulletin:
                     bulletin[new.id] = new
            else:
               self.vibs[vib.id] = vib

      self._metalocations[realurl] = localdir

      return localdir

   def load(self, iface=None):
      """Populates this instance's vibs and bulletins attributes with
         information from the meta-data URLs. Also sets the VIB's
         remotelocation attribute, since this may be relative to the
         meta-data's location.
            Parameters:
               * iface - An instance of the Iface.Iface class. Used primarily
                         for showing a nice progress bar duing downloading.
            Raises:
               * MetadataDownloadError - If one or more of the URLs specified
                                         in the constructor cannot be
                                         retrieved.
               * MetadataFormatError   - If one or more of the URLs specified
                                         in the constructor cannot be
                                         extracted or parsed.
      """
      progress = None
      if iface:
         progress = iface.GetProgress()

      for metaurl, cachelocation in self._metalocations.items():
         if cachelocation and os.path.isdir(cachelocation):
            continue

         localdir = self._loadmeta(metaurl, progress)

         self.pkgscanner += self._createscanner(localdir)

   def _createscanner(self, metalocation):
      'Creates a PkgScanner instance and returns it.'
      name = os.path.basename(metalocation)
      if config.GetPlatform() == 'cos':
         metatype = 'rpm-md'
      else:
         metatype = 'apt-deb'
      return scan.PkgScanner(name=name,
                             metatype=metatype,
                             metabase=metalocation,
                             usesyschannel=self.usepkgdb)
