#######################################################################
# Copyright (C) 2009 VMWare, Inc.
# All Rights Reserved
########################################################################
#
# PackageDatabase.py
#

"""This module implements classes for populating VIB information from a host's
   package database.
"""

__all__ = ['PackageDatabase', 'RPMDatabase', 'ipkgDatabase']

import logging
import os

import Vib
import VIBScanner
import config
import errors
import platform
import util.runcommand
import version

try:
   import rpm
   HAVE_RPM = True
except:
   HAVE_RPM = False

log = logging.getLogger(__name__)

class RPMDatabase(VIBScanner.VIBScanner):
   """This class provides methods for creating VIB data based on the host's
      RPM database.

      Attributes:
         * bulletins - An instance of BulletinCollection. (Not used.)
         * vibs      - An instance of VIBCollection, containing the VIBs parsed
                       from the package database.
   """

   def load(self):
      """Populates this instance's vibs and bulletins attributes with
         information from the host database.
            Raises:
               * PackageManagerError - If the package manager database cannot
                                       be read.
      """
      if not HAVE_RPM:
         raise RuntimeError("Required rpm module missing.")

      try:
         ts = rpm.ts()
         for hdr in ts.dbMatch():
            ver = version.RPMVersion(hdr[rpm.RPMTAG_EPOCH],
                                     hdr[rpm.RPMTAG_VERSION],
                                     hdr[rpm.RPMTAG_RELEASE],
                                     hdr[rpm.RPMTAG_ARCH])
            vid = Vib.VIB.idfromtnv(Vib.VIB.TYPE_RPM, hdr[rpm.RPMTAG_NAME], ver)
            v = Vib.VIB(vid, Vib.VIB.TYPE_RPM, persist=False)
            v.state.add(v.state.INSTALLED)
            v.name = hdr[rpm.RPMTAG_NAME]
            v.version = ver
            v.description = hdr[rpm.RPMTAG_DESCRIPTION]
            v.summary = hdr[rpm.RPMTAG_SUMMARY]
            v.installdate = float(hdr[rpm.RPMTAG_INSTALLTIME])
            v.releasedate = float(hdr[rpm.RPMTAG_BUILDTIME])
            v.vendor = hdr[rpm.RPMTAG_VENDOR]
            self.vibs[vid] = v
      # Any other type of error is a programming error.
      except (EnvironmentError, rpm.error), e:
         raise errors.PackageManagerError(e)


class RPMDatabaseCLI(VIBScanner.VIBScanner):
   """This class provides methods for creating VIB data based on the host's
      RPM database. Unlike the RPMDatabase class, this one obtains data using
      the rpm command-line utility, as opposed to using the rpm module for
      Python.

      Attributes:
         * bulletins - An instance of BulletinCollection. (Not used.)
         * vibs      - An instance of VIBCollection, containing the VIBs parsed
                       from the package database.
   """
   def _rpmquery(self):
      rpmfields = ('epoch', 'version', 'release', 'arch', 'name', 'description',
                   'summary', 'installtime', 'buildtime', 'vendor')
      qf = '\xff'.join(tuple('%%{%s}' % x for x in rpmfields)) + '\xfe'
      args = ['/bin/rpm', '-qa', '--qf', qf]
      try:
         # This is already halfway to the 1MB max output that runcommand will
         # store in memory. Should find a more efficient way to do this.
         rc, out = util.runcommand.runcommand(args)
         if rc:
            msg = ("rpm command exited with non-zero status (%d): '%s'."
                   % (rc, out))
            raise errors.PackageManagerError(msg)
         for record in out.split('\xfe'):
            if record:
               yield dict(zip(rpmfields, record.split('\xff')))
      except Exception, e:
         msg = "Error retrieving data from rpm command-line: %s." % e
         raise errors.PackageManagerError(msg)

   def load(self):
      """Populates this instance's vibs and bulletins attributes with
         information from the host database.
            Raises:
               * PackageManagerError - If the package manager database cannot
                                       be read.
      """
      for values in self._rpmquery():
         try:
            epoch = values.get('epoch')
            if epoch == '(none)':
               epoch = None
            ver = version.RPMVersion(epoch, values['version'],
                                     values['release'], values['arch'])
            vid = Vib.VIB.idfromtnv(Vib.VIB.TYPE_RPM, values['name'], ver)
            v = Vib.VIB(vid, Vib.VIB.TYPE_RPM, persist=False)
            v.state.add(v.state.INSTALLED)
            v.name = values['name']
            v.version = ver
            v.description = values['description']
            v.summary = values['summary']
            try:
               v.installdate = float(values['installtime'])
            except:
               pass
            try:
               v.releasedate = float(values['buildtime'])
            except:
               pass
            v.vendor = values['vendor']
            self.vibs[vid] = v
         except KeyError, e:
            log.warn("Could not parse RPM value for '%s' in '%s' package" %
                     (e, values.get('name', 'unknown')))

class ipkgDatabase(VIBScanner.VIBScanner):
   """This class provides methods for creating VIB data based on the host's
      ipkg database.

      Attributes:
         * bulletins - An instance of BulletinCollection. (Not used.)
         * vibs      - An instance of VIBCollection, containing the VIBs parsed
                       from the package database.
   """

   def _parsestatus(self, path):
      """Generator function to yield values from an ipkg status file."""
      try:
         f = open(path)
      except Exception, e:
         msg = "Cannot open ipkg status file: %s." % e
         raise errors.PackageManagerError(msg)

      try:
         values = dict()
         name = None
         for line in f:
            line = line.strip('\n')
            if line:
               if line.startswith(' ') and name in ('Description', 'Conffiles'):
                  values[name] = '\n'.join((values[name], line))
               else:
                  try:
                     name, value = (x.strip() for x in line.split(':'))
                     values[name] = value
                  except:
                     msg = ("ipkg status: skipping invalid line ('%s')." %
                            line)
                     log.warn(msg)
            elif values:
               yield values
               values = dict()
      except Exception, e:
         f.close()
         msg = "Error reading ipkg status file: %s" % e
         raise errors.PackageManagerError(msg)

      f.close()

   def load(self):
      """Load VIB data from the ipkg status file(s).
            Note: This method reads ipkg databases directly, without checking or
                  obtaining a lock.  (This isn't meant to infer infer that ipkg
                  itself uses locking, either, because it doesn't).  It assumes
                  that the ipkg database has been staged, and that it will not
                  change while this method is executing.
            Raises:
               * PackageManagerError - If the package manager database cannot
                                       be read.
      """

      for statusfn in platform.conf['latestDBs']:
         if not os.path.exists(statusfn):
            log.warn("ipkg status file '%s' does not exist." % statusfn)
            continue

         for values in self._parsestatus(statusfn):
            name = values.get('Package')
            ver = version.DebVersion(values.get('Version', ''))

            if not name or not ver.versionstring:
               log.warn("Invalid package data parsed from '%s'." % statusfn)
               continue

            vid = Vib.VIB.idfromtnv(Vib.VIB.TYPE_DEB, name, ver)

            v = Vib.VIB(vid, Vib.VIB.TYPE_DEB, persist=False)
            v.name = name
            v.version = ver

            try:
               # This is here to be consistent with smart, which assumes
               # anything that isn't strictly 'installed' is not installed.
               # (ipkg defines various intermediate and failure states for
               # installation and removal).
               if values['Status'].split()[2] == 'installed':
                  v.state.add(v.state.INSTALLED)
            except:
               pass

            try:
               v.installdate = float(values['Installed-Time'])
            except Exception, e:
               pass

            self.vibs[vid] = v

if config.GetPlatform() == 'cos':
   # We can use the RPMDatabase class when we resolve the issues with accessing
   # the RPM database from both 32- and 64-bit processes. See PR 457206.
   #PackageDatabase = RPMDatabase
   PackageDatabase = RPMDatabaseCLI
else:
   PackageDatabase = ipkgDatabase
