#######################################################################
# Copyright (C) 2009 VMWare, Inc.
# All Rights Reserved
########################################################################
#
# Transaction.py
#

"""This module provides the Transaction class, a high-level interface to
   esxupdate functions.
"""

__all__ = ['Transaction']

import logging
import os

import Database
import Iface
import Metadata
import PackageDatabase
import VIBCollection
import VIBScanner
import Vib
import config
import const
import depsolver
import errors
import platform
import scan
import util.runcommand

log = logging.getLogger(__name__)

platform.PlatformInit()
## TODO:
# This is just a workaround for PR 462281.
# We need a proper way of setting the HWPLATFORM at initialization
# time, and also should move the static setting out of the Vib
# module into a centralized configuration engine.  For now that is
# too much work.
biosvendor, biosmodel = platform.GetBiosVendorModel()
log.info('BIOS reports Vendor %s, Model %s' % (biosvendor, biosmodel))
Vib.HWPLATFORM = Vib.HWPlatform(biosvendor, biosmodel)
Vib.SWPLATFORM = Vib.SWPlatform(config.GetConf().platform.product,
                                config.GetConf().platform.version)


class Transaction(object):
   """The Transaction class provides a high-level interface to esxupdate
      functions, including enumerating VIB and Bulletin objects from the
      esxupdate database and from meta-data, as well as updating or removing
      VIBs on the host.
         Class Variables:
            * SUCCESS                        - Indicates a successful
                                               installation of one or more VIBs.
            * SUCCESS_REQUIRES_REBOOT        - Indicates one or more VIBs were
                                               successfully installed, but that
                                               the host must be rebooted before
                                               the changes are effective.
            * SUCCESS_REQUIRES_HOSTD_RESTART - Indicates one or more VIBs were
                                               successfully installed, but that
                                               the host agent daemon must be
                                               restarted before the changes are
                                               effective.
            * SUCCESS_HOST_NOT_CHANGED       - Indicates that a transaction
                                               succeeded, but no VIBs were
                                               installed or removed.
         Attributes:
            * metadata   - An instance of Metadata class containing the combined
                           metadatas passed into the constructor, plus VIBs for
                           installed packages that are not in the metadatas.
            * database   - An instance of the Database class initialized from
                           the local esxupdate database.
            * pkgdb      - An instance of PackageDatabase with VIBs generated
                           for installed packages.  Only VIBs which are not in
                           metadata are kept however.
   """

   SUCCESS                        = 0
   SUCCESS_REQUIRES_REBOOT        = 80
   SUCCESS_REQUIRES_HOSTD_RESTART = 81
   SUCCESS_HOST_NOT_CHANGED       = 82

   VIBCONF = '/etc/vmware/esxupdate/vib.conf'

   def __init__(self, *args, **kwargs):
      """Class constructor.
            Parameters:
               * args   - Positional (non-keyword) arguments giving URL or
                          local paths to meta-data. If there are no positional
                          arguments, VIB and Bulletin information will only be
                          loaded from the esxupdate database.
               * kwargs - One or more of the following optional keyword
                          arguments:
                             * dbpath    - As specified for the Database class
                                           constructor. Defaults to the value
                                           returned by platform.GetDbPath().
                             * dbcreate  - As specified for the Database class
                                           constructor.
                             * usepkgdb  - As specified for the Metadata class
                                           constructor.
                             * iface     - An instance of the Iface.Iface class.
                                           This parameter is implemented
                                           primarily for backwards compatibility
                                           with the esxupdate command-line
                                           utility. Its use is otherwise
                                           discouraged, and there is no warranty
                                           that this parameter will be provided
                                           in future revisions or that the
                                           interface will remain compatible.
                             * loadpkgdb - A boolean specifying whether to
                                           create VIBs from packages loaded from
                                           the host's package database. If this
                                           is True, VIB objects will be created
                                           for each package in the host's
                                           package database. Defaults to False.
            Returns: A new Transaction object.
            Raises:
               * DatabaseIOError       - If database lock cannot be obtained, or
                                         a file cannot be read.
               * DatabaseFormatError   - If vibs.xml or bulletins.zip cannot be
                                         parsed, or format is otherwise invalid.
               * MetadataDownloadError - If one or more of the URLs specified
                                         in the constructor cannot be
                                         retrieved.
               * MetadataFormatError   - If one or more of the URLs specified
                                         in the constructor cannot be
                                         extracted or parsed.
      """
      databasekwargs = {'dbpath': platform.GetDbPath()}
      for key in ('dbpath', 'dbcreate'):
         if key in kwargs:
            databasekwargs[key] = kwargs.pop(key)

      metadatakwargs = dict()
      for key in ('usepkgdb',):
         if key in kwargs:
            metadatakwargs[key] = kwargs.pop(key)

      self.iface = kwargs.pop('iface', Iface.Iface())

      self.loadpkgdb = kwargs.pop('loadpkgdb', False)

      if kwargs:
         msg = "Unrecognized keyword argument(s): '%s'" % ', '.join(kwargs)
         raise TypeError(msg)

      self._updates = VIBScanner.VIBScanner()
      self._removes = VIBScanner.VIBScanner()

      self._skipstage = VIBCollection.VIBCollection()

      self.database = Database.Database(**databasekwargs)
      self.metadata = Metadata.Metadata(*args, **metadatakwargs)

      if self.loadpkgdb:
         self.pkgdb = PackageDatabase.PackageDatabase()
      else:
         self.pkgdb = VIBScanner.VIBScanner(**metadatakwargs)

      self.metadata.load(self.iface)
      self.load()

      #
      # pkgdb will end up with VIBs for host packages except for
      # those in metadata already.  If any pkgdb packages are also
      # in Database, the VIB from Database will replace the one
      # generated in pkgdb.  This ensures no duplicates due to
      # the same name and version but different VIB types.
      # Also, the original metadata before host packages were added
      # can be obtained by removing the pkgdb keys from metadata.
      self._trimpkgdb(self.metadata.vibs)
      self._trimpkgdb(self.database.vibs, replace=True)
      self.metadata += self.pkgdb

      log.debug('Final VIB counts: metadata %d  pkgdb %d' % (
         len(self.metadata.vibs), len(self.pkgdb.vibs)))

   def load(self):
      """Loads or reloads data from the database and any metadata."""
      dbupdated = self.database.load()

      self.database.sync(self.metadata)

      # Due to the performance penalty of calling pkgdb.load(), we only do so
      # if the database has changed (or if it's the first time we're loading
      # the database).
      if dbupdated and self.loadpkgdb:
         self.pkgdb.load()

      log.debug('VIBs loaded: database %d  metadata %d  pkgdb %d' % (
         len(self.database.vibs), len(self.metadata.vibs), len(self.pkgdb.vibs)))

   def _trimpkgdb(self, vibcollection, replace=False):
      # Only keep things in self.pkgdb if we don't have an entry for them
      # in vibcollection. If replace is True, an entry from vibcollection 
      # replaces the entry in pkgdb if the name and version are identical.
      # This is useful when the vibtypes are different, say Database has
      # cross_raj_1 and pkgdb has rpm_raj_1.
      for v in self.pkgdb.vibs.values():
         v2 = vibcollection.getbynameversion(v.name, v.version)
         if v2:
            del self.pkgdb.vibs[v.id]
            if replace:
               self.pkgdb.vibs[v2.id] = v2

   @property
   def bulletins(self):
      """A property returning a BulletinCollection instance of the union of
         bulletins from metadata and the database. References in the returned
         data may be objects from the database, from the metadata, or new
         objects created by merging two objects with the same ID.  (See the
         documentation for BulletinCollection.__add__.)
      """
      return self.database.bulletins + self.metadata.bulletins

   @property
   def vibs(self):
      """A property returning a VIBCollection instance of the union of VIBs
         from metadata and the database. References in the returned data may be
         objects from the database, from the metadata, or new objects created
         by merging two objects with the same ID.  (See the documentation for
         VIBCollection.__add__.)
      """
      return self.database.vibs + self.metadata.vibs + self.pkgdb.vibs

   def get(self, name, default=None):
      """Retrieve an object with the given ID. Collections are searched in this
         order: database.bulletins, database.vibs, metadata.bulletins,
         metadata.vibs, pkgdb.vibs. If no match is found, the default value is
         returned.
            Parameters:
               * name    - A string giving the ID of the desired object.
               * default - The value to return if the specified ID is not
                           found.  (Defaults to None.)
            Returns:
               * A VIB or Bulletin object, or the default value.
      """
      return (self.database.get(name) or self.metadata.get(name) or
              self.pkgdb.get(name) or default)

   def update(self, arg):
      """Schedule the selected VIB or Bulletin for updating.
            Parameters:
               * arg - The ID of a bulletin or VIB in metadata.
            Raises:
               * NoMatchError - If the given argument does not match a bulletin
                                or VIB ID in metadata.
      """
      bulletin = self.metadata.bulletins.get(arg)
      if bulletin:
         self._updates.bulletins[bulletin.id] = bulletin
      else:
         vib = self.metadata.vibs.get(arg)
         if vib:
            self._updates.vibs[vib.id] = vib
            # This would be a rare case, but better to avoid potentially
            # confusing behavior.
            self._skipstage.pop(vib.id, None)
         else:
            msg = "No Bulletin or VIB found with ID '%s'." % arg
            raise errors.NoMatchError(arg, msg)

   def updatevibfromurl(self, url, skipsigcheck = False, nocache = False):
      """Schedule the VIB at the selected URL for updating.
            Parameters:
               * url          - A string giving the URL or local path of a VIB
                                file.
               * skipsigcheck - Do not validate VIB signed hash data.
               * nocache      - Download VIB to local cache, even if it already
                                exists in the cache.
            Raises:
               * VibIOError         - If an error is encountered attempting to
                                      read or write from the file system.
               * VibDownloadError   - If an error is encountered attempting to
                                      retrieve the remote file.
               * VibFormatError     - If an error is encountered parsing the ar
                                      format or the descriptor XML data.
               * VibSigMissingError - If skipsigcheck is False, and the VIB
                                      archive does not contain signature
                                      information.
               * VibSigFormatError  - If skipsigcheck is False, and the VIB
                                      archive's signature data is not in the
                                      correct format.
               * VibSigInvalidError - If skipsigcheck is False, and the VIB
                                      archive's signature data is not signed by
                                      a trusted publisher.
               * VibSigDigestError  - If skipsigcheck is False, and the
                                      checksum data in the VIB archive's
                                      signature data does not match the
                                      checksum computed from the VIB archive
                                      data.
      """
      cachedir = config.GetConf().default.cachedir
      # Added to work around PR 529643. Otherwise, we will always try to
      # download any VIB passed as a URL to the -b argument.
      vib = None
      for v in self.database.vibs.values():
         if v.remotelocation == url:
            vib = v
            break
      if vib is not None:
         vib.stage(cachedir, self.iface.GetProgress(), skipsigcheck, nocache)
         if nocache:
            self._skipstage[vib.id] = vib
      else:
         # Vib.VIB.fromurl always downloads a VIB, so it is safe to add it to
         # _skipstage, and avoid re-staging it later.
         vib = Vib.VIB.fromurl(url, cachedir, self.iface.GetProgress())
         # Set nocache = False here; only using vib.stage() to extract payload.
         vib.stage(cachedir, self.iface.GetProgress(), skipsigcheck, False)
         self._skipstage[vib.id] = vib

      if config.GetPlatform() == const.PLATFORMS.cos:
         pkgdir = os.path.dirname(vib.packagelocation)
         pkgscanner = scan.PkgScanner(vib.id, 'rpm-dir', pkgdir)
      else:
         pkgscanner = scan.PkgScanner(vib.id, 'deb-file', vib.packagelocation)

      # When loading something from a URL, assume the caller knows what they're
      # doing, and clobber anything already in metadata.vibs.
      self.metadata.replacevib(vib)
      self.metadata.pkgscanner += pkgscanner
      if self.database.vibs.has_key(vib.id):
         self.database.replacevib(vib)
      self._updates.vibs[vib.id] = vib

   def updateall(self):
      """Schedule all bulletins in metadata which meet system requirements for
         updating.
            Returns:  Number of bulletins scheduled for update.
      """
      # Need to scan metadata to make meetssystemreqs meaningful
      self.metadata.scan()

      count = 0
      for bulletin in self.metadata.bulletins.meetssystemreqs.values():
         self._updates.bulletins[bulletin.id] = bulletin
         count += 1
      log.debug('%d bulletins are scheduled for update.' % (count))
      return count

   def remove(self, arg, force=False):
      """Schedule the specified bulletin or VIB for removal.
            Parameters:
               * arg - A bulletin or VIB ID.
            Raises:
               * NoMatchError - The specified argument did not match a bulletin
                                or VIB ID from the metadata or database.
      """
      bulletin = self.bulletins.get(arg)
      if bulletin:
         if not force and bulletin.releasetype != 'extension':
            raise errors.NoMatchError(arg,
                                      'remove command can be used only for '
                                      'extensions which were added after the '
                                      'initial install, not for removing '
                                      'ESX patch and update bulletins.')
         self._removes.bulletins[bulletin.id] = bulletin
      else:
         vib = self.vibs.get(arg)
         if vib:
            if not force:
               # chech whether the vib belogns to 'extension' bulletin.
               for bul in self.bulletins.values():
                  if bul.releasetype == 'extension':
                     continue
                  if vib in bul.vibids:
                     raise errors.NoMatchError(
                        arg,
                        'remove command can be used only for '
                        'extensions which were added after the '
                        'initial install, not for removing '
                        'ESX patch and update bulletin vibs. '
                        'However, the specified vib is a member '
                        'of %s bulletin, %s.' % (bul.releasetype,
                                                 bul.id))
            self._removes.vibs[vib.id] = vib
         else:
            msg = "No Bulletin or VIB found with ID '%s'." % arg
            raise errors.NoMatchError(arg, msg)

   def _solvevibs(self, olderversion):
      for vib in self._removes.vibs.values():
         if vib.state.INSTALLED not in vib.state:
            msg = "Not removing VIB %s; it is not installed." % vib.id
            log.info(msg)
            self.iface.ShowMessage(msg)
            del self._removes.vibs[vib.id]

   def _solvebulletins(self, olderversion):
      for bulletin in self._updates.bulletins.values():
         if not bulletin.meetssystemreqs:
            msg = "Bulletin %s does not meet system requirements." % bulletin.id
            raise errors.NoMatchError(bulletin.id, msg)
         vibcount = 0
         for vib in bulletin.values():
            if vib.obsoletedbyhost and not olderversion:
               msg = ("VIB %s in bulletin %s is obsoleted by the host."
                      % (vib.id, bulletin.id))
               log.debug(msg)
               continue
            if vib.systemreqfailures:
               msg = ("VIB %s in bulletin %s does not meet system requirements."
                      % (vib.id, bulletin.id))
               log.debug(msg)
               continue
            if vib.state.INSTALLED in vib.state:
               msg = ("VIB %s in bulletin %s is already installed."
                      % (vib.id, bulletin.id))
               log.debug(msg)
               continue
            add = True
            for other in self._updates.vibs.values():
               if other.id in vib.oldervibs:
                  del self._updates.vibs[other.id]
               elif vib.id in other.oldervibs:
                  add = False
            if add:
               self._updates.vibs[vib.id] = vib
               vibcount += 1
         if not vibcount:
            del self._updates.bulletins[bulletin.id]
            msg = ("Skipping bulletin %s; it is installed or obsoleted."
                   % bulletin.id)
            self.iface.ShowException(errors.HostNotChanged(msg))

      for bulletin in self._removes.bulletins.values():
         bulletin.scan()
         numapplicable = len(bulletin.meetssystemreqs)
         if bulletin.obsoleted:
            msg = ("%d of %d VIBs in bulletin %s have been superseded. "
                   "Please try removing the newest bulletin for this "
                   "component instead."
                   % (len(bulletin.obsoleted), numapplicable, bulletin.id))
            raise errors.NoMatchError(bulletin.id, msg)
         elif not bulletin.installed:
            msg = ("No VIBs from this bulletin are installed.")
            raise errors.NoMatchError(bulletin.id, msg)
         for vib in bulletin.values():
            if vib.state.INSTALLED in vib.state:
               self._removes.vibs[vib.id] = vib

   def solve(self, **kwargs):
      """Resolve all bulletins to a list of VIBs, and check dependencies.
            Parameters:
               * nodeps       - Do not check for unmet dependencies.
               * noconflicts  - Do not check for conflicts.
               * olderversion - Do not check for obsoletion by an already
                                installed package.
            Returns: A VIBCollection instance containing any additional VIBs
                     suggested to resolve dependencies.
            Raises:
               * DependencyError - If dependency resolution fails.
               * NoMatchError    - If one or more selected VIBs or Bulletins
                                   does not match the host platform.
      """
      olderversion = kwargs.get('olderversion', False)

      self.metadata.scan()

      # At this point, everything here was specifically added with the update()
      # method, as we haven't done any bulletin expansion yet.
      vibmatchfailures = self._updates.vibs.failssystemreqs
      bulletinmatchfailures = self._updates.bulletins.failssystemreqs

      if vibmatchfailures:
         msg = ("One or more requested VIBs did not match the host "
                "platform:\n%s" % '\n'.join(vibmatchfailures))
         raise errors.NoMatchError(None, msg)
      if bulletinmatchfailures:
         msg = ("One or more requested bulletins did not match the host "
                "platform:\n%s" % '\n'.join(bulletinmatchfailures))
         raise errors.NoMatchError(None, msg)

      self._solvevibs(olderversion)
      self._solvebulletins(olderversion)

      # Fix me: Removals are not factored into dependency resolution, and they
      # really should be. For now, we rely on the package manager to complain
      # if we're trying to remove something that we shouldn't be. Due to some
      # issues on Visor, however, ipkg will remove things with blithe
      # indifference to dependencies.
      solver = depsolver.DepSolver(self._updates.vibs, self.metadata, **kwargs)
      deps = solver.Resolve()

      return deps

   def clear(self):
      "Clear the list of Bulletins and VIBs schedule for update or removal."
      self._updates.vibs.clear()
      self._updates.bulletins.clear()
      self._removes.vibs.clear()
      self._removes.bulletins.clear()

   def updatedb(self):
      """Scan the host against metadata and database, and add compliant
         Bulletins and installed and obsoleted VIBs to the database.
            Raises:
               * DatabaseIOError     - On a failure to write to a file, or if
                                       database has changed since last read
                                       (serial number has been incremented).
      """
      log.info('Updating esxupdate database...')
      self.database.scan()
      self.metadata.scan()
      for bulletin in self.bulletins.values():
         if not bulletin.persist:
            continue
         if bulletin.compliant:
            self.database.bulletins[bulletin.id] = bulletin
            for vib in bulletin.values():
               # Note that we do not check the vib.persist value here.
               self.database.vibs[vib.id] = vib
      for vib in self.vibs.values():
         if not vib.persist:
            continue
         if vib.state & set((vib.state.INSTALLED, vib.state.RETIRED)):
            self.database.vibs[vib.id] = vib
            # This shouldn't really go here...
            vib.state.discard(vib.state.STAGED)

      self.database.save()

   def _stage(self, skipsigcheck = False, nocache = False):
      cachedir = config.GetConf().default.cachedir
      for v in self._updates.vibs.values():
         # Short-circuit to avoid re-staging things already downloaded by
         # addvibfromurl. PR 529643.
         if v.id in self._skipstage and id(v) == id(self._skipstage[v.id]):
            pass
         elif nocache or v.state.STAGED not in v.state:
            v.stage(cachedir, self.iface.GetProgress(), skipsigcheck, nocache)
            # Avoid staging twice in any case:
            if nocache:
               self._skipstage[v.id] = v
         self.database.vibs[v.id] = v

   def stage(self, skipsigcheck = False, nocache = False):
      """Solve the current transaction, and download any VIBs pending update
         to the local cache.
            Parameters:
               * skipsigcheck - Do not validate VIB signed hash data.
               * nocache      - Download VIBs to local cache, even if they
                                already exist in the cache.
            Raises:
               * UnsatisifedDependencies - If there are unresolved dependencies
                                           in the set of VIBs pending update.
               * DependencyError         - If dependency resolution fails.
               * NoMatchError            - If one or more selected VIBs or
                                           Bulletins do not match the host
                                           platform.
               * VibIOError              - If an error is encountered attempting
                                           to read or write from the file
                                           system.
               * VibDownloadError        - If an error is encountered attempting
                                           to retrieve a remote file.
               * VibFormatError          - If an error is encountered parsing
                                           the ar format or the descriptor XML
                                           data.
               * VibSigMissingError      - If skipsigcheck is False, and the VIB
                                           archive does not contain signature
                                           information.
               * VibSigFormatError       - If skipsigcheck is False, and the VIB
                                           archive's signature data is not in
                                           the correct format.
               * VibSigInvalidError      - If skipsigcheck is False, and the VIB
                                           archive's signature data is not
                                           signed by a trusted publisher.
               * VibSigDigestError       - If skipsigcheck is False, and the
                                           checksum data in the VIB archive's
                                           signature data does not match the
                                           checksum computed from the VIB
                                           archive data.
      """
      deps = self.solve()
      if deps:
         msg = 'Stage cannot complete due to unmet dependencies.  Esxupdate ' \
               'suggests adding the following VIBs to resolve the ' \
               'dependencies: %s' % (', '.join(deps.keys()))
         raise errors.UnsatisfiedDependencies(deps, msg)

      try:
         self.database.load(keeplock = True)
         self._stage(skipsigcheck, nocache)
         self.database.save()
         if config.GetPlatform() == 'visor':
            platform.conf.UpdateStagedDb()
      finally:
         self.database.unlock()

   def _getremovepkgs(self):
      removepkgs = set()
      for v in self._removes.vibs.values():
         try:
            vibtype, name, version = v.tnv
            # rpm expects [epoch:]ver-rel.arch not @arch
            version = version.replace('@', '.')
            if config.GetPlatform() == 'cos':
               removepkgs.add('-'.join((name, version)))
            else:
               removepkgs.add(name)
         except:
            log.warning('Unable to get package name for VIB %s' % v.id)
      return removepkgs

   def _run(self, nodeps=False, olderversion=False, noconflicts=False):
      removepkgs = self._getremovepkgs()
      log.debug('Removing %s' % (removepkgs))

      config.CreateVibEnv(self.VIBCONF)

      try:
         platform.PackageTransaction(self._updates.vibs, removepkgs,
                                     progress=self.iface.GetProgress(),
                                     nodeps=nodeps, olderversion=olderversion
                                     # TODO: hook up 'noconflicts'
                                    )
      finally:
         try:
            self.updatedb()
         except Exception, e:
            log.warn('Failed to update esxupdate database after a package '
                     'transaction failure: %s. The database may be '
                     'inconsistent with the package manager state.' % str(e))

      for script in config.GetPostUpdateScripts(self.VIBCONF):
         self.iface.ShowMessage('Running [%s]...' % (script))
         log.info('Running [%s]...' % (script))
         rc, out = util.runcommand.runcommand(script)
         log.info(out)
         if rc != 0:
            raise errors.PostScriptError(script, rc)
         else:
            self.iface.ShowMessage('ok.')

      # Clear cache.
      try:
         self._skipstage.clear()
         platform.CheckCache(cachesize=None, cleancache=True, doumount=True)
      except Exception, e:
         log.warning('Error cleaning up cache dir: %s' % e)

      # Finish visor installation
      if config.GetPlatform() == 'visor':
         platform.conf.UpdateBootbank()

   def run(self, **kwargs):
      """Run the transaction, solving dependencies and removing or installing
         VIBs scheduled for removal or update.
            Keyword Parameters:
               * installdeps  - If True, causes unmet dependencies that can be
                                resolved from metadata to be added to the
                                the transaction. Defaults to True.
               * nommode      - If True, does not verify that the host is in
                                maintenance mode when one or more VIBs require
                                it. Defaults to False.
               * skipsigcheck - If True, signed hash data for VIBs will not be
                                verified. Defaults to False.
               * nocache      - If True, VIBs which are already in the cache
                                will be downloaded again. Defaults to False.
               * nodeps       - Do not check for unmet dependencies, and pass
                                the nodeps flag to the package manager.
               * noconflicts  - Do not check for conflicts.
               * olderversion - Do not check for obsoletion by an already
                                installed package.
            Returns: SUCCESS, SUCCESS_REQUIRES_REBOOT,
                     SUCCESS_REQUIRES_HOSTD_RESTART or SUCCESS_HOST_NOT_CHANGED.
            Raises:
               * UnsatisifedDependencies - If there are unresolved dependencies
                                           in the set of VIBs pending update.
               * DependencyError         - If dependency resolution fails.
               * NoMatchError            - If one or more selected VIBs or
                                           Bulletins do not match the host
                                           platform.
               * VibIOError              - If an error is encountered attempting
                                           to read or write from the file
                                           system.
               * VibDownloadError        - If an error is encountered attempting
                                           to retrieve a remote file.
               * VibFormatError          - If an error is encountered parsing
                                           the ar format or the descriptor XML
                                           data.
               * VibSigMissingError      - If skipsigcheck is False, and the VIB
                                           archive does not contain signature
                                           information.
               * VibSigFormatError       - If skipsigcheck is False, and the VIB
                                           archive's signature data is not in
                                           the correct format.
               * VibSigInvalidError      - If skipsigcheck is False, and the VIB
                                           archive's signature data is not
                                           signed by a trusted publisher.
               * VibSigDigestError       - If skipsigcheck is False, and the
                                           checksum data in the VIB archive's
                                           signature data does not match the
                                           checksum computed from the VIB
                                           archive data.
               * MaintenanceModeError    - If one or more VIBs require
                                           maintenance mode, nommode is False,
                                           and the host is not in maintenance
                                           mode (or host's maintenance mode
                                           status could not be determined).
               * DatabaseIOError         - On a failure to write to write the
                                           database.
      """
      installdeps  = kwargs.pop('installdeps',  True)
      nommode      = kwargs.pop('nommode',      False)
      skipsigcheck = kwargs.pop('skipsigcheck', False)
      nocache      = kwargs.pop('nocache',      False)

      solverkwargs = dict()
      for arg in ('nodeps', 'noconflicts', 'olderversion'):
         if arg in kwargs:
            solverkwargs[arg] = kwargs.pop(arg)

      if kwargs:
         msg = "Unrecognized keyword argument(s): '%s'" % ', '.join(kwargs)
         raise TypeError(msg)

      deps = self.solve(**solverkwargs)
      if deps and not installdeps and not self.iface.DependencyPrompt(deps):
         raise errors.UnsatisfiedDependencies(deps, 'Exiting on user input.')

      self._updates.vibs.update(deps)

      if not self._updates.vibs and not self._removes.vibs:
         log.debug("Nothing to update or remove.")
         try:
            self.database.load(keeplock = True)
            self.updatedb()
         finally:
            self.database.unlock()
         return self.SUCCESS_HOST_NOT_CHANGED

      mmodevibs = set(v.id for v in self._updates.vibs.values()
                      if v.requiresmaintenancemode)
      mmodevibs.update(v.id for v in self._removes.vibs.values()
                      if v.requiresmaintenancemode)

      if not nommode and config.IsESXhost() and mmodevibs:
         if not platform.GetMaintenanceMode():
            msg = ("The following VIBs require this host to be in maintenance "
                   "mode: %s. Please put the host in maintenance mode to "
                   "proceed." % ', '.join(mmodevibs))
            raise errors.MaintenanceModeError(msg)

      self._stage(skipsigcheck = skipsigcheck, nocache = nocache)
      self.database.save()
      #
      # Make sure we write updated database to /(alt)bootbank on Visor
      # so in case of failure during install, we can revert the db
      #
      if config.GetPlatform() == 'visor':
         platform.conf.UpdateStagedDb()

      try:
         self.database.load(keeplock = True)
         try:
            self._run(**solverkwargs)
         except Exception:
            if config.GetPlatform() == 'visor':
               platform.conf.Revertdb()
            raise
      finally:
         #
         # Cleanup after an update on visor
         #
         if config.GetPlatform() == 'visor':
            platform.conf.CleanupStagingdir()

         self.database.unlock()

      if self._updates.vibs.rebootrequired or self._removes.vibs.rebootrequired:
         return self.SUCCESS_REQUIRES_REBOOT
      elif self._updates.vibs.hostdrestart or self._removes.vibs.hostdrestart:
         return self.SUCCESS_REQUIRES_HOSTD_RESTART
      else:
         return self.SUCCESS
