#######################################################################
# Copyright (C) 2009 VMWare, Inc.
# All Rights Reserved
########################################################################
#
# VIBCollection.py
#

__all__ = ['VIBCollection']

import config
import const

class VIBCollection(dict):
   """This class represents a collection of VIB objects and provides
      methods and properties for modifying the collection and determining
      the relationships between them.
   """
   @property
   def meetssystemreqs(self):
      return VIBCollection((k, v) for (k, v) in self.items()
                           if v.meetssystemreqs)

   # Note that bool(self.meetssystemreqs) == True does not imply that
   # bool(self.failssystemreqs) == False.

   @property
   def failssystemreqs(self):
      return VIBCollection((k, v) for (k, v) in self.items()
                           if not v.meetssystemreqs)

   @property
   def newervibs(self):
      return VIBCollection((k, v) for (k, v) in self.matches.items()
                           if v.state.INSTALLED not in v.state
                           and not v.obsoletedbyhost)

   @property
   def rebootrequired(self):
      for vib in self.values():
         if vib.rebootrequired:
            return True
      return False

   @property
   def hostdrestart(self):
      for vib in self.values():
         if vib.hostdrestart:
            return True
      return False

   def update(self, other):
      # Think about whether we need to change this.
      dict.update(self, other)

   def sync(self, other):
      """Updates the common keys in this object and other to point to the same
         VIB objects. It does not affect values that are not common between
         the two.
            Parameters:
               * other - Another VIBCollection instance.
      """
      for key in set(self.keys()) & set(other.keys()):
         if id(self[key]) != id(other[key]):
            v = self[key] + other[key]
            self[key] = v
            other[key] = v

   def __add__(self, other):
      """Merge this collection with another to form a new collection consisting
         of the union of VIBs from both. The new collection may contain
         references to VIBs from the left operand, the right, or neither.  Keys
         which exist in both the left and right operands and do not reference
         the same objects will result in a new object being created. (See notes
         for VIB.__add__ method.)
            Parameters:
               * other - another VIBCollection instance.
            Returns: A new VIBCollection instance.
      """
      new = VIBCollection()
      new.update(self)
      for v in other.values():
         if v.id in new and id(v) != id(new[v.id]):
            new[v.id] += v
         else:
            new[v.id] = v
      return new

   def getbynameversion(self, name, version):
      """Returns a VIB instance matching (name, version), or None if no such
         VIB exists.
      """
      # This is the only way we tie a VIB to its corresponding package.  A
      # little flimsy, na? -RS
      ver = str(version)

      for t in const.SUPPORTED[config.GetConf().platform.product]:
         if t == const.VIBTYPES.cross and '@' in ver:
            ver = ver[:ver.index('@')]
         key = '%s_%s_%s' % (t, name, ver)
         if key in self:
            return self[key]
      return None      

   def getbyname(self, name):
      """Returns a list of Vibs with the name matching name.
            Parameters:
               * name - string giving the VIB name.
            Returns: A list of matching VIBs.
      """
      # TODO: optimize this if it's gonna be used a lot
      return [v for v in self.values() if v.name == name]
