#######################################################################
# Copyright (C) 2009 VMWare, Inc.
# All Rights Reserved
########################################################################
#
# VIBScanner.py
#

__all__ = ['VIBScanner']

"""This module defines a base class for objects which support scanning over a
   set of VIB and Bulletin objects.

   You probably do not want to import this module directly.  Instead, use the
   appropriate metadata or database module(s).
"""

import logging
log = logging.getLogger(__name__)

import Bulletin
import Vib
import VIBCollection
import config
import errors
import scan

class VIBScanner(object):
   """This is a base class which implements methods to scan a group of VIBs and
      Bulletins.
         Attributes:
            * bulletins  - A BulletinCollection instance containing the
                           Bulletins being scanned.
            * vibs       - A VIBCollection instance containing the VIBs being
                           scanned.
            * brokenvibs - A VIBCollection of installed VIBs with dependencies
                           that may be broken by replacing an installed package.
            * brokendeps - A dictionary of Requires dependencies that may be
                           broken, keyed by requirement name.
   """
   def __init__(self, vibs=None, bulletins=None, usepkgdb=True):
      """Class constructor.
            Parameters:
               * vibs      - If specified, must be an instance of VIBCollection
                             to bind to this object's 'vibs' attribute.
               * bulletins - If specified, must be an instance of
                             BulletinCollection to bind to this object's
                             'bulletins' attribute.
               * usepkgdb  - A boolean to determine whether to include VIB
                             relationships with packages already installed on
                             the host.
            Returns: A new VIBScanner instance.
            Note:
               The values for the vibs and bulletins parameters are bound to
               this object's attributes without copying, so this object's
               'vibs' and 'bulletins' attributes will simply be references to
               the objects you passed to the constructor.
      """
      self.vibs = vibs or VIBCollection.VIBCollection()
      self.bulletins = bulletins or Bulletin.BulletinCollection()
      # Always set default PkgScanner...
      self.pkgscanner = scan.PkgScanner(usesyschannel=usepkgdb)
      self.usepkgdb = usepkgdb
      self.brokenvibs = VIBCollection.VIBCollection()
      self.brokendeps = dict()

   def scan(self):
      """Calls the appropriate PkgScanner instance for this collection, and
         maps packages and package constraints to VIBs. Fills out VIB system
         requirements and other scan data. Fills out Bulletin scan data.  The
         following VIB attributes may be modified by this method:
            * requires
            * conflicts
            * conflictswithhost
            * oldervibs
            * obsoletedbyhost
            * pkgstates

      """

      log.debug("Loading repomd/apt metadata...")
      try:
         self.pkgscanner.Load()
      except Exception, e:
         msg = "Failed scanning package data: %s." % e
         raise errors.MetadataFormatError(None, msg)

      # Fill out system req's, installed for each Vib
      for vib in self.vibs.values():
         # In case Scan() is run multiple times, clear results first
         vib.scanned = False
         # Why do we do this in scan?
         if not vib.swplatforms.matcheshost:
            for reason in vib.swplatforms.problems:
               vib.systemreqfailures.add(reason)
         # Skip hwplatform check if config.IsESXhost is False (including
         # offline case for vibddi).
         if config.IsESXhost() and not vib.hwplatforms.matcheshost:
            for reason in vib.hwplatforms.problems:
               vib.systemreqfailures.add(reason)

      # Now do a package scan and map results to VIB IDs, if possible
      #
      # NOTE: For obsoletes and conflicts, sets of host package (name, version)
      # tuples are kept. These include both packages with VIB metadata and host
      # packages with no VIB data. The (n,v) pairs allow tracing of
      # dependencies via packages with no Vib instances, such as:
      #    * VIB A conflicts with host pkg (Bn, Bv), which is obsoleted by
      #      VIB C.
      #    * (Bn, Bv) would be in VIB A's hostconflicts and VIB C's
      #      hostobsoleted
      log.debug("Computing rpm/dpkg dependencies...")
      pkglist = self.pkgscanner.Scan()
      log.log(1, "Mapping pkg scan results back to VIB level...")

      # Use this to track VIBs whose corresponding package is no longer
      # installed. See comment after the 'for pkg in pkglist' loop.
      vibs_with_noninstalled_pkg = set()

      reqcache = {}

      for pkg in pkglist:
         if pkg.name.startswith('cisco-vem') and pkg.installed:
            Vib.cisco_packages[pkg.name] = pkg
         vib = self.vibs.getbynameversion(pkg.name, pkg.version)

         # Which packages are older?
         for obsoletion in pkg.obsoletes:
            # Save each obsolete constraint info in the VIB, so obsolescence
            # chain can be recreated when native package metadata is missing
            # (eg for VibDB).
            if vib:
               otuple = (obsoletion.name, obsoletion.relation,
                         obsoletion.version)
               vib.replaces.append(otuple)
            for olderpkg in obsoletion.packages:
               oldvib = self.vibs.getbynameversion(olderpkg.name,
                                                   olderpkg.version)
               if oldvib:
                  if vib:
                     vib.oldervibs.add(oldvib.id)
                  if pkg.installed:
                     oldvib.hostobsoleting.add((pkg.name, pkg.version))
               if vib and olderpkg.installed:
                  vib.hostobsoleted.add((olderpkg.name, olderpkg.version))

         # Go through conflicts
         for conflict in pkg.conflicts:
            for confpkg in conflict.packages:
               confvib = self.vibs.getbynameversion(confpkg.name,
                                                    confpkg.version)
               if confvib and vib:
                  vib.conflicts.add(confvib.id)
                  confvib.conflicts.add(vib.id)
               if vib and confpkg.installed:
                  vib.hostconflicts.add((confpkg.name, confpkg.version))
               if confvib and pkg.installed:
                  confvib.hostconflicts.add((pkg.name, pkg.version))

         if vib is None:
            log.log(1, "Couldn't find VIB matching (%s, %s)" % (pkg.name,
                                                                pkg.version))
            continue

         # 
         # Go through requires, translate to vibs, compute providedByHost
         for pkgreq in pkg.requires:
            # Use req from cache to save time & memory
            if pkgreq.key in reqcache:
               vib.requires.append(reqcache[pkgreq.key])
            else:
               req = Vib.Requires(pkgreq.key)
               for prov in pkgreq.packages:
                  provvib = self.vibs.getbynameversion(prov.name, prov.version)
                  if provvib:
                     req.providedby.add(provvib.id)
                  if prov.installed:
                     req.hostprovides.add((prov.name, prov.version))
               reqcache[pkgreq.key] = req
               vib.requires.append(req)

         if pkg.installed:
            vib.setpkgstate(vib.state.INSTALLED)
         else:
            # We don't yet know whether the correct pkgstate is retired or
            # uninstalled, so we just record for now the fact that this VIB
            # doesn't have its pkg installed on the host. See comment following
            # this loop.
            vibs_with_noninstalled_pkg.add(vib)

         vib.scanned = True

      # VIBs which no longer have their corresponding pkg installed on the host
      # may fall into two categories:  uninstalled or retired. We should rely
      # on the 'obsoletedByHost' attribute to know which is the case. However,
      # we need to finish our iteration through the 'for pkg in pkglist' loop
      # before 'obsoletedByHost' can be accurate. This is why we record the
      # list of VIBs with no installed pkg during the loop, and set the correct
      # pkgstate once the loop has finished.
      for vib in vibs_with_noninstalled_pkg:
         # Note that if our Vib is in the vibs_with_noninstalled_pkg set, it
         # means that we actually have a Pkg object from pkgscanner, and so the
         # obsoletedByHost attribute should be authoritative.
         if vib.obsoletedbyhost:
            vib.setpkgstate(vib.state.RETIRED)
            # May be staged, which seems a bit silly for being obsoleted.
            vib.state.discard(vib.state.STAGED)
         else:
           vib.setpkgstate(vib.state.UNINSTALLED)

      # All the vibs without package scan results must be of the wrong platform
      # (no yum meta for visor vibs), or for Visor the arch is unsupported.
      # Right now, there's no easy way to distinguish the two.
      for vib in self.vibs.values():
         if not vib.scanned:
            # This stuff doesn't get written back to the database, but is used
            # when providing scan results output to VUM.  If we ever decide to
            # add this info to the esxupdate database, we will need to find a
            # way to differentiate cases where the package is missing in the
            # meta-data, or the arch doesn't match the platform.
            if config.GetPlatform() == 'cos':
               reason = "No rpm metadata in repodata/*"
            else:
               reason = "Vib not in Packages.gz or of the wrong arch"
            vib.systemreqfailures.add(reason)
            vib.scanned = True

      #
      # Compute indexes for checking host dpendency breakage
      self.brokenvibs.clear()
      self.brokendeps.clear()
      for hostreq, brokenpkgs in self.pkgscanner.replaceddeps.items():
         if not hostreq in reqcache:
            continue
         #
         # Get the Requires() object for each potentially broken
         # dependency.  Brokendeps becomes a convenient dict with an entry
         # for each broken dep and the providedby and hostprovides can
         # be accessed from the Requires() object.
         brokendep = self.brokendeps.setdefault(hostreq, reqcache[hostreq])
         for pkg in brokenpkgs:
            brokenvib = self.vibs.getbynameversion(pkg.name, pkg.version)
            #
            # Add the VIB IDs for up/downgrades that break an existing
            # dep by not providing it
            if brokenvib:
               brokendep.breaks.add(brokenvib.id)
         for pkg in self.pkgscanner.requiredby[hostreq]:
            #
            # Add the host package names and versions requiring this dep
            brokendep.hostrequiredby.add((pkg.name, pkg.version))
            depvib = self.vibs.getbynameversion(pkg.name, pkg.version)
            if depvib:
               self.brokenvibs[depvib.id] = depvib

      for bul in self.bulletins.values():
         bul.scan()

      self.pkgscanner.RemoveAllChannels()

   def filteroutbyid(self, ids):
      """Prunes the scan results, setting scanned status to False for VIBs
         matching ids, and removing any references to VIBs matching ids
         in existing requires, obsoletes, conflicts vib lists.
         * ids   - an iterable of VIB ID strings to prune
      """
      idset = set(ids)
      count = 0
      for vib in self.vibs.values():
         if vib.id in idset:
            vib.scanned = False
            count += 1
         elif vib.scanned:
            vib.oldervibs -= idset
            vib.conflicts -= idset
            # It'd be far more efficient to go through reqcache, but oh well.
            for req in vib.requires:
               req.providedby -= idset
      log.debug('Pruned %d VIBs from scan results' % (count))

   def sync(self, other=None):
      # Create a temporary collection to hold VIB references.
      mergedvibs = VIBCollection.VIBCollection()

      # Add each VIBCollection in self to the temporary collection. The __add__
      # method automatically takes care of calling VIB.__add__ when the key
      # values do not reference the same objects.
      mergedvibs += self.vibs
      for bulletin in self.bulletins.values():
         mergedvibs += bulletin

      # Repeat for 'other', if necessary.
      if other:
         mergedvibs += other.vibs
         for bulletin in other.bulletins.values():
            mergedvibs += bulletin

      # Now replace all of the VIB references in 'self' with the cumulative
      # values in the temporary collection.
      for vid in self.vibs:
         self.vibs[vid] = mergedvibs[vid]
      for bulletin in self.bulletins.values():
         for vid in bulletin:
            bulletin[vid] = mergedvibs[vid]

      # Do the same for 'other', if necessary.
      if other:
         for vid in other.vibs:
            other.vibs[vid] = mergedvibs[vid]
         for bulletin in other.bulletins.values():
            for vid in bulletin:
               bulletin[vid] = mergedvibs[vid]
         # Finally, sync the bulletin collections between self and other. We
         # don't need to worry about VIB references changing here, since they
         # should already all be the same references from mergedvibs.
         self.bulletins.sync(other.bulletins)

   def get(self, name, default=None):
      return self.bulletins.get(name) or self.vibs.get(name) or default

   def __add__(self, other):
      """Merge this scanner with another.
            Parameters:
               * other - another VIBScanner instance.
            Returns: A new VIBScanner instance.
      """
      new = VIBScanner()
      new.vibs = self.vibs + other.vibs
      new.bulletins = self.bulletins + other.bulletins
      new.sync()
      new.pkgscanner = self.pkgscanner + other.pkgscanner
      return new

   def replacevib(self, vib):
      """Replaces a VIB in self.vibs as well as in each bulletin that
         references the same VIB ID.
            Parameters:
               vib - The VIB object to insert as a replacement. Note that if
                     a VIB with the same ID does not already exist in vibs,
                     this method adds it.
      """
      self.vibs[vib.id] = vib
      for bulletin in self.bulletins.values():
         if vib.id in bulletin:
            bulletin[vib.id] = vib
            if not bulletin.id in vib.bulletins:
               vib.bulletins[bulletin.id] = bulletin

   def replacebulletin(self, bulletin):
      """Replaces a bulletin in self.bulletins. Each VIB in the replaced
         bulletin object will displace any other VIBs in this object or in any
         bulletin in this object.
            Parameters:
               bulletin - The Bulletin object to insert as a replacement. Note
                          that if a Bulletin with the same ID does not already
                          exist in bulletins, this method adds it.
      """
      self.bulletins[bulletin.id] = bulletin
      for vib in bulletin.values():
         self.replacevib(vib)
      for vib in self.vibs.values():
         if vib.id not in bulletin and bulletin.id in vib.bulletins:
            del vib.bulletins[bulletin.id]
