########################################################################
# Copyright (C) 2009 VMWare, Inc.
# All Rights Reserved
########################################################################

"""This module contains classes necessary to parse, extract, and validate
   VIB archives and VIB XML meta-data.
"""

__all__ = ['VIB']

import binascii
import datetime
import gzip
import os
import shutil
import tarfile
import tempfile
import time

import Bulletin
import Descriptor
import config
import const
import downloader
import errors
import util
import util.arfile
import util.cpio
import version

try:
   from lxml import etree
   xmlerrclass = etree.XMLSyntaxError
except ImportError:
   try:
      import xml.etree.cElementTree as etree
   except ImportError:
      import cElementTree as etree
   from xml.parsers.expat import ExpatError
   xmlerrclass = ExpatError

# used to keep track of any cisco packages which are installed
# PR499703
cisco_packages = {}

try:
   import rpm
   HAVE_RPM = True
except:
   HAVE_RPM = False

try:
   import PyPKCS
   HAVE_PYPKCS = True
except:
   HAVE_PYPKCS = False

HASH_CHUNK_SIZE = 16 * 1024

# Visor has py2.5, COS has M2Crypto.
# Should this be in util module? -RS
try:
   import hashlib
   hasher = hashlib.sha256
except ImportError:
   from M2Crypto import EVP
   class hasher:
      def __init__(self):
         self.md = EVP.MessageDigest('sha256')

      def update(self, chunk):
         self.md.update(chunk)

      def hexdigest(self):
         digest = self.md.final()
         return "".join(["%02x" % ord(ch) for ch in digest])

class SWPlatform(object):
   """A class for representing a software platform.
         Attributes (read-only Properties):
            * productid - A string representing the product ID. Currently,
                          'esx', 'embeddedEsx' and 'vima' are in use.
            * productname - A string representing the public product name
                          corresponding to the productid.  The list of public
                          names can be found in const.py.
            * version   - An instance of ESXVersion, representing the version
                          of this software platform specification.
            * locale    - An string representing a locale name, or None if
                          unused. (Currently not implemented.)
            * matcheshost - A boolean; True if this SWPlatform matches the
                          host platform SWPLATFORM;  False if this SWPlatform
                          does not match the host;  None if SWPLATFORM is not
                          defined.
   """
   def __init__(self, productid='', ver=version.ESXVersion(), locale=''):
      """Class constructor.
            Parameters:
               * productid - An optional value to assign to the productid
                             attribute.
               * ver       - Either an instance of ESXVersion, or an
                             appropriately formatted string which can be
                             used to create an ESXVersion object using
                             ESXVersion.fromstring(). Assigned to version
                             attribute.
               * locale    - A string to be assigned to the locale attribute.
            Raises:
               * ValueError - if 'ver' parameter is not an ESXVersion object,
                              and cannot be parsed using
                              ESXVersion.fromstring().
      """
      if isinstance(ver, version.ESXVersion):
         self._version = ver
      else:
         self._version = version.ESXVersion.fromstring(ver)
      self._productid = productid
      self._locale = locale

      # Profiling shows that SWPlatform.matcheshost() is called many times
      # and is slow cumulatively, so pre-compute the answer for speed.
      if SWPLATFORM:
         self._matcheshost = self.matches(SWPLATFORM)
      else:
         self._matcheshost = None

   def __hash__(self):
      # We want to support adding this to a set(), so we must be hashable.
      # Incidentally, this is also why our attributes are made available as
      # read-only properties. -RS
      return hash((self._productid, self._version, self._locale))

   def __cmp__(self, other):
      # This doesn't really make sense in terms of comparing SWPlatform objects;
      # it is only here to satisfy the requirement that objects with the same
      # hash are equal.
      mine = (self._productid, self._version, self._locale)
      theirs = (other.productid, other.version, other.locale)
      return cmp(mine, theirs)

   def __str__(self):
      ver = self._version.versionstring
      return '-'.join(x for x in (self._productid, ver, self._locale) if x)

   productid = property(lambda self: self._productid)
   version   = property(lambda self: self._version)
   locale    = property(lambda self: self._locale)

   @property
   def productname(self):
      for productname in const.EXT2INT:
         if self.productid == const.EXT2INT[productname]:
            return productname
      # shouldn't be here
      return ''

   def _problemstr(self, host):
      if self._productid and self._productid != host.productid:
         return ("This VIB supports product ID '%s', but host reports product "
                 "ID as '%s'." % (self._productid, host.productid))
      if not self.version.matches(host.version):
         return ("This VIB supports product version '%s', but host reports "
                 "product version as '%s'." % (self._version, host.version))
      if self._locale and self._locale != host.locale:
         return ("This VIB supports locale '%s', but host reports locale as "
                 "'%s'." % (self._locale, host.locale))
      return None

   def matches(self, host):
      """Determine if this SWPlatform object matches another.
            Parameters:
               * other - Another instance of SWPlatform.
            Returns:
               * True  - If self.productid, self.version and self.locale
                         match the software platform information in 'host'.
                         Note that if self.locale is None, it will match
                         any value of host.locale. If self.productid is '',
                         it will match any value of host.productid. The
                         ESXVersion.matches() method is used to determine
                         matching between self.version and host.version.
               * False - False if this SWPlatform object does not match the
                         'host' parameter.
      """
      return self._problemstr(host) is None

   matcheshost = property(lambda self: self._matcheshost)

   @property
   def problem(self):
      """Return a string describing why this SWPlatform does not match the
         host.
      """
      host = SWPlatform(config.GetConf().platform.product,
                        config.GetConf().platform.version)
      return self._problemstr(host)

SWPLATFORM = None

class HWPlatform(object):
   """A class for representing a hardware platform.
         Attributes (read-only Properties):
            * vendor - A string representing the hardware vendor.
            * model  - A string representing the hardware model.
   """
   def __init__(self, vendor='', model=''):
      """Class constructor.
            Parameters:
               * vendor - An optional string value to assign to the vendor
                          attribute.
               * model  - An optional string value to assign to the model
                          attribute.
      """
      self._vendor = vendor
      self._model = model

   def __hash__(self):
      # We want to support adding this to a set(), so we must be hashable.
      # Incidentally, this is also why our attributes are made available as
      # read-only properties. -RS
      return hash((self._vendor, self._model))

   def __cmp__(self, other):
      mine = (self._vendor, self._model)
      theirs = (other._vendor, other._model)
      return cmp(mine, theirs)

   def __str__(self):
      return '-'.join(x for x in (self._vendor, self._model) if x)

   vendor = property(lambda self: self._vendor)
   model  = property(lambda self: self._model)

   def _problemstr(self, host):
      if self._vendor and self._vendor != host.vendor:
         return ("This VIB supports hardware vendor '%s', but host reports "
                 "vendor as '%s'." % (self._vendor, host.vendor))
      if self._model and self._model != host.model:
         return ("This VIB supports hardware model '%s', but host reports "
                 "model as '%s'." % (self._model, host.model))
      return None

   def matches(self, host):
      """Determine if this HWPlatform object matches another.
            Parameters:
               * other - Another instance of HWPlatform.
            Returns:
               * True  - If self.vendor and self.model match the hardware
                         platform information in 'host'. Note that if
                         self.vendor is '', it will match any value of
                         host.vendor. If self.model is '', it will match any
                         value of host.model.
               * False - False if this HWPlatform object does not match the
                         'host' parameter.
      """
      return self._problemstr(host) is None

   @property
   def matcheshost(self):
      """Determine if this HWPlatform object matches the host.
            Returns: True if matches, False if not.
            Raises:
               * RuntimeError - if setswplatform module function has not been
                                called (HWPLATFORM is None).
         See matches() method for more information.
      """
      if HWPLATFORM is None:
         msg = "Attempt to match host platform before HWPLATFORM is set."
         raise RuntimeError(msg)
      return self.matches(HWPLATFORM)

   @property
   def problem(self):
      """Return a string describing why this HWPlatform does not match the
         host.
      """
      return self._problemstr(HWPLATFORM)

HWPLATFORM = HWPlatform()

class SWPlatformCollection(set):
   """A collection of 0 or more SWPlatform objects."""
   # We could do some type validation here, but we assume input is trusted.
   def __str__(self):
      return ', '.join(str(x) for x in self)

   @property
   def matcheshost(self):
      """Determine if any SWPlatform in this Collection matches the host.
         Returns: True or False
      """
      if not self:
         return True
      for swplatform in self:
         if swplatform.matcheshost:
            return True
      return False

   @property
   def problems(self):
      rc = set(x.problem for x in self)
      rc.discard(None)
      return rc

class HWPlatformCollection(set):
   """A collection of 0 or more HWPlatform objects."""
   def __str__(self):
      return ', '.join(str(x) for x in self)

   @property
   def matcheshost(self):
      """Determine if any HWPlatform in this Collection matches the host.
         Returns: True or False
      """
      if not self:
         return True
      for hwplatform in self:
         if hwplatform.matcheshost:
            return True
      return False

   @property
   def problems(self):
      rc = set(x.problem for x in self)
      rc.discard(None)
      return rc

class VIBState(set):
   """The VIBState class is used to track package states. Values must be one of
      UNINSTALLED, INSTALLED, RETIRED, PENDING or STAGED. UNINSTALLED,
      INSTALLED and RETIRED are exclusive to one another.  Calling add() or
      update() with one of these states will cause the other two to be removed.
      Calling update() or __init__() with more than one of these results in a
      ValueError exception.
   """
   UNINSTALLED = 'uninstalled'
   INSTALLED   = 'installed'
   RETIRED     = 'retired'
   PENDING     = 'pending'
   STAGED      = 'staged'

   states = (UNINSTALLED, INSTALLED, RETIRED, PENDING, STAGED)
   _exclusives = (UNINSTALLED, INSTALLED, RETIRED)

   def _validate_values(self, values):
      exclvals = set([x for x in values if x in self._exclusives])
      if len(exclvals) > 1:
         raise ValueError("Cannot update with multiple exclusive values: %s" %
                          ', '.join(exclvals))
      return exclvals

   def __init__(self, values=()):
      values = list(values)
      self._validate_values(values)
      set.__init__(self, values)

   def add(self, value):
      if value in self._exclusives:
         for x in self._exclusives:
            self.discard(x)
      set.add(self, value)

   def update(self, values):
      values = list(values)
      if self._validate_values(values):
         for x in self._exclusives:
            self.discard(x)
      set.update(self, values)

   def __str__(self):
      return ', '.join(self)

# NOTE: New-style class is needed for decorators/property to work, hence a
# subclass of built-in object type.
class VIB(object):
   """The VIB class represents a VIB object, either in metadata, or actual
      .vib files, or both.

      VIB objects are intended to be hashable, based on the 'id' attribute.
      This implies that once initialized, the 'id' attribute cannot change.

      Class Variables:
         * VIB_VERSION        - A constant string for the version of the
                                current VIB Specification
         * CROSS_ARCH         - A constant signifying the architecture value
                                used for cross-platform VIBs and VIBs on ESXi
         * TYPE_RPM           - A constant representing a VIB type that
                                requires RPM for installation.
         * TYPE_DEB           - A constant representing a VIB type that
                                requires ipkg for installation.
         * TYPE_CROSS         - A constant representing a VIB type with a
                                payload that can be installed with either RPM
                                or ipkg.
         * TYPE_FW            - A constant representing a VIB type containing a
                                firmware image.
         * DEST_ROOT          - A constant signifying the VIB payload should be
                                installed to the root (/) file system.
         * DEST_STAGE         - A constant signifying the VIB payload should be
                                persisted directly to a boot bank.
         * DEST_OEM           - A constant signifying the VIB payload should be
                                installed to the storage location for OEM
                                extensions.
         * DEST_CUSTOM        - A constant signifying the VIB payload should be
                                installed to the storage location for custom
                                (third-party) extensions.
         * DEST_LOCKER        - A constant signifying the VIB payload should be
                                installed to locker storage.
         * XML_DEST_META      - A constant to signify XML output destined for
                                vmware.xml (metadata.zip).
         * XML_DEST_VIB       - A constant to signify XML output destined for
                                descriptor.xml (inside a .vib file).
         * XML_DEST_DB        - A constant to signify XML output destined for
                                vibs.xml (in the esxudpate database).

      Attributes:
         * id                - A string giving the unique ID of the VIB. Can
                               only be set by the constructor.
         * type              - One of TYPE_RPM, TYPE_DEB, TYPE_CROSS or
                               TYPE_FW, specifying the payload type of the VIB.
                               Can only be set by the constructor.
         * name              - A user-friendly string giving the name of the
                               VIB. Not required to be unique.
         * version           - An instance of Version, specifying the version
                               information for the VIB.
         * arch              - A read-only string returning the architecture of
                               the VIB.
         * state             - An instance of VIBState, specifying the current
                               state(s) of the VIB, relative to the host.
         * description       - A string giving a user-friendly, detailed
                               description of the software in the VIB.
         * summary           - A string giving a user-friendly, one-line
                               summary of the VIB contents.
         * installdate       - A float value giving the date and time the VIB
                               was installed (as returned by time.time()). May
                               also be None if VIB was not installed, or the
                               value is otherwise not applicable.
         * releasedate       - A float value giving the date and time the VIB
                               was built or released (as returned by
                               time.time()). May also be None if unknown.
         * swplatforms       - An instance of SWPlatformCollection, indicating
                               the platforms for which the VIB is intended.
         * hwplatforms       - An instance of HWPlatformCollection, indicating
                               the platforms for which the VIB is intended.
         * maintenancemode   - A boolean indicating whether the host must be in
                               maintenance mode before installing the VIB.
         * packagefile       - A string indicating the name of the package in
                               the VIB payload.
         * bulletins         - An instance of BulletinCollection containing
                               bulletins of which this VIB is a member.
         * sourceurl         - An absolute URL to the location of the VIB.
         * relativepath      - A path relative to the location of metadata.zip
                               giving the location of the VIB.
         * remotelocation    - The URL or path of the .vib file. May be the
                               same as locallocation if a .vib is being
                               installed from a local path. May be an empty
                               string ('') if the remotelocation is unknown.
         * locallocation     - The path of the .vib on the local file system.
                               May be an empty string ('') if the VIB is not
                               available on the local file system.
         * packagelocation   - The location of the package payload for the VIB.
                               (Implies the VIB has already been extracted.)
         * scanned           - A boolean indicating whether this VIB has been
                               scanned against the host or not.
         * systemreqfailures - A set of strings indicating 0 or more reasons
                               why the host does not meet the VIB's system
                               requirements.
         * requires          - A list of Requires instances, specifying
                               dependency information for the VIB.
         * conflicts         - A set of VIB IDs for VIBs which conflict with
                               this VIB object.
         * oldervibs         - A set of VIB IDs for VIBS which are replaced
                               (obsoleted) by this VIB.
         * hostconflicts     - A set of (name, version) tuples indicating which
                               packages on the host conflict with this VIB.
         * hostobsoleted     - A set of (name, version) tuples indicating which
                               packages on the host are obsoleted by this VIB.
         * hostobsoleting    - A set of (name, version) tuples indicating which
                               packages on the host obsolete this VIB.
         * replaces          - A list of (name, conditional operator, version)
                               tuples representing which packages this VIB
                               replaces.
         * visordestination  - One of DEST_ROOT, DEST_STAGE, DEST_OEM,
                               DEST_CUSTOM, or DEST_LOCKER, specifying the
                               target destination on Visor hosts. (This
                               currently is not used for other hosts.)
         * rebootrequired    - A boolean specifying whether a reboot is or is
                               not required after installing the VIB.
         * hostdrestart      - A boolean specifying whether or not the host
                               agent must be restarted after installing the
                               VIB.
         * vendor            - String representing the vendor (publisher) of
                               the VIB.
         * persist           - A boolean specifying whether this VIB should be
                               persisted to the esxupdate database or not.
   """
   #TODO: Daming: Please bump this for cond maint mode. Any other changes we forgot?
   VIB_VERSION = '1.4.5'
   CROSS_ARCH = 'x86_64'

   TYPE_RPM   = 'rpm'
   TYPE_DEB   = 'deb'
   TYPE_CROSS = 'cross'
   TYPE_FW    = 'fw'
   types = (TYPE_RPM, TYPE_DEB, TYPE_CROSS) # TYPE_FW

   DEST_ROOT   = 'root'
   DEST_STAGE  = 'stage'
   DEST_OEM    = 'oem'
   DEST_CUSTOM = 'custom'
   DEST_LOCKER = 'locker'
   destinations = (DEST_ROOT, DEST_STAGE, DEST_OEM, DEST_CUSTOM, DEST_LOCKER)

   XML_DEST_META  = 'meta'
   XML_DEST_VIB   = 'vib'
   XML_DEST_DB    = 'db'

   def __init__(self, id, type, **kwargs):
      """Class constructor.
            Parameters:
               * id     - A string giving the VIB's unique ID.
               * type   - One of the TYPE_ constants specifying the VIB type.
               * kwargs - A list of keyword arguments used to populate the
                          VIB's attributes.
            Returns: A new VIB object.
            Raises:
               * TypeError - On any unrecognized keyword argument.
      """
      assert id is not None, "id parameter cannot be None"

      self._id = id
      self._type = type
      self.name              = kwargs.pop('name', '')
      self.version           = kwargs.pop('version', version.Version())
      self.state             = kwargs.pop('state', VIBState())
      self.description       = kwargs.pop('description', '')
      self.summary           = kwargs.pop('summary', '')
      self.installdate       = kwargs.pop('installdate', None)
      self.releasedate       = kwargs.pop('releasedate', None)
      self.swplatforms       = kwargs.pop('swplatforms', SWPlatformCollection())
      self.hwplatforms       = kwargs.pop('hwplatforms', HWPlatformCollection())
      self.maintenancemode   = kwargs.pop('maintenancemode', True)
      self.packagefile       = kwargs.pop('packagefile', '')
      self.bulletins    = kwargs.pop('bulletins', Bulletin.BulletinCollection())
      self.sourceurl         = kwargs.pop('sourceurl', '')
      self.relativepath      = kwargs.pop('relativepath', '')
      self.remotelocation    = kwargs.pop('remotelocation', self.sourceurl)
      self.locallocation     = kwargs.pop('locallocation', '')
      self.packagelocation   = kwargs.pop('packagelocation', '')
      self._scanned          = kwargs.pop('scanned', False)
      self.systemreqfailures = kwargs.pop('systemreqfailures', set())
      self.requires          = kwargs.pop('requires', list())
      self.conflicts         = kwargs.pop('conflicts', set())
      self.oldervibs         = kwargs.pop('oldervibs', set())
      self.hostconflicts     = kwargs.pop('hostconflicts', set())
      self.hostobsoleted     = kwargs.pop('hostobsoleted', set())
      self.hostobsoleting    = kwargs.pop('hostobsoleting', set())
      self.replaces          = kwargs.pop('replaces', list())
      self.visordestination  = kwargs.pop('visordestination', self.DEST_STAGE)
      self.rebootrequired    = kwargs.pop('rebootrequired', True)
      self.hostdrestart      = kwargs.pop('hostdrestart', True)
      self.vendor            = kwargs.pop('vendor', '')
      self.persist           = kwargs.pop('persist', True)

      if kwargs:
         badkws = ', '.join("'%s'" % kw for kw in kwargs)
         raise TypeError("Unrecognized keyword argument(s): %s." % badkws)

      # Should these be pseudo-private?
      self._vibfile = None
      self._meta = None
      self._cached = False

   __hash__ = lambda self: hash(self.id)
   __str__  = lambda self: self.id
   __repr__ = lambda self: '%s:%s' % (self.id, ', '.join(self.state))
   __cmp__  = lambda self, other: cmp(str(self), str(other))

   type = property(lambda self: self._type)
   id   = property(lambda self: self._id)

   def setscanned(self, value):
      if not value:
         self.systemreqfailures.clear()
         del self.requires[:]
         self.conflicts.clear()
         self.oldervibs.clear()
         self.hostconflicts.clear()
         self.hostobsoleted.clear()
         self.hostobsoleting.clear()
         del self.replaces[:]
      self._scanned = value

   scanned = property(lambda self: self._scanned, setscanned)

   @property
   def tnv(self):
      """Returns a tuple containing the type, name and version of the VIB
         payload. Will never raise an exception, but may return empty strings
         ('') for one or more of the values. For legacy reasons, this will
         parse the VIB ID to determine the type, name and version if any of
         those attributes are not set in the VIB object. However, this logic is
         deprecated and may be removed in the future.
      """
      if self.type and self.name and self.version.versionstring:
         return self.type, self.name, self.version.versionstring
      # If the name or version fields has a '_' in it, this will return the
      # wrong info - only do this when id is there but not name and version
      # (old VibDB format).
      try:
         vibtype, vibname, vibversion = self._id.split('_', 2)
         return vibtype, vibname, vibversion
      except:
         return self.type, self.name, self.version.versionstring

   @property
   def arch(self):
      """Returns the architecture of the VIB.
           RPM-based:  This returns the arch field from the version string.
           Cross and DEB-based:  This returns the arch constant used for these
                                 types of VIBs
           FW:  Returns ''.
         Returns '' if the version is None or arch attr is None.
      """
      if self.type == self.TYPE_FW:
         return ''
      if self.version:
         if hasattr(self.version, 'arch'):
            if self.version.arch:
               return self.version.arch
            else:
               return ''
         else:
            #
            # If the CROSS_ARCH ever changes, this is the place to return the proper
            # string for older VIBs, probably differentiated by vibversion.
            return self.CROSS_ARCH
      else:
         return ''

   @property
   def requiresmaintenancemode(self):
      if self.meetssystemreqs and self._meta:
         if self.obsoleteshost or self._ciscoupgrade():
            return self._meta.systemReqs.CalculateMMode('update')
         elif not self.obsoletedbyhost and \
              not self.state.INSTALLED in self.state:
            return self._meta.systemReqs.CalculateMMode('install')
      return self.maintenancemode

   @property
   def meetssystemreqs(self):
      # Note that we do not use self.swplatforms.matcheshost and
      # self.hwplatforms.matcheshost directly. Doing so does not give us the
      # opportunity to skip the evaluation when running in offline mode.
      # Instead, scanning will set systemreqfailures based on hwplatforms
      # and swplatforms, depending on the value of config.IsESXHost(). See
      # PR 482703 and PR 484128, as well as VIBScanner.scan().
      return not self.systemreqfailures

   @property
   def pkgdepsmetbyhost(self):
      """Returns True if all of the VIB's package dependencies are satisfied by
         the host, and none of the host's packages conflict with the VIB's
         payload. Returns False otherwise.
      """
      for req in self.requires:
         if not req.providedbyhost:
            return False
      return not self.conflictswithhost

   conflictswithhost = property(lambda self: bool(self.hostconflicts))
   obsoleteshost     = property(lambda self: bool(self.hostobsoleted))
   obsoletedbyhost   = property(lambda self: bool(self.hostobsoleting))

   @classmethod
   def idfromtnv(cls, type, name, ver):
      """Generate a VIB ID from a package name and version.
            Parameters:
               * type - One of TYPE_RPM, TYPE_DEB or TYPE_CROSS.
               * name - A string giving the package name.
               * ver  - An instance of Version or a string giving the package
                        version.
            Returns: A string giving the generated VIB ID.
      """
      verstr = str(ver)
      if type == cls.TYPE_CROSS and verstr:
         verstr = verstr.split('@')[0]
      return '_'.join((type, name, verstr))

   def __add__(self, other):
      """Merges attributes between two VIBS.
            Parameters:
               * other - another VIB object.
            Returns: A new VIB object. Note that the new VIB will have to be
                     re-scanned, since we have no way of ensuring that
                     system requirements meta-data has not changed.
      """
      assert self.id == other.id, "Cannot merge VIBs with different IDs."
      assert self.type == other.type, "Cannot merge VIBs with different types."

      v = VIB(self.id, self.type)

      if other.scanned and not self.scanned:
         v.state.update(self.state)
         v.state.update(other.state)
      elif self.scanned and not other.scanned:
         v.state.update(other.state)
         v.state.update(self.state)
      # Numbers can be compared to None. NB: 0 > None.
      elif other.installdate > self.installdate:
         v.state.update(self.state)
         v.state.update(other.state)
      else:
         v.state.update(other.state)
         v.state.update(self.state)
      v.installdate = max(self.installdate, other.installdate)

      metaattrs = ('name', 'version', 'description', 'summary', 'swplatforms',
                   'hwplatforms', 'maintenancemode', 'packagefile',
                   'sourceurl', 'relativepath', 'visordestination',
                   'rebootrequired', 'hostdrestart', 'vendor', '_meta')
      for attr in metaattrs:
         mine = getattr(self, attr)
         theirs = getattr(other, attr)
         if self.releasedate > other.releasedate:
            setattr(v, attr, mine)
         elif self.releasedate < other.releasedate:
            setattr(v, attr, theirs)
         else:
            setattr(v, attr, mine or theirs)
      v.releasedate = max(self.releasedate, other.releasedate)

      stageattrs = ('locallocation', 'packagelocation', 'remotelocation',
                    '_vibfile')
      staged = VIBState.STAGED
      for attr in stageattrs:
         mine = getattr(self, attr)
         theirs = getattr(other, attr)
         if staged in self.state and not staged in other.state:
            setattr(v, attr, mine)
         elif staged in other.state and not staged in self.state:
            setattr(v, attr, theirs)
         else:
            setattr(v, attr, mine or theirs)
      if staged in self.state | other.state:
         v.state.add(staged)

      v.persist = self.persist or other.persist

      v.bulletins = self.bulletins + other.bulletins

      return v

   def setpkgstate(self, value):
      """Sets the appropriate pkgstate, updating timestamp if the state has
         changed.
      """
      if value not in self.state:
         self.state.add(value)
         self.updatetimestamp()

   def updatetimestamp(self):
      self.installdate = time.time()

   def toxml(self, target=XML_DEST_DB):
      """Expresses VIB attributes as an ElementTree element.
            Parameters:
               * target - One of XML_DEST_VIB, XML_DEST_META, or XML_DEST_DB,
                          specifying the intended target of the output. The
                          target controls which attributes are expressed in the
                          ElementTree element.
            Returns: An object conforming to the ElementTree element API.
      """
      if target == self.XML_DEST_DB:
         descriptorclass = Descriptor.DescriptorBase
      elif target == self.XML_DEST_VIB:
         descriptorclass = Descriptor.DescriptorFile
      elif target == self.XML_DEST_META:
         descriptorclass = Descriptor.VibMeta
      else:
         raise RuntimeError("Unknown XML destination '%s'" % target)

      if self._meta is None:
         meta = descriptorclass()
         self._meta = meta
      elif self._meta.__class__ != descriptorclass:
         try:
            # Should we even try this? This is primarily useful for preserving
            # things that are common to VibMeta and DescriptorBase, but which
            # our own VIB class doesn't know how to deal with. -RS
            meta = descriptorclass()
            meta.FromXml(self._meta.ToXml())
         except:
            meta = descriptorclass()
      else:
         meta = self._meta

      # All Descriptor objects should have these things.
      meta.vibID = self._id
      meta.vibType = self._type
      meta.name = self.name
      meta.vendor = self.vendor
      meta.version = self.version.versionstring
      if self.releasedate:
         dt = datetime.datetime.utcfromtimestamp(self.releasedate)
         meta.buildDate = dt.isoformat()

      for swplatform in self.swplatforms:
         item = (swplatform.productid, swplatform.version.versionstring,
                 swplatform.locale)
         if not item in meta.systemReqs.swplatforms:
            meta.systemReqs.swplatforms.append(item)
      for hwplatform in self.hwplatforms:
         item = (hwplatform.vendor, hwplatform.model)
         if not item in meta.systemReqs.hwplatforms:
            meta.systemReqs.hwplatforms.append(item)

      meta.systemReqs.maintenanceMode = self.maintenancemode
      meta.rebootRequired = self.rebootrequired
      meta.hostdRestart = self.hostdrestart
      meta.visorDestination = self.visordestination
      meta.pkgfile = self.packagefile

      # Target-specific things:
      if target == self.XML_DEST_DB:
         # For vibs.xml, just write the ts as a float value.
         node = etree.Element('vib', id=self._id, state=','.join(self.state))
         if self.installdate:
            node.set('ts', '%.6f' % self.installdate)
         if self.state.STAGED in self.state:
            if self.remotelocation:
               etree.SubElement(node, 'remoteLocation').text = \
                  self.remotelocation
            if self.locallocation:
               etree.SubElement(node, 'localLocation').text = self.locallocation
            if self.packagelocation:
               etree.SubElement(node, 'packageLocation').text = \
                  self.packagelocation
         descriptor = meta.ToXml()
         node.append(descriptor)
         for package, op, ver in self.replaces:
            replaces = etree.SubElement(node, 'replaces', package=package)
            if op and ver:
               replaces.set('op', op)
               replaces.set('ver', ver)
         return node
      elif target == self.XML_DEST_VIB:
         return meta.ToXml()
      elif target == self.XML_DEST_META:
         meta.sourceUrl = self.sourceurl
         meta.relativePath = self.relativepath
         return meta.ToXml()
      else:
         # Redundant
         raise RuntimeError("Unknown XML source '%s'" % target)

   @classmethod
   def fromxml(cls, xml, target=XML_DEST_DB, **kwargs):
      """Creates a VIB instance from XML.
            Parameters:
               * xml    - Must be either an instance of ElementTree, or a
                          string of XML-formatted data.
               * kwargs - Initialize constructor arguments from keywords.
                          Primarily useful to provide required id or type
                          arguments when XML data is from a template.
            Returns: A new VIB object.
            Raises:
               * VibFormatError - A string was passed, but could not be parsed
                                  as valid XML data; or the parsed XML did not
                                  include required tags or attributes, or was
                                  otherwise not in the expected format.
      """
      if etree.iselement(xml):
         node = xml
      else:
         try:
            node = etree.fromstring(xml)
         except Exception, e:
            msg = "Error parsing XML data: %s." % e
            raise errors.VibFormatError(None, msg)

      # Why do we have 80% of the XML parsing in Descriptor.py, and this
      # little bit here? -RS
      if target == cls.XML_DEST_DB:
         # VIB ID
         kwargs['id'] = node.get('id')

         # VIB state (Always assume uninstalled?)
         kwargs['state'] = VIBState((VIBState.UNINSTALLED,))
         for state in node.get('state', '').split(','):
            state = state.strip()
            if state in VIBState.states:
               kwargs['state'].add(state)

         # VIB timestamp
         timestamp = node.get('ts', '')
         try:
            kwargs['installdate'] = float(timestamp)
         except:
            try:
               kwargs['installdate'] = util.isotofloat(timestamp)
            except:
               pass

         # Other DB-specific stuff not handled by Descriptor module...
         kwargs['remotelocation'] = node.findtext('remoteLocation', '').strip()
         kwargs['locallocation'] = node.findtext('localLocation', '').strip()
         kwargs['packagelocation'] = \
            node.findtext('packageLocation', '').strip()

         # Stuff handled by Descriptor.py
         dnode = node.find('descriptor')
         if dnode:
            try:
               meta = Descriptor.DescriptorBase(dnode)
            except Exception, e:
               msg = "Unable to parse VIB data from XML: %s." % e
               raise errors.VibFormatError(None, msg)
         else:
            # Old versions (vMA 4.0 GA) do not save descriptor in the database.
            meta = Descriptor.DescriptorBase()
            meta.vibID = kwargs.get('id', '')
            meta.vibType = meta.vibID.split('_', 1)[0]
            # An RPM name, version, release and even arch can have an '_' in
            # it, which means it is impossible to reliably parse a VIB ID into
            # a package name and version.
            meta.name = '(unknown)'
            meta.version = '0-0'

         # Replaces (also specific to the DB)
         kwargs['replaces'] = []
         for rnode in node.findall('replaces'):
            kwargs['replaces'].append((rnode.get('package'),
                                       rnode.get('op'), rnode.get('ver')))
      elif target == cls.XML_DEST_VIB:
         try:
            meta = Descriptor.DescriptorFile()
            meta.FromXml(node)
         except Exception, e:
            msg = "Unable to parse VIB data from XML: %s." % e
            raise errors.VibFormatError(None, msg)
      elif target == cls.XML_DEST_META:
         try:
            meta = Descriptor.VibMeta(node)
         except Exception, e:
            msg = "Unable to parse VIB data from XML: %s." % e
            raise errors.VibFormatError(None, msg)
         kwargs['sourceurl'] = meta.sourceUrl
         kwargs['relativepath'] = meta.relativePath
      else:
         raise RuntimeError("Unknown XML source '%s'" % target)

      kwargs.setdefault('id', meta.vibID)
      if 'id' not in kwargs or not kwargs['id'] or '/' in kwargs['id']:
         msg = "Could not parse valid VIB ID from XML data."
         raise errors.VibFormatError(None, msg)

      # All Descriptor objects should have these things.
      kwargs['type'] = meta.vibType
      if meta.vibType not in cls.types:
         msg = "Could not parse valid VIB type from XML data."
         raise errors.VibFormatError(None, msg)

      if meta.vibType == cls.TYPE_RPM:
         kwargs['version'] = version.RPMVersion.fromstring(meta.version)
      elif meta.vibType == cls.TYPE_DEB:
         kwargs['version'] = version.DebVersion.fromstring(meta.version)
      elif meta.vibType == cls.TYPE_CROSS:
         kwargs['version'] = version.Version.fromstring(meta.version)
      #elif meta.vibType == cls.TYPE_FW:
      #   kwargs['version'] = FWVersion.fromstring(meta.version)

      kwargs['name'] = meta.name
      kwargs['vendor'] = meta.vendor

      try:
         kwargs['releasedate'] = util.isotofloat(meta.buildDate)
      except:
         pass

      kwargs['swplatforms'] = SWPlatformCollection()
      for productid, ver, locale in meta.systemReqs.swplatforms:
         kwargs['swplatforms'].add(SWPlatform(productid, ver, locale))

      kwargs['hwplatforms'] = HWPlatformCollection()
      for vendor, model in meta.systemReqs.hwplatforms:
         kwargs['hwplatforms'].add(HWPlatform(vendor, model))

      kwargs['maintenancemode'] = meta.systemReqs.maintenanceMode
      kwargs['rebootrequired'] = meta.rebootRequired
      kwargs['hostdrestart'] = meta.hostdRestart
      if meta.visorDestination:
         if meta.visorDestination not in cls.destinations:
            msg = "%s is an invalid visor destination." % meta.visorDestination
            raise errors.VibFormatError(None, msg)
         kwargs['visordestination'] = meta.visorDestination
      kwargs['packagefile'] = meta.pkgfile

      vib = cls(**kwargs)
      vib._meta = meta
      return vib

   @classmethod
   def fromfile(cls, name=None, fileobj=None):
      """Initializes a VIB object from a .vib file.
            Parameters:
               * name    - The path to a .vib file. May be None if fileobj is
                           specified.
               * fileobj - A file or file-like object from which to read VIB
                           data.
            Returns: A new VIB object.
            Raises:
               * VibIOError     - If an error is encountered attempting to
                                  read or write from the file system.
               * VibFormatError - If an error is encountered parsing the ar
                                  format or the descriptor XML data.
      """
      if name:
         locallocation = name
      elif hasattr(fileobj, 'name') and fileobj.name:
         locallocation = fileobj.name

      try:
         vibfile = VibFile(name=name, fileobj=fileobj)
         descriptor = etree.parse(vibfile.extractfile('descriptor.xml'))
         vibfile.close()
      except EnvironmentError, e:
         msg = "fromfile: Error reading VIB file: %s." % e
         raise errors.VibIOError(locallocation, msg)
      except KeyError, e:
         msg = "fromfile: descriptor.xml not found in VIB archive."
         raise errors.VibFormatError(locallocation, msg)
      except Exception, e:
         msg = "fromfile: Error reading VIB file format: %s." % e
         raise errors.VibFormatError(locallocation, msg)

      vib = cls.fromxml(descriptor.getroot(), cls.XML_DEST_VIB)

      if locallocation:
         vib.locallocation = locallocation

      return vib

   @classmethod
   def fromurl(cls, url, destdir, progress=None):
      """Download the VIB to the local file system and create new VIB object.
            Parameters:
               * url      - The remote location of the file.
               * destdir  - The local directory to which to download the .vib
                            file.
               * progress - An object implementing the progress API used by
                            the downloader.Downloader class. See that class'
                            documentation.
            Returns: A new VIB object.
            Raises:
               * VibIOError       - If an error is encountered attempting to
                                    read or write from the file system.
               * VibDownloadError - If an error is encountered attempting to
                                    retrieve the remote file.
               * VibFormatError   - If an error is encountered parsing the ar
                                    format or the descriptor XML data.
      """
      if not os.path.isdir(destdir):
         try:
            os.makedirs(destdir)
         except Exception, e:
            msg = "fromurl: Cannot create destination directory: %s" % e
            raise errors.VibIOError(url, msg)

      tmploc = os.path.join(destdir, str(hash(url)))
      d = downloader.Downloader(url, tmploc, progress=progress)

      try:
         locallocation = d.Get()
      except downloader.DownloaderError, e:
         raise errors.VibDownloadError(url, tmploc, str(e))

      try:
         vibfile = VibFile(name=locallocation)
         tree = etree.parse(vibfile.extractfile('descriptor.xml'))
         descriptor = tree.getroot()
         vibfile.close()
      except EnvironmentError, e:
         msg = "fromurl: Error reading VIB file: %s." % e
         raise errors.VibIOError(url, msg)
      except KeyError, e:
         msg = "fromurl: descriptor.xml not found in VIB archive."
         raise errors.VibFormatError(url, msg)
      except Exception, e:
         msg = "fromurl: Error reading VIB file format: %s." % e
         raise errors.VibFormatError(url, msg)

      vib = cls.fromxml(descriptor, target=cls.XML_DEST_VIB)

      if locallocation == tmploc:
         locallocation = os.path.join(destdir, vib.id + '.vib')
         try:
            os.rename(tmploc, locallocation)
            vib._cached = True
         except Exception, e:
            msg = "fromurl: Error renaming temporary file: %s" % e
            raise errors.VibIOError(url, msg)

      vib.locallocation = locallocation
      vib.remotelocation = url

      return vib

   def _ciscoupgrade(self):
      """ This method determines if this vib is an "upgrade" to
          an already installed cisco vib.

          Checks that the current vib name is Cisco's.
          Checks the cisco_packages dictionary for any installed
          cisco vibs, if any are found, then True is returned.
      """
      if self.name.startswith("cisco-vem"):
         if self.state.INSTALLED in self.state:
             return False

         if len(cisco_packages) > 0:
             return True

      return False

   def converttodeb(self, destdir, symlink=False):
      """Create packagelocation file with 'deb' extension, either by rename
         or create a symbolic link to locallocation.
         Parameters:
            * destdir - The directory in which to convert the deb
            * symlink - True:  create symlink from vibID.deb to vib file
                        False: rename vib file as vibID.deb
      """
      if self.type in [self.TYPE_DEB, self.TYPE_CROSS] \
            and self.locallocation is not None:

         filename = self.id + ".deb"
         dst = os.path.abspath(os.path.join(destdir, filename))
         if symlink:
            os.symlink(os.path.abspath(self.locallocation), dst)
         else:
            os.rename(self.locallocation, dst)
            self.locallocation = dst
            self._closevibfile()
            self._openvibfile()
         self.packagelocation = dst

   def createfullrpm(self, destdir):
      """Converts the data.tar.gz in a cross-platform VIB to a cpio.gz and
         appends it to the rpm at packagelocation.
            Parameters:
               * destdir - Directory in which to create the RPM.
            Raises:
               * RuntimeError   - The method is called and the locallocation
                                  attribute is not set or does not refer to a
                                  valid path on the file system.
               * VibIOError     - If an error is encountered attempting to
                                  read or write from the file system.
               * VibFormatError - If an error is encountered parsing the ar
                                  format or the descriptor XML data.
      """
      if not self.type == self.TYPE_CROSS:
         # Raise an exception here.
         msg = "Cannot call createfullrpm on VIB with type = %s" % self.type
         raise RuntimeError(msg)

      if not HAVE_RPM:
         raise RuntimeError("Required rpm module missing.")

      try:
         tmpdir = tempfile.mkdtemp()
      except Exception, e:
         msg = "createfullrpm: Cannot create temporary directory: %s" % e
         raise errors.VibIOError(self.locallocation, msg)

      # This always sets self.packagelocation. Also note that the call to this
      # method is not conditional, since we cannot guarantee that any previous
      # call used the same 'destdir' parameter. -RS
      self.extractrpm(destdir)

      # Obtain list of files from RPM header. Don't rely on tarfile list as
      # some dirs may be pruned - PR 319135.
      try:
         f = open(self.packagelocation)
         ts = rpm.ts()
         hdr = ts.hdrFromFdno(f.fileno())
         f.close()
      except Exception, e:
         msg = "Error reading RPM header: %s" % e
         raise errors.VibIOError(self.locallocation, msg)

      # Format of cpio filelist: ./<path> for each file to archive
      sources = [os.path.join(tmpdir, fn.lstrip(os.sep))
                 for fn in hdr[rpm.RPMTAG_FILENAMES]]

      self.extractdatagz(tmpdir)

      try:
         # create cpio archive and reconstruct entire .rpm
         gz = gzip.open(self.packagelocation, 'ab')
         # This will automatically close gz:
         util.cpio.mkrpmcpio(sources, gz, recurse=False, root=tmpdir)
      except Exception, e:
         shutil.rmtree(tmpdir)
         self.packagelocation = ''
         msg = "Error appending payload to .rpm file: %s" % e
         raise errors.VibIOError(self.locallocation, msg)

      shutil.rmtree(tmpdir)

   def _removeintermediatefiles(self):
      """Removes intermediate files created by the stage process. Should only
         be called after the VIB has been successfully extracted.
      """
      locallocation = os.path.realpath(self.locallocation)
      packagelocation = os.path.realpath(self.packagelocation)

      # On Visor, packagelocation == locallocation. In that case, don't remove
      # the package. In the case where the VIB was not downloaded from a remote
      # location, then self._cached is False, and we don't want to remote that,
      # either.
      if self._cached and locallocation != packagelocation:
         try:
            os.unlink(locallocation)
         except:
            log.warn("Unable to remove cached VIB at %s." % locallocation)

   def stage(self, destdir, progress=None, skipsigcheck=False, nocache=False):
      """Download and unpack a VIB.
            Parameters:
               * destdir      - The local directory to which to download the
                                .vib file.
               * progress     - An object implementing the progress API used by
                                the downloader.Downloader class. See that
                                class' documentation.
               * skipsigcheck - A boolean specifying whether to skip signature
                                validation. Defaults to False, meaning an
                                exception will be raised if VIB does not have
                                a valid and properly signed checksum.
               * nocache      - Re-stage if already in cache.
            Raises:
               * RuntimeError       - The method is called and the
                                      locallocation attribute is not set or
                                      does not refer to a valid path on the
                                      file system.
               * VibIOError         - If an error is encountered attempting to
                                      read or write from the file system.
               * VibDownloadError   - If an error is encountered attempting to
                                      retrieve the remote file.
               * VibFormatError     - If an error is encountered parsing the ar
                                      format or the descriptor XML data.
               * VibSigMissingError - If skipsigcheck is False, and the VIB
                                      archive does not contain signature
                                      information.
               * VibSigFormatError  - If skipsigcheck is False, and the VIB
                                      archive's signature data is not in the
                                      correct format.
               * VibSigInvalidError - If skipsigcheck is False, and the VIB
                                      archive's signature data is not signed by
                                      a trusted publisher.
               * VibSigDigestError  - If skipsigcheck is False, and the
                                      checksum data in the VIB archive's
                                      signature data does not match the
                                      checksum computed from the VIB archive
                                      data.
      """
      if not os.path.isdir(destdir):
         try:
            os.makedirs(destdir)
         except Exception, e:
            msg = "stage: Cannot create destination directory: %s" % e
            raise errors.VibIOError(self.remotelocation, msg)

      # If self.locallocation is set and is a valid file, we assume that it
      # has previously been downloaded successfully, unless nocache = True. -RS
      if os.path.exists(self.locallocation or '') and not nocache:
         newvib = self.fromfile(self.locallocation)
      elif self.remotelocation:
         newvib = self.fromurl(self.remotelocation, destdir, progress)
      else:
         msg = ("stage: Cannot download VIB '%s': Remote location unknown."
                % self._id)
         raise errors.VibDownloadError(None, None, msg)

      checkattrs = ('id', 'type', 'visordestination')
      for attr in checkattrs:
         if getattr(newvib, attr) != getattr(self, attr):
            msg = ("stage: VIB %s from descriptor does not match VIB object %s."
                   % (attr, attr))
            raise errors.VibFormatError(newvib.locallocation, msg)

      self.locallocation = newvib.locallocation
      self._cached = newvib._cached
      del newvib

      if not skipsigcheck:
         self.verifysignature()

      if progress:
         progress.start()
         progress.setTopic("Unpacking %s" % self._id)
         progress.set(0, 1)
         progress.show()

      # I don't really think that figuring out the platform based on a config
      # file is the best way to go. API users should have a way to configure
      # module parameters like this, otherwise unit testing or other use of the
      # library outside of the ESX environment becomes difficult. -RS
      if config.GetPlatform() == 'cos':
         stagedir = os.path.join(destdir, self._id)
         if not os.path.isdir(stagedir):
            try:
               os.makedirs(stagedir)
            except Exception, e:
               msg = "stage: Cannot create stage directory: %s" % e
               raise errors.VibIOError(self.locallocation, msg)

         if self.type == self.TYPE_RPM:
            self.extractrpm(stagedir)
         elif self.type == self.TYPE_CROSS:
            self.createfullrpm(stagedir)
         else:
            msg = "VIB type '%s' is not appropriate for this host." % self.type
            raise errors.VibFormatError(self.locallocation, msg)
      elif config.GetPlatform() == 'visor':
         if self.type in (self.TYPE_CROSS, self.TYPE_DEB):
            self.packagelocation = self.locallocation
         else:
            msg = "VIB type '%s' is not appropriate for this host." % self.type
            raise errors.VibFormatError(self.locallocation, msg)

      if progress:
         progress.set(1, 1)
         progress.show()
         progress.stop()

      self.state.add(VIBState.STAGED)
      self.updatetimestamp()
      self._removeintermediatefiles()
      self._closevibfile()

   def _openvibfile(self):
      # This method just ensures that self._vibfile is bound to a valid
      # VibFile object.
      if self._vibfile is not None:
         return

      if not self.locallocation or not os.path.isfile(self.locallocation):
         msg = "Unable to open .vib archive; no valid location is set."
         raise RuntimeError(msg)

      try:
         self._vibfile = VibFile(self.locallocation)
      except EnvironmentError, e:
         msg = "Error reading .vib file: %s" % e
         raise errors.VibIOError(self.locallocation, msg)
      except Exception, e:
         msg = "Error reading data from .vib file: %s" % e
         raise errors.VibFormatError(self.locallocation, msg)

   def _closevibfile(self):
      """Close self._vibfile if it is not None"""
      if self._vibfile is not None:
         self._vibfile.close()
         self._vibfile = None

   def computedigest(self):
      """Returns the sha256 digest of the unsigned portion of the VIB.
            Returns: A string representing the digest (in hex notation).
            Raises:
               * RuntimeError   - The method is called and the locallocation
                                  attribute is not set or does not refer to a
                                  valid path on the file system.
               * VibIOError     - If an error is encountered attempting to read
                                  or write from the file system.
               * VibFormatError - If an error is encountered parsing the ar
                                  format.
      """
      self._openvibfile()
      return self._vibfile.ComputeDigest()

   def extractrpm(self, destdir):
      """Extracts the .rpm file in the vib archive and sets the packagelocation
         attribute.
            Parameters:
               * destdir - The directory in which to extract the RPM file.
            Raises:
               * RuntimeError   - The method is called and the locallocation
                                  attribute is not set or does not refer to a
                                  valid path on the file system.
               * VibIOError     - If an error is encountered attempting to read
                                  or write from the file system.
               * VibFormatError - If an error is encountered parsing the ar
                                  format.
      """
      self._openvibfile()
      dst = self._vibfile.ExtractRpm(destdir)
      self.packagelocation = dst

   def extractdatagz(self, destdir):
      """Extracts all files and directories from data.tar.gz to destdir.
            Parameters:
               * destdir - Specifies the directory to extract data.tar.gz to.
            Raises:
               * RuntimeError   - The method is called and the locallocation
                                  attribute is not set or does not refer to a
                                  valid path on the file system.
               * VibIOError     - If an error is encountered attempting to read
                                  or write from the file system.
               * VibFormatError - If an error is encountered parsing the ar
                                  format.
      """
      self._openvibfile()
      self._vibfile.ExtractDataGz(destdir)

   def verifysignature(self):
      """Verifies the VIB's signature data. This method extracts the sig.pkcs7
         file from the VIB archive data, verifies that the PKCS7 data has a
         valid signature, and then verifies that the checksum contained in the
         PKCS7 data matches the checksum computed for the VIB data.
            Raises:
               * RuntimeError       - The method is called and the
                                      locallocation attribute is not set or
                                      does not refer to a valid path on the
                                      file system.
               * VibIOError         - If an error is encountered attempting to
                                      read or write from the file system.
               * VibFormatError     - If an error is encountered parsing the ar
                                      format.
               * VibSigMissingError - If the VIB archive does not contain
                                      signature information.
               * VibSigFormatError  - If the VIB archive's signature data is
                                      not in the correct format.
               * VibSigInvalidError - If the VIB archive's signature data is
                                      not signed by a trusted publisher.
               * VibSigDigestError  - If the checksum data in the VIB archive's
                                      signature data does not match the
                                      checksum computed from the VIB archive
                                      data.
      """
      self._openvibfile()
      self._vibfile.VerifySignature()

class VibFile(util.arfile.ArFile):
   """ Contains methods specific to VIB files, such as
       digital signatures and extracting RPMs.  Has these
       attributes:
         desc   - Descriptor instance from VIB
         rpm    - name of .rpm entry in archive, or None
         signed - True if Vib is signed
         
       Only reading and adding signatures is supported
       right now.
   """
   ELFSIGN_MAJOR_VERSION = 0    #sha256 digest.

   def __init__(self, name=None, mode='rb', fileobj=None):
      """ Construct a VibFile instance.
          name or fileobj is expected to refer to a
          completely built .vib.
          mode:  rb - read-only;  a+b - can add signatures.
                 Only the above two modes are supported.
      """
      if mode not in ('rb', 'a+b', 'wb'):
         raise ValueError, "Incompatible VibFile mode"
      try:
         util.arfile.ArFile.__init__(self, name, mode, fileobj)
      except util.arfile.ArError, e:
         raise errors.VibFormatError(e.filename,
               "Vib is not a valid ar file :%s" % (str(e)))
      self.rpm = None
      self.signed = False
      self.desc = None
      self._unsigned_length = 0
      if len(self.filelist) > 0:
         self._parseVib()

   def _parseVib(self):
      '''Verifies Vib format, sets vars and reads in descriptor'''
      filenames = [info.filename for info in self.filelist]
      if len(set(filenames)) < len(filenames):
         raise errors.VibFormatError(self._filename,
                                     "Duplicate file entries in Vib")
      if 'descriptor.xml' not in filenames:
         raise errors.VibFormatError(self._filename,
                                     "Vib missing descriptor.xml")
      try:
         self.desc = Descriptor.DescriptorBase(etree.fromstring(
            self.read('descriptor.xml')))
      except (xmlerrclass, Descriptor.BadDescriptorError), e:
         raise errors.VibFormatError(self._filename,
                                     "Cannot parse Vib descriptor.xml")

      for f in filenames:
         if f.endswith('.rpm'):
            if not self.rpm:
               self.rpm = f
            else:
               raise errors.VibFormatError(self._filename,
                                           "More than one .rpm found")

      if self.desc.vibType in (VIB.TYPE_CROSS, VIB.TYPE_RPM):
         if not self.rpm:
            raise errors.VibFormatError(self._filename,
                "Vibs of type '%s' must have one rpm" % (self.desc.vibType))
      if self.desc.vibType in (VIB.TYPE_CROSS, VIB.TYPE_DEB):
         #
         # Check for .deb compatibility.  Since arfile's getFilelist()
         # skips over long filename table, we must check _longfns
         # to detect for its absence.
         #
         if self._longfns:
            raise errors.VibFormatError(self._filename,
                                        "Long filenames not allowed for Vibs")
         if filenames[0:3] != ['debian-binary', 'control.tar.gz', 'data.tar.gz']:
            raise errors.VibFormatError(self._filename,
                                        "Vib is not Debian-compatible")

      self.signed = 'sig.pkcs7' in filenames
      if self.signed:
         if filenames[-1] != 'sig.pkcs7':
            raise errors.VibFormatError(self._filename,
                                        "Signature file not the last member")
         lastinfo = self.filelist[-2]
      else:
         lastinfo = self.filelist[-1]
      self._unsigned_length = lastinfo.offset + lastinfo.size

      
   def ComputeDigest(self):
      ''' Returns the sha256 digest of the unsigned portion of the vib
          as a hex string
      '''
      self._fileobj.seek(0, 0)
      h = hasher()
      remaining = self._unsigned_length
      while remaining > 0:
         chunk = self._fileobj.read(min(remaining, HASH_CHUNK_SIZE))
         h.update(chunk)
         remaining -= len(chunk)
      return h.hexdigest()

   def ExtractRpm(self, destdir):
      """ Extracts the .rpm file in the vib archive. """
      if self.rpm:
         if self.desc.pkgfile:
            dst = os.path.join(destdir, self.desc.pkgfile)
         else:
            dst = os.path.join(destdir, os.path.basename(self.rpm))
         f = open(dst, 'wb')
         src = self.extractfile(self.rpm)
         chunk = src.read(HASH_CHUNK_SIZE)
         while chunk:
            f.write(chunk)
            chunk = src.read(HASH_CHUNK_SIZE)
         f.close()
         return dst

   def ExtractDataGz(self, destdir):
      """ Extracts all files and directories from data.tar.gz
          to destdir
      """
      try:
         data = self.extractfile('data.tar.gz')
         #
         # Open tar.gz in block streaming mode and stream
         # straight from Vib file to minimize memory footprint
         # 
         tar = tarfile.open(mode="r|gz", fileobj=data)
         for tarinfo in tar:
            tar.extract(tarinfo, destdir)
         tar.close()
      except KeyError, e:
         raise errors.VibFormatError(self._filename,
               'Missing data.tar.gz in VIB :%s' % (str(e)))
      except tarfile.TarError, e:
         raise errors.VibIOError(self._filename,
               'Error extracting data.tar.gz from vib %s: %s' %
               (self._filename, str(e)) )

   def AddSignature(self, pkcs7file):
      """ Adds a PKCS7 signature file to an unsigned Vib.
          If the Vib is signed, then the signature is replaced.
          The pkcs7file should be the file output by elfsign -V.

          Note: No protection for the original .vib is handled
          here, as VibFile is designed to handle one physical
          file only.  For best practice, the caller should
          modify a copy of the original vib only.
      """
      if self.signed:
         #
         # In 'a' mode, it's necessary to truncate the file
         # if we don't want to write to the end.
         # Seek to beginning of sig.pkcs7 Ar header
         #
         import struct
         hdr_len = struct.calcsize(util.arfile.STRUCT_AR_HEADER)
         #self._fileobj.seek(self._unsigned_length, 0)
         self._fileobj.seek(self.filelist[-1].offset - hdr_len, 0)
         self._fileobj.truncate()
         self.signed = False
         self.filelist.pop()
      else:
         self._fileobj.seek(0, 2)

      fp = open(pkcs7file, 'r')
      self.writestr('sig.pkcs7', fp.read())
      fp.close()
      self.signed = True

   def VerifySignature(self):
      """ Verifies the Vib signature.
          It extracts out the sig.pkcs7 file, verifies that the PKCS7
          sig is valid (the key can be traced back to the VMware or partner
          root certificate, and the data contained is signed correctly),
          and then verifies that the checksum contained in the PKCS7
          matches the checksum of the Vib itself.
      Raises:
          VibSignatureError and its subclassed errors
          VibIOError - could not unpack sig.pkcs7
      """
      if not HAVE_PYPKCS:
         msg = ("verifysignature: Required PyPKCS module not present.")
         raise RuntimeError(msg)

      if not self.signed:
         raise errors.VibSigMissingError(self._filename)
      #
      # mkstemp() offers various security benefits
      #
      fd, pkcs7 = tempfile.mkstemp()
      try:
         #
         # First four bytes of sig.pkcs7 is version string
         #
         info = self.nameToInfo('sig.pkcs7')
         offset = info.offset
         siglen = info.size
         if siglen <= 4:
            raise errors.VibSigFormatError(self._filename,
                                           "Signature file too short")
         self._fileobj.seek(offset, 0)
         sigversion = self._fileobj.read(4)
         if ord(sigversion[0]) != self.ELFSIGN_MAJOR_VERSION:
            raise errors.VibSigVersionError(self._filename,
                                            "Incorrect signature version(%d)"
                                            % (ord(sigversion[0])) )
         try:
            os.write(fd, self._fileobj.read(siglen - 4))
            os.close(fd)
         except (EnvironmentError, KeyError), e:
            raise errors.VibIOError(self._filename,
                                    "Cannot extract signature file: %s" % (e))
         try:
            if not PyPKCS.IsPKCS7SignatureValid(pkcs7):
               raise errors.VibSigInvalidError(self._filename,
                                                "PKCS7 signature is invalid")

            sigchecksum = binascii.hexlify(PyPKCS.GetDataFromPKCS7File(pkcs7))
         except PyPKCS.Exception, e:
            raise errors.VibSigInvalidError(self._filename,
                  'Error checking PKCS7 signature: %s' % (str(e)))

         ourchecksum = self.ComputeDigest()
         if sigchecksum != ourchecksum:
            raise errors.VibSigDigestError(self._filename,
                                           ourchecksum, sigchecksum)
      finally:
         os.unlink(pkcs7)


class Requires(object):
   """A class representing a dependency.
         Attributes:
            * name           - A name describing the requirement.
            * providedby     - A set of VIB IDs, representing the VIBs which
                               can be use to satisfy the requirement.
            * providedbyhost - A boolean. True if the required dependency is
                               already installed on the host.
            * hostprovides   - A set of (name, version) tuples indicating which
                               packages on the host satisfy this dependency.
            * hostrequiredby - A set of (name, version) tuples indicating which
                               packages on the host require this dependency.
            * breaks         - A set of VIB IDs, representing the VIBs upgrading
                               or downgrading the packages in hostprovides,
                               and which do NOT satisfy this dependency
   """
   def __init__(self, name):
      self.name = name
      self.providedby = set()
      self.hostprovides = set()
      self.hostrequiredby = set()
      self.breaks = set()

   @property
   def providedbyhost(self):
      return len(self.hostprovides) > 0
