#######################################################################
# Copyright (C) 2008 VMWare, Inc.
# All Rights Reserved
########################################################################

import logging
import os
import shutil

import Iface
import Metadata
import Transaction
import config
import const
import errors
import platform

'''This module contains children of the Command class.  Each instance
   represents a command primitive to the esxupdate command line.'''


# TODO: have the logger name be tied to each command
log = logging.getLogger('cmdline')


class Command(object):
   '''An interface to specific logic for a given command based on command-
      line arguments.  This is an abstract base class.  Override it with your
      own implementation.'''

   GLOBAL_OPTIONS = ('--meta', '--bundle', '--loglevel', '--HA', '--http_proxy',
         '--timeout', '--retry')
   COMMAND_OPTIONS = ()

   hidden = False

   PLATFORM_UNSUPPORTED_CMDS = {'cos': tuple(),
                                'visor-usb': tuple(),
                                'visor-thin': tuple(),
                                'visor-offline' : tuple(),
                                'visor-pxe': ('scan', 'stage', 'update', 'remove')}

   def __init__(self, alias, parser, optvalues):
      '''Class constructor.  The 'alias' argument is the command-line argument
         which was mapped to this command.  The 'parser' argument is an
         instance of optparse.OptionParser.  The 'optvalues' argument is an
         instance of optparse.Values.
      '''
      #
      # filter unsupported command
      #
      self.CheckCommand(alias)

      import downloader
      self.alias = alias
      self.parser = parser
      self.optvalues = optvalues
      #
      # Select the correct output interface
      #
      if self._getopt('HA'):
         self.iface = Iface.HostdIface()
      else:
         self.iface = Iface.CommandLineIface()
      #
      # Set the HTTP proxy
      #
      downloader.SetProxy(optvalues.proxyurl)

      # Set downloader timeout
      downloader.SetTimeout(optvalues.timeout)

      downloader.SetRetry(optvalues.retry)

   def _getopt(self, opt):
      return getattr(self.optvalues, opt, None)

   def CheckArgs(self):
      '''Make sure arguments are valid for this command.  Raise error if
         they are not by calling parser.error().
      '''
      # This is a bit cumbersome, but optparse does not give us an easy way
      # to do this.
      validopts = list()
      for opt in self.GLOBAL_OPTIONS + self.COMMAND_OPTIONS:
         option = self.parser.get_option(opt)
         if option and option not in validopts:
            validopts.append(option)
      for option in self.parser.option_list:
         if option in validopts:
            continue
         if option.dest and getattr(self.optvalues, option.dest):
            self.parser.error("'%s' invalid with '%s' command" %
                              (option, self.alias))
      # These cannot be combined...
      if self._getopt('meta') and self._getopt('bundlezips'):
         self.parser.error("Options '%s' and '%s' cannot be used together" %
                           (self.parser.get_option('--meta'),
                            self.parser.get_option('--bundle')))

   def CheckCommand(self, cmd):
      ''' Check whether the command is supported by the running platform.'''
      mode = platform.GetRunningMode()
      if cmd in self.PLATFORM_UNSUPPORTED_CMDS[mode]:
         platforminfo = "platform running mode: %s" % (mode)
         raise errors.UnsupportedCommandError(cmd,
               'Command "%s" is not supported on :\n%s' %
               (cmd, platforminfo))

   def DownloadBundleZip(self, bundleURL):
      """
      Downloads an offline bundle .zip at bundleURL, returning
      the local file path of the metadata.zip in the .zip
      Throws:
       BundleDownloadError, BundleFormatError, FileIOError
      """
      import os
      import downloader

      cachedir = config.GetConf().default.cachedir

      try:
         location = os.path.join(cachedir,
                                 'bundle%s.zip' % str(hash(bundleURL)))
         d = downloader.Downloader(bundleURL, location,
                                   progress=self.iface.GetProgress())
         location = d.Get()
      except downloader.DownloaderError, e:
         raise errors.BundleDownloadError(e.url, e.local, e.err)

      return 'zip:%s?metadata.zip' % location

   def Run(self):
      '''Implements the code necessary to run the command.'''
      return NotImplemented

class QueryCommand(Command):
   COMMAND_OPTIONS = ('--vib-view', '--compliant', '--all')

   def Run(self):
      if self._getopt('vibview'):
         t = Transaction.Transaction(loadpkgdb=True)
         self.iface.AddVibQueryResponse(t.vibs)
      else:
         t = Transaction.Transaction()
         if self._getopt('all'):
            buls = t.bulletins
         elif self._getopt('compliant'):
            buls = t.bulletins.compliant
         else:
            buls = t.bulletins.succinct
         self.iface.AddQueryResponse(buls)

class InfoCommand(Command):
   COMMAND_OPTIONS = ('-b',)

   def Run(self):
      bzipurls = self._getopt('bundlezips') or list()
      metaurls = self._getopt('meta') or list()
      selected = self._getopt('bundles')

      urls = metaurls + [self.DownloadBundleZip(url) for url in bzipurls]

      # We keep metadata separate from database stuff.  If -m is specified
      # (urls contains something), then look only in the metadata.  Otherwise,
      # look only in the database.
      if urls:
         metadata = Metadata.Metadata(*urls)
         metadata.load(self.iface)
         bulletins = metadata.bulletins
         vibs = metadata.vibs
      else:
         t = Transaction.Transaction(iface=self.iface)
         bulletins = t.bulletins
         vibs = t.vibs

      if not selected:
         selected = bulletins.keys()

      for bundlespec in selected:
         if bundlespec in bulletins:
            self.iface.ShowBulletinDetails(bulletins[bundlespec])
         elif bundlespec in vibs:
            self.iface.ShowMessage("VIBs are not supported by the info "
                                   "command; ignoring %s." % bundlespec)
         else:
            self.iface.ShowMessage("Unknown bulletin %s." % bundlespec)


class UpdateCommand(Command):
   COMMAND_OPTIONS = ('-b', '--noobsoletes', '--reinstall', '--nodeps',
                      '--nosigcheck', '--nocache', '--maintenancemode',
                      '--olderversion')

   def CheckArgs(self):
      Command.CheckArgs(self)
      bundleopt = self.parser.get_option('-b')
      metaopt = self.parser.get_option('--meta')
      zipopt = self.parser.get_option('--bundle')
      if not (self._getopt('bundles') or self._getopt('meta')
              or self._getopt('bundlezips')):
         self.parser.error("'%s' command requires one of '%s', '%s' or '%s'" %
                           (self.alias, bundleopt, metaopt, zipopt))

   def Run(self):
      if self.alias == 'stage':
         nocache = True
      else:
         nocache = self._getopt('nocache') or False
      skipsigcheck = self._getopt('nosigcheck') or False
      nommode = self._getopt('maintenancemode') or False
      hamode = self._getopt('HA') or False
      bzipurls = self._getopt('bundlezips') or list()
      metaurls = self._getopt('meta') or list()
      selected = self._getopt('bundles') or list()
      nodeps = self._getopt('nodeps') or False
      olderversion = self._getopt('olderversion') or False
      noconflicts = False

      urls = metaurls + [self.DownloadBundleZip(url) for url in bzipurls]

      # Mount default ramdisk for visor if scratch is not set up and cache dir
      # is not mounted yet; do nothing for other cases
      platform.CheckCache(cachesize=0, cleancache=False)

      t = Transaction.Transaction(iface=self.iface, loadpkgdb=True, *urls)

      for bundlespec in selected:
         if os.path.exists(bundlespec) or bundlespec.count('/'):
            # Yes, if nocache is True, you will download the VIB twice.
            t.updatevibfromurl(bundlespec, skipsigcheck=skipsigcheck,
                               nocache=nocache)
         else:
            t.update(bundlespec)

      if not selected and not t.updateall():
         msg = 'No bulletins for this platform could be found. Nothing to do.'
         raise errors.NoMatchError('', msg)

      if self.alias == 'stage':
         t.stage(skipsigcheck=skipsigcheck, nocache=nocache)
      else:
         rc = t.run(skipsigcheck=skipsigcheck, nocache=nocache, nommode=nommode,
                    nodeps=nodeps, olderversion=olderversion,
                    noconflicts=noconflicts, installdeps=False)

         if rc == t.SUCCESS_REQUIRES_REBOOT:
            raise errors.NeedsRebootResult()
         elif config.IsESXhost() and rc == t.SUCCESS_REQUIRES_HOSTD_RESTART:
            if hamode:
               # hostd will know to restart itself with this code
               raise errors.HostdRestartResult()
            else:
               platform.RestartHostd()

class ScanCommand(Command):
   COMMAND_OPTIONS = ('--all', '--vib-view', '--cleancache', '--cachesize')

   def CheckArgs(self):
      Command.CheckArgs(self)
      metaopt = self.parser.get_option('--meta')
      zipopt = self.parser.get_option('--bundle')

      if self._getopt('cachesize') is not None or self._getopt('cleancache'):
         if not self.alias == 'check':
            self.parser.error("'%s' command does not support either "
                              "--cachesize or --cleancache option" %
                              (self.alias))
         elif self._getopt('meta') or self._getopt('bundlezips'):
            log.warning('--meta and --bundle options will be ignored when '
                        '--cachesize or --cleancache is specified')
      else:
         if not (self._getopt('meta') or self._getopt('bundlezips')):
            self.parser.error("'%s' command requires '%s' or '%s'" %
                              (self.alias, metaopt, zipopt))

   def Run(self):
      cleancache = self._getopt('cleancache') or False
      cachesize = self._getopt('cachesize')
      vibview = self._getopt('vibview') or False
      all = self._getopt('all') or False
      bzipurls = self._getopt('bundlezips') or list()
      metaurls = self._getopt('meta') or list()

      urls = metaurls + [self.DownloadBundleZip(url) for url in bzipurls]

      if self.alias == 'check':
         t = Transaction.Transaction(iface=self.iface, usepkgdb=False, *urls)
      elif self._getopt('HA'):    # scan --HA
         t = Transaction.Transaction(iface=self.iface, loadpkgdb=True, *urls)
      else:
         # For CLI scan, do not create VIBs for all host pkgs, just report
         # on the metadata passed
         t = Transaction.Transaction(iface=self.iface, *urls)

      t.metadata.scan()
      #
      # PR 460862: We will be reporting VIBs not in the metadata to VUM,
      #   but VUM wants to limit these extra VIBs just to the potentially
      #   broken ones, so we have to filter out all other extra VIBs.
      #
      extravibs = set(t.pkgdb.vibs.keys())
      if extravibs:
         filterout = extravibs - set(t.metadata.brokenvibs.keys())
         t.metadata.filteroutbyid(filterout)

      self.iface.AddScanResponse(t.metadata, vibview=vibview, all=all,
            scanmode=self.alias)

      ret = platform.CheckCache(cachesize, cleancache, doumount=True)

      # vihostupdate only support /tmp dir..
      log.info("Set up symbolic link to cache dir...")
      if os.path.isfile(const.TMP_CACHE_PATH) \
         or (os.path.islink(const.TMP_CACHE_PATH)
             and os.path.isdir(const.TMP_CACHE_PATH)):
         os.unlink(const.TMP_CACHE_PATH)
      elif os.path.isdir(const.TMP_CACHE_PATH):
         shutil.rmtree(const.TMP_CACHE_PATH)
      os.symlink(ret['location'], const.TMP_CACHE_PATH)
      ret['location'] = const.TMP_CACHE_PATH

      self.iface.AddCheckRespose(ret)

class RemoveCommand(Command):
   COMMAND_OPTIONS = ('-b', '--nodeps', '--maintenancemode', '--force')

   def CheckArgs(self):
      Command.CheckArgs(self)
      bundleopt = self.parser.get_option('-b')
      selected = self._getopt('bundles')
      if not selected:
         self.parser.error("'%s' command requires '%s'" %
                           (self.alias, bundleopt))
      if len(set(selected)) > 1:
         self.parser.error("'%s' command only supports one '%s' option" %
                           (self.alias, bundleopt))

   def Run(self):
      selected = self._getopt('bundles')
      hamode = self._getopt('HA') or False
      nommode = self._getopt('maintenancemode') or False
      nodeps = self._getopt('nodeps') or False
      force = self._getopt('force') or False

      t = Transaction.Transaction(iface=self.iface)

      for bundlespec in selected:
         t.remove(bundlespec, force)

      rc = t.run(nommode=nommode, nodeps=nodeps)

      if rc == t.SUCCESS_REQUIRES_REBOOT:
         raise errors.NeedsRebootResult()
      elif config.IsESXhost() and rc == t.SUCCESS_REQUIRES_HOSTD_RESTART:
         if hamode:
            # hostd will know to restart itself with this code
            raise errors.HostdRestartResult()
         else:
            platform.RestartHostd()

class ClearPendingCommand(Command):
   hidden = True

   def Run(self):
      t = Transaction.Transaction()
      changed = 0
      for v in t.database.vibs.values():
         if v.state.PENDING in v.state:
            log.debug("Clearing 'pending' state on VIB ID '%s'" % v.id)
            v.state.discard(v.state.PENDING)
            changed += 1
      if changed:
         log.info("Updating database to clear 'pending' state on %d entries" %
                  changed)
         t.database.save()
         # Ugh.  So lame...
         if (hasattr(platform, 'conf') and
             hasattr(platform.conf, 'UpdateStagedDb')):
            log.info("Writing staged database to persistent tgz")
            platform.conf.UpdateStagedDb()

class SyncDBCommand(Command):
   hidden = True

   def Run(self):
      bzipurls = self._getopt('bundlezips') or list()
      metaurls = self._getopt('meta') or list()

      urls = metaurls + [self.DownloadBundleZip(url) for url in bzipurls]

      t = Transaction.Transaction(*urls)
      t.database.lock()
      t.updatedb()

      # The syncdb command isn't really intended for Visor, but what the heck.
      if (hasattr(platform, 'conf') and
          hasattr(platform.conf, 'UpdateStagedDb')):
         log.info("Writing staged database to persistent tgz")
         platform.conf.UpdateStagedDb()

      t.database.unlock()

COMMANDS = {'query':  QueryCommand,
            'info':   InfoCommand,
            'update': UpdateCommand,
            'stage':  UpdateCommand,
            'scan':   ScanCommand,
            'check':  ScanCommand,
            'remove': RemoveCommand,
            'clearpending':  ClearPendingCommand,
            'syncdb': SyncDBCommand}
