#######################################################################
# Copyright (C) 2008 VMWare, Inc.
# All Rights Reserved
########################################################################
#
# config.py
#
# Implements EsxupdateConfig and EsxupdateConfigSection classes.
# Sample usage:
# >>> from config import EsxupdateConfig
# >>> c = EsxupdateConfig('/etc/vmware/esxupdate/esxupdate.conf')
# >>> print c.log.level
# DEBUG
#

from ConfigParser import ConfigParser
import errors
from const import *
import os
import shutil

class EsxupdateConfigSection(dict):
   '''Representation of esxupdate configuration section.  Each attribute is
      an option name/value within the section.'''
   def __getattr__(self, name):
      if name not in self:
         raise AttributeError("%s instance has no attribute '%s'" %
                              (self.__class__.__name__, name))
      return self[name]

class EsxupdateConfig(ConfigParser):
   '''Representation of esxupdate configuration, ini-style format.  Each
      attribute is an EsxupdateConfigSection object.'''
   def __init__(self, *args):
      'Accepts as a parameter list one or more configuration files to parse.'
      ConfigParser.__init__(self)
      # These are defaults, and we want to be certain they're always defined.
      self.add_section('platform')
      self.set('platform', 'product', '')
      self.set('platform', 'version', '')
      self.set('platform', 'build', '')
      self.set('platform', 'offline', '')
      self.set('platform', 'restricted', '')
      self.add_section('log')
      self.set('log', 'file', '/var/log/vmware/esxupdate.log')
      self.set('log', 'level', 'DEBUG')
      self.set('log', 'size', '1024000')
      self.add_section('visor')
      self.set('visor', 'bootbankdir', '/bootbank')
      self.set('visor', 'altbootbankdir', '/altbootbank')
      self.set('visor', 'lockerdir', '/locker')
      self.set('visor', 'packagingdir', '/tmp/stage')
      self.set('visor', 'stagedir', '/tmp/stage/firmware')
      self.set('visor', 'oemstgdir', '/tmp/stage/oemstage')
      self.set('visor', 'customstgdir', '/tmp/stage/customstage')
      self.set('visor', 'distribution', 'ESXi')
      self.set('visor', 'defaultdb', '/etc/vmware/esxupdate/status')
      self.set('visor', 'helperscript', '/sbin/bootbankstage-install.sh')
      self.set('visor', 'ipkgconf', '/tmp/ipkg.conf')
      self.set('visor', 'rootdb', '/usr/lib/ipkg/status')
      self.add_section('default')
      self.set('default', 'cachedir', '/var/tmp/esxupdate')
      self.set('default', 'defaultcachemb', '400')
      self.set('default', 'mincachemb', '100')

      self._sectioncache = dict()

      try:
         self.read(args)
      except Exception, e:
         raise errors.ConfigError(args, e)

   def __getattr__(self, name):
      if not self.has_section(name):
         raise AttributeError("%s instance has no attribute '%s'" %
                              (self.__class__.__name__, name))
      if name not in self._sectioncache:
         self._sectioncache[name] = \
            EsxupdateConfigSection([(x, self.get(name, x)) for x in
                                    self.options(name)])
      return self._sectioncache[name]

_default_config = None
def GetConf(configfiles = CONFIGFILE):
   '''Retrieves the instance of the default configuration.  Useful to avoid
      parsing the configuration multiple times when various modules import this
      one within the same script.
      Args:
         configfiles - If specified, a single file or list of files to parse
                       the default configuration from.
      Returns:
         An instance of the default configuration.
      Raises:
         errors.ConfigError on a parsing error.
   '''
   global _default_config
   if not _default_config:
      _default_config = EsxupdateConfig(configfiles)
   return _default_config


def CreateVibEnv(vibConf, sourceConf = CONFIGFILE):
   '''Creates vib.conf for consumption by vib-env / Vib
      pre/post scripts, by copying from sourceConf.
      Existing vib.conf is overwritten.
      Directories are created as needed.
   '''
   vibConfDir = os.path.dirname(vibConf)
   if not os.path.isdir(vibConfDir):
      try:
         os.makedirs(vibConfDir)
      except EnvironmentError, e:
         raise errors.FileIOError(vibConfDir, 'Unable to create dir %s: %s' %
                                  (vibConfDir, str(e)) )
   try:
      shutil.copy(sourceConf, vibConf)
   except EnvironmentError, e:
      raise errors.FileIOError(vibConf, 'Error copying %s to %s: %s' %
                               (sourceConf, vibConf, str(e)) )

   #
   # Make vib.conf writeable so scripts can set vars
   #
   os.chmod(vibConf, 0644)


def GetPostUpdateScripts(vibConf):
   ''' Returns a list of scripts to run at the end of the update
       transaction.  The list is from the post-update section of
       vibConf, sorted alphanumerically by var name.  For ex,:
         [post-update]
         100_mkinitrd = /usr/bin/mkinitrd -v /boot/kernelz-4.2.3-99
         101_pciid = /usr/sbin/esxcfg-pciid
       If there is no such section, an empty list is returned.
   '''
   conf = ConfigParser()
   try:
      conffile = file(vibConf)
      conf.readfp(conffile)
      conffile.close()
   except Exception, e:
      raise errors.FileIOError(vibConf, "Cannot read from %s: %s" %
                                (vibConf, e))

   if not conf.has_section('post-update'):
      return []
   items = conf.items('post-update')
   items.sort(lambda a, b: cmp(a[0], b[0]))
   return [item[1] for item in items]

def GetPlatform():
   try:
      product = GetConf().platform.product
   except Exception, e:
      raise errors.ConfigError(message='product must be specified in platform '
            'section in esxupdate.conf file')

   PRODUCT2PLATFORM = {'esx': PLATFORMS.cos,
                       'embeddedEsx': PLATFORMS.visor,
                       'vima': PLATFORMS.cos}
   if product in PRODUCT2PLATFORM:
      return PRODUCT2PLATFORM[product]
   else:
      raise errors.ConfigError(message='product must be a known product '
            'type: %s\nGot: %s' %(' or '.join(INT_PRODUCTS), product))

def IsESXhost():
   product = GetConf().platform.product
   return product in (INT_PRODUCTS.esx, INT_PRODUCTS.embeddedEsx) and \
          not GetConf().platform.offline
