#######################################################################
# Copyright (C) 2008 VMWare, Inc.
# All Rights Reserved
########################################################################

__all__ = ['DepSolver']

import VIBCollection
import errors
from const import *

class Requirement(object):
   '''Represents a generic requirement, along with references to the Vibs that
      require it and those that provide it.'''
   def __init__(self, name, requirers, providers):
      self.name = name
      self.requirers = requirers
      self.providers = providers

   def __str__(self):
      return self.name

class RequirementsList(dict):
   '''Represents a collection of 'Requirement' objects, and provides some
      convenience functions for managing them.'''

   @property
   def providers(self):
      '''Get a VIBCollection instance of all providers in the RequirementList.'''
      providerscache = VIBCollection.VIBCollection()
      for req in self.values():
         for provider in req.providers:
            providerscache[provider.id] = provider
      return providerscache

   @property
   def requirers(self):
      '''Get a VIBCollection instance of all requirers in the RequirementsList.'''
      requirerscache = VIBCollection.VIBCollection()
      for req in self.values():
         for requirer in req.requirers:
            requirerscache[requirer.id] = requirer
      return requirerscache

   def getByProvider(self, provider):
      '''Return a RequirementsList object with only the subset of Requirements
         that have a given provider.'''
      reqlist = RequirementsList()
      for req in self:
         if provider in self[req].providers:
            reqlist[req] = self[req]
      return reqlist

   def getByRequirer(self, requirer):
      '''Return a RequirementsList object with only the subset of Requirements
         that have a given requirer.'''
      reqlist = RequirementsList()
      for req in self:
         if requirer in self[req].requirers:
            reqlist[req] = self[req]
      return reqlist

class Problem(object):
   '''A base class representing a dependency problem.'''
   def __eq__(self, other):
      return str(self) == str(other)

   def __ne__(self, other):
      return str(self) != str(other)

class Conflict(Problem):
   '''A class representing a dependency that could not be satisfied due to a
      conflict with another VIB.'''
   def __init__(self, req, vib, other):
      self.req = req
      self.vib = vib
      self.other = other

   def __str__(self):
      requirers = ', '.join([str(v) for v in self.req.requirers])
      return "%s provides '%s' (required by %s), but conflicts with %s" \
             % (self.vib, self.req.name, requirers, self.other)

class HostConflict(Problem):
   '''A class representing a dependency that could not be satisfied due to a
      conflict with the host.'''
   def __init__(self, req, vib):
      self.req = req
      self.vib = vib

   def __str__(self):
      requirers = ', '.join([str(v) for v in self.req.requirers])
      return "%s provides '%s' (required by %s), but conflicts with the host" \
             % (self.vib, self.req.name, requirers)

class HostBreakage(Problem):
   '''A class representing VIBs that could not be installed due to a
      breakage with an API used by host packages.'''
   def __init__(self, vibs, api, broken):
      self.vibs = vibs
      self.api = api
      self.broken = broken

   def __str__(self):
      if len(self.vibs) > 1:
         text = "VIBs %s break" % (', '.join(str(v) for v in self.vibs))
      else:
         text = "VIB %s breaks" % (', '.join(str(v) for v in self.vibs))
      brokenpkgs = ', '.join('%s %s' % (n, v) for n, v in self.broken)
      return "%s host API %s used by host packages %s" % (text, self.api,
                                                          brokenpkgs)

class Obsoletion(Problem):
   '''A class representing a dependency that could not be satisfied due to an
      obsoletion by another VIB.'''
   def __init__(self, req, vib, by):
      self.req = req
      self.vib = vib
      self.by = by

   def __str__(self):
      requirers = ', '.join([str(v) for v in self.req.requirers])
      return "%s provides '%s' (required by %s), but is obsoleted by %s" \
             % (self.vib, self.req.name, requirers, self.by)

class HostObsoletion(Problem):
   '''A class representing a dependency that could not be satisfied due to an
      obsoletion by the host.'''
   def __init__(self, req, vib):
      self.req = req
      self.vib = vib

   def __str__(self):
      requirers = ', '.join([str(v) for v in self.req.requirers])
      return "%s provides '%s' (required by %s), but is obsoleted by the host" \
             % (self.vib, self.req, requirers)

class NoProvider(Problem):
   '''A class representing a dependency that could not be satisfied due to a
      missing provider.'''
   def __init__(self, req):
      self.req = req

   def __str__(self):
      requirers = ', '.join([str(v) for v in self.req.requirers])
      return "No VIB provides '%s' (required by %s)" \
             % (self.req.name, requirers)

class RequestedVibConflict(Problem):
   '''A class representing a user-requested VIB that can not be installed
      because it conflicts with another VIB.'''
   def __init__(self, vib, other):
      self.vib = vib
      self.other = other

   def __str__(self):
      return "Requested VIB %s conflicts with %s" % (self.vib, self.other)

class RequestedVibHostConflict(Problem):
   '''A class representing a user-requested VIB that can not be installed
      because it conflicts with the host.'''
   def __init__(self, vib):
      self.vib = vib

   def __str__(self):
      return "Requested VIB %s conflicts with the host" % self.vib

class RequestedVibObsoletion(Problem):
   '''A class representing a user-requested VIB that can not be installed
      because it is obsoleted by another VIB.'''
   def __init__(self, vib, by):
      self.vib = vib
      self.by = by

   def __str__(self):
      return "Requested VIB %s is obsoleted by %s" % (self.vib, self.by)

class RequestedVibHostObsoletion(Problem):
   '''A class representing a user-requested VIB that can not be installed
      because it is obsoleted by the host.'''
   def __init__(self, vib):
      self.vib = vib

   def __str__(self):
      return "Requested VIB %s is obsoleted by the host" % self.vib

class RequestedVibAlreadyInstalled(Problem):
   '''A class representing a user-requested VIB that can not be installed
      because it is already installed on the host.'''
   def __init__(self, vib):
      self.vib = vib

   def __str__(self):
      return "Requested VIB %s is already installed" % self.vib

class ProblemSet(set):
   '''A collection of one or more problems encountered during dependency
      resolution.'''
   pass

class DepSolver(object):
   '''A class for resolving dependencies.
      Arguments:
         * vibs     - The VIBs for which we want to resolve dependencies.
                      Should be a VIBCollection or dict of VIBs keyed by id.
         * metadata - The additional VIBs from which we can obtain dependencies.
                      Should be a VIBScanner with VIBs from both metadata and
                      installed packages.  Should have been scanned.
         * nodeps   - Skip reporting of unsatisfied dependencies as problems
         * noconflicts - Skip reporting of conflicts and breakages as problems
         * olderversion - Do not check VIB obsoletion.
   '''
   def __init__(self, vibs, metadata, nodeps=False, noconflicts=False,
                olderversion=False):
      self.vibs = vibs
      self.metavibs = metadata.vibs
      self.requirements = RequirementsList()
      self.deps = VIBCollection.VIBCollection()
      self.nodeps = nodeps
      self.noconflicts = noconflicts
      self.olderversion = olderversion
      self.brokendeps = metadata.brokendeps

   def _CalculateReplacements(self, vibs):
      '''Computes _pendingreplacement based on the list or set of vibs.
         We need to know if things are going to be replaced on the host by our
         pending transaction, so that we do not raise an error for things that
         conflict with or are obsoleted by them.
      '''
      if not self.noconflicts:
         for v in vibs:
            self._pendingreplacement.update(v.hostobsoleted)
         if self.olderversion:
            for v in vibs:
               self._pendingreplacement.update(v.hostobsoleting)

   def _GetRequirements(self):
      '''Iterate over the VIBs in self.vibset and self.deps to popluate the
         RequirementsList in self.requirements.  Return True if new items
         have been populated, or False if nothing has changed from previous
         invocation.
      '''
      changed = False
      for vib in self.vibs.values() + self.deps.values():
         for requires in vib.requires:
            # provided by host package that we aren't replacing
            if requires.hostprovides - self._pendingreplacement:
               continue
            if requires.providedby:
               found = False
               for p in requires.providedby:
                  if p in self.vibs or p in self.deps.keys():
                     found = True
                     break
               if found:
                  # already provided by something in self.vibset or self.deps.
                  continue # for requires in vib.requires
            if requires.name in self.requirements:
               requirement = self.requirements[requires.name]
               if vib.id not in requirement.requirers:
                  requirement.requirers[vib.id] = vib
                  changed = True
            else:
               requirers = VIBCollection.VIBCollection()
               providers = VIBCollection.VIBCollection()
               requirers[vib.id] = vib
               for provider in requires.providedby:
                  pvib = self.vibs.get(provider) or self.metavibs.get(provider)
                  #
                  # Don't count already installed VIBs as providers
                  if pvib and not pvib.state.INSTALLED in pvib.state:
                     providers[provider] = pvib
               requirement = Requirement(requires.name, requirers, providers)
               self.requirements[requires.name] = requirement
               changed = True
      return changed

   def _scoreProvider(self, provider):
      '''A function used to score a providers.  Returns a list.  Can be used
         to sort a list based on
         cmp(_scoreProvider(provider1), _scoreProvider(provider2)).
         Documentation is the comments in this code.
      '''
      # This function probably doesn't scale well for huge sets of VIBs, but it
      # doesn't need to, since an entire ESX host has only about 400 packages.

      # Calculate scores based on various weights.  This uses a rather nifty
      # property of sequence comparisons:
      # [a, b, c] > [d, e f] is equivalent to
      # a > d or (a == d and b > e) or (a == d and b == e and c > f)
      # Note that higher scores are less desirable.
      score = list()

      # Avoid conflicts at all costs, both with self.vibset and with the host.
      conflicts = set(
         [v for v in self.vibs.keys() if v in provider.conflicts]
      )
      conflicts.update(
         [v for v in self.vibs.values() if provider.id in v.conflicts]
      )

      score.append(len(provider.hostconflicts) + len(conflicts))

      # If an item is obsoleted by something in self.vibset or the host,
      # then it's going to be impossible anyway, so it should compare
      # unfavorably.
      obsoletedby = \
         set([v for v in self.vibs.values() if provider.id in v.oldervibs])
      score.append(len(provider.hostobsoleting) + len(obsoletedby))

      # Score according to number of things that we obsolete.  Note that this
      # takes care of ordering obsoletion chains, as obsoletions are cumulative
      # per smart and our scan results.  I.e. if B -obs-> C and A -obs-> B,
      # then A.olderVibs contains B and C.
      score.append(len(provider.oldervibs))

      # If a VIB satisfies requirements for more VIBs than another, that would
      # seem a better choice.
      requirers = set()
      for req in self.requirements.getByProvider(provider.id).values():
         requirers.update(req.requirers)
      score.append(-len(requirers))

      # If a VIB gets along better with others (fewer conflicts), that should
      # compare favorably.  Not that 'conflicts' set may already contain things
      # from previous search of self.vibset.
      conflicts.update(
         [v for v in self.metavibs.values() if provider.id in v.conflicts]
      )
      score.append(len(conflicts))

      # If a VIB is obsoleted by more things than another...  Note that
      # obsoletedby still contains things from previous search of self.vibset.
      obsoletedby.update(
         [v for v in self.metavibs.values() if provider.id in v.oldervibs]
      )
      score.append(len(obsoletedby))

      # If we've gotten this far, it's pretty arbitrary, but let's just add
      # something to make comparisons deterministic.
      score.append(provider.id)

      return score

   def _ResolveRequirement(self, requirement):
      '''Find the best match for a given requirement.
         Argument:  requirement = a Requirement object.
         Returns:  a Vib instance, representing the best VIB to satisfy the
                   requirement, or None if the req cannot be satisfied
                   and nodeps is True.
         Raises:  DependencyError if the requirement cannot be resolved.
      '''
      # Get a list of things by score.
      provideslist = list()
      for provider in requirement.providers.values():
         provideslist.append((provider, self._scoreProvider(provider)))
      provideslist.sort(lambda x, y: cmp(x[1], y[1]))
      if not provideslist:
         if self.nodeps:
            return None
         else:
            problems = ProblemSet((NoProvider(requirement),))
            raise errors.DependencyError(problems)

      return provideslist[0][0]

   def _CheckRequirement(self, vib, requirement):
      '''Check the Vib object for conflicts and obsoletions.  Intended to
         do a sanity check on the list of VIBs added as dependencies, to
         make sure they don't conflict with or obsolete each other or the
         selected VIBs, and do not conflict with nor are obsoleted by the host.
         Argument: vib, a Vib object.
         Returns:  Nothing.
         Raises:   DependencyError if one or more problems found.
      '''
      problems = ProblemSet()

      if not self.noconflicts:
         if vib.hostconflicts - self._pendingreplacement:
            problems.add(HostConflict(requirement, vib))

         conflicts = set([v for v in self.vibs.values() + self.deps.values()
                          if vib.id in v.conflicts or v.id in vib.conflicts])
         if conflicts:
            for c in conflicts:
               problems.add(Conflict(requirement, vib, c))

      if not self.olderversion:
         if vib.obsoletedbyhost:
            problems.add(HostObsoletion(requirement, vib))

         obsoletions = set([v for v in self.vibs.values()
                            if vib.id in v.oldervibs])
         if obsoletions:
            for o in obsoletions:
               problems.add(Obsoletion(requirement, vib, o))

      if problems:
         raise errors.DependencyError(problems)

   def _CheckRequested(self, vib):
      '''Check the Vib object for conflicts and obsoletions.  Intended to
         do a sanity check on the list of VIBs passed in to the resolver, to
         make sure they don't conflict with or obsolete each other, and do not
         conflict with nor are obsoleted by the host.
         Argument: vib, a Vib object.
         Returns:  Nothing.
         Raises:   DependencyError if one or more problems found.
      '''
      # This looks a lot like the code above in _CheckRequirement, but there
      # are enough subtle differences that it seemed a little difficult to
      # abstract it all out to a common method.
      problems = ProblemSet()

      if vib.state.INSTALLED in vib.state:
         problems.add(RequestedVibAlreadyInstalled(vib))

      if not self.noconflicts:
         if vib.hostconflicts - self._pendingreplacement:
            problems.add(RequestedVibHostConflict(vib))

         conflicts = set([v for v in self.vibs.values()
                          if vib.id in v.conflicts or v.id in vib.conflicts])
         if conflicts:
            for c in conflicts:
               problems.add(RequestedVibConflict(vib, c))

      if not self.olderversion:
         if vib.obsoletedbyhost:
            problems.add(RequestedVibHostObsoletion(vib))

         obsoletions = set([v for v in self.vibs.values()
                            if vib.id in v.oldervibs])
         if obsoletions:
            for o in obsoletions:
               problems.add(RequestedVibObsoletion(vib, o))

      if problems:
         raise errors.DependencyError(problems)

   def _CheckBrokenHostDeps(self):
      '''Checks to see if any host dependencies will be broken
         by any of the requested or pulled-in VIBs.
         Returns:  Nothing.
         Raises:   DependencyError if one or more problems found.
      '''
      problems = ProblemSet()
      instlist = set(self.vibs.keys() + self.deps.keys())

      # Iterate through the deps considered potentially broken
      for req, brokendep in self.brokendeps.items():
         # Skip if all host packages requiring the dep will be
         # replaced
         stillbroken = brokendep.hostrequiredby - self._pendingreplacement
         if not stillbroken:
            continue
         # Are any incompatible VIBs in our install list?
         incompatibles = brokendep.breaks & instlist
         if not incompatibles:
            continue
         # If so, will another VIB on our install list or that
         # is installed satisfy the dependency?
         if brokendep.hostprovides - self._pendingreplacement:
            continue
         if brokendep.providedby & instlist:
            continue

         problems.add(HostBreakage(incompatibles, req, stillbroken))

      if problems:
         raise errors.DependencyError(problems)

   def Resolve(self):
      '''The method used to resolve dependencies for the given set of VIBs
         (see class doc string).
         Returns: A VIBCollection instance of the dependencies required.
         Raises:  DependencyError if one or more dependencies could not be
                  resolved.
      '''
      problems = ProblemSet()
      self._pendingreplacement = set()
      self._CalculateReplacements(self.vibs.values())

      # If self.nodeps is set, we assume that the user/caller knows what
      # they're doing, and we will not attempt to resolve any dependencies,
      # even though they may exist in available metadata.

      if not self.nodeps:
         # This dict maps each requirement to the VIB we chose to satisfy it.
         requirements_map = dict()
         # Make sure _GetRequirements() always returns true the first time
         # and Resolve() can be called safely more than once
         self.deps.clear()
         self.requirements.clear()
         while self._GetRequirements():
            for requirement in self.requirements.values():
               try:
                  match = self._ResolveRequirement(requirement)
                  if match:
                     self.deps.setdefault(match.id, match)
                     requirements_map[requirement] = match
               except errors.DependencyError, e:
                  problems.update(e.problemset)

      self._CalculateReplacements(self.deps.values())

      # Don't trust that the list of VIBs we're given is without its own
      # problems.
      for vib in self.vibs.values():
         try:
            self._CheckRequested(vib)
         except errors.DependencyError, e:
            problems.update(e.problemset)

      # If we've pulled in requirements, check that our matches don't have
      # issues.
      if not self.nodeps:
         for req, vib in requirements_map.items():
            try:
               self._CheckRequirement(vib, req)
            except errors.DependencyError, e:
               problems.update(e.problemset)

      # Check that our VIBs (requested + pulled) do not break host dependencies
      if not self.noconflicts:
         try:
            self._CheckBrokenHostDeps()
         except errors.DependencyError, e:
            problems.update(e.problemset)

      if problems:
         raise errors.DependencyError(problems)

      return self.deps
