#######################################################################
# Copyright (C) 2008 VMWare, Inc.
# All Rights Reserved
########################################################################
#
# errors.py
#

class EsxupdateError(Exception):
   description = ('Abstract base class for esxupdate exceptions.  Each '
                  'exception type has a class variable, errno, which is an '
                  'error code unique to the related error, and description, '
                  'which is a string description of the exception.')
   errno = NotImplemented # Should be overridden in inheriting classes.
   attrs = tuple()

   def __init__(self, *args, **kwargs):
      # 'self' here is likely to be an instance of a child class, hence attrs
      # will actually contain something.
      arglen = len(args) + len(kwargs)
      if arglen > len(self.attrs):
         raise TypeError('%s expected %s arguments, got %s' %
                         (self.__class__, len(self.attrs), arglen))
      alreadyset = list()
      for i in range(len(args)):
         setattr(self, self.attrs[i], args[i])
         alreadyset.append(self.attrs[i])
      for k, v in kwargs.items():
         if k not in self.attrs:
            raise TypeError("%s got an unexpected keyword argument '%s'" %
                            (self.__class__, k))
         if k in alreadyset:
            raise TypeError("%s got multiple values for keyword argument '%s'" %
                            (self.__class__, k))
         setattr(self, k, v)
      # Note:  Built-in Exception() class does not normally accept kw args.
      Exception.__init__(self, *(list(args) + kwargs.values()))

   def __str__(self):
      if self.attrs:
         return Exception.__str__(self)
      else:
         return self.description

class NotRootError(EsxupdateError):
   description = 'The current user is not UID 0 (root).'
   errno = 1

class LockingError(EsxupdateError):
   description = 'Another instance of esxupdate is currently running.'
   errno = 3
   attrs = ('message',)

class MetadataDownloadError(EsxupdateError):
   description = 'Failed to download metadata.'
   errno = 4
   attrs = ('url', 'localfile', 'message')

class MetadataFormatError(EsxupdateError):
   description = 'The format of the metadata is invalid.'
   errno = 5
   attrs = ('filename', 'message')

class VibDownloadError(EsxupdateError):
   description = 'Failed to download VIB.'
   errno = 7
   attrs = ('url', 'localfile', 'message')

class VibFormatError(EsxupdateError):
   description = 'The format of the VIB is invalid.'
   errno = 8
   attrs = ('filename', 'message')

class VibIOError(EsxupdateError):
   description = 'Error reading or writing VIB data.'
   errno = 9
   attrs = ('filename', 'message')

class FileIOError(EsxupdateError):
   description = 'Unable to create, write or read a file as expected.'
   errno = 10
   attrs = ('filename', 'message')

class DatabaseFormatError(EsxupdateError):
   description = 'The format of the database is invalid.'
   errno = 11
   attrs = ('filename', 'message')

class DatabaseIOError(EsxupdateError):
   description = 'Unable to read or write to database.'
   errno = 12
   attrs = ('filename', 'message')

class NoMatchError(EsxupdateError):
   """ Raised due to the following (matching refers to Vibs that have a swPlatform
       supporting the ESX host platform;  A bulletin matches if one or more of
       its Vibs matches the platform):  
     1. [-m meta -b vibID update/stage] - vibID not found in meta
     2. [-m meta -b bulletin update/stage] - bulletin not found in meta
     3. -b vibspec / -b bulletin specifies a nonmatching vib/bulletin
     4. [-m meta update/stage] - no -b, and no matching Vibs found in meta
   """
   description = 'No matching bulletin or VIB was found in the metadata.'
   errno = 13
   attrs = ('id', 'message')

class DependencyError(EsxupdateError):
   description = 'There was an error resolving dependencies.'
   errno = 14
   attrs = ('problemset',)

   # This should maybe go in Iface.py, but for now it's here.
   def __str__(self):
      return '\n'.join([str(p) for p in self.problemset])

class PackageManagerError(EsxupdateError):
   description = 'The package manager transaction failed.'
   errno = 15
   attrs = ('vibs', 'message')

class ConfigError(EsxupdateError):
   description = 'Failed to parse configuration files.'
   errno = 16
   attrs = ('files', 'message')

class VisorSetupError(EsxupdateError):
   description = 'There was an error setting up ESXi installation destination'
   errno = 17
   attrs = ('message',)

class MaintenanceModeError(EsxupdateError):
   description = 'Maintenance mode is not enabled or could not be determined.'
   errno = 18
   attrs = ('message',)

class PostScriptError(EsxupdateError):
   description = 'A post-transaction script failed with a nonzero exit code. '\
                 'Examine esxupdate logs for more details. '
   errno = 19
   attrs = ('script', 'exitcode')

class VibSignatureError(EsxupdateError):
   """ Base class for all Vib signature errors """
   pass

class VibSigMissingError(VibSignatureError):
   description = 'Vib signature is missing.'
   errno = 20
   attrs = ('vib',)

class VibSigVersionError(VibSignatureError):
   description = 'Unsupported Vib signature version.'
   errno = 21
   attrs = ('vib', 'message')

class VibSigFormatError(VibSignatureError):
   description = 'pkcs7.sig file is too short or corrupted.'
   errno = 22
   attrs = ('vib', 'message')

class VibSigInvalidError(VibSignatureError):
   description = 'PKCS7 certificate is not valid (expired, etc) or ' \
                 'the public key is missing.'
   errno = 23
   attrs = ('vib', 'message')

class VibSigDigestError(VibSignatureError):
   description = 'Vib digest does not match the digest in the PKCS7 signature.'
   errno = 24
   attrs = ('vib', 'vibdigest', 'sigdigest')

class UnsatisfiedDependencies(EsxupdateError):
   description = 'Additional VIBs are required to satisfy dependencies.'
   errno = 25
   attrs = ('suggestedvibs', 'message')

class BundleDownloadError(EsxupdateError):
   description = 'Failed to download offline bundle .zip.'
   errno = 26
   attrs = ('url', 'localfile', 'message')

class BundleFormatError(EsxupdateError):
   description = 'Invalid bundle ZIP archive, or missing metadata.zip inside.'
   errno = 27
   attrs = ('filename', 'message')

class UnsupportedCommandError(EsxupdateError):
   description = 'The requested command is not supported on the platform.'
   errno = 28
   attrs = ('command', 'message')

class RestrictedVibError(EsxupdateError):
   description = 'The VIB is not supported on the platform.'
   errno = 29
   attrs = ('vibs', 'message')

class NormalExit(EsxupdateError):
   pass

class NeedsRebootResult(NormalExit):
   description = 'The update completed successfully, ' \
                 'but the system needs to be ' \
                 'rebooted for the changes to be effective.'
   errno = 80

class HostdRestartResult(NormalExit):
   description = 'Hostd needs to be restarted to complete installation.'
   errno = 81

class HostNotChanged(NormalExit):
   description = 'Host was not updated, no changes required.'
   errno = 82
   attrs = ('message',)
