#######################################################################
# Copyright (C) 2008 VMWare, Inc.
# All Rights Reserved
########################################################################
#
# platform.py
#   Platform-specific stuff.
#

from vmware.esx45update.const import *
from vmware.esx45update import config

if config.GetPlatform() == PLATFORMS.cos:
   from _cos import *
elif config.GetPlatform() == PLATFORMS.visor:
   from _visor import *


#
# Methods that apply to both COS and Visor platforms
#

from vmware.esx45update import errors
from vmware.esx45update.util import runcommand

import os

#
#  Checking for the # of running VMs == 0 is not a substitute for
#  maintenance mode.  It does not cover many use cases such as
#  VMotion in progress.
#
#  Even better would be to enter maintenance mode ourselves -- that has a
#    much stronger guarantee, that no ops are in flight to start/stop a VM,
#    for example.  However, we do not do that for these reasons:
#    i) we'd have to exit maint mode, even for exceptions & errors;
#          - what happens if we have errors exiting maint mode?
#    ii) it's much more testing
#
#    TODO: If we do ever get around to entering maint mode, we should use
#    pyVim/pyVmomi, and connect to hostd via a local ticket (or file a
#    PR to do this).
#
def GetMaintenanceMode():
   """ Returns True if the ESX host is in maintenance mode, and
       False otherwise.
       MaintenanceModeError is thrown if vim-cmd has a non-zero return code,
       such as (most likely) hostd isn't up.
   """
   cmd = ['/usr/bin/vim-cmd', 'hostsvc/runtimeinfo']
   rc, out = runcommand.runcommand(cmd)
   if rc != 0:
      log.error('vim-cmd returned nonzero status %d\nOutput:\n%s' % (
         rc, out))
      raise errors.MaintenanceModeError(
         "Unable to determine if the system is in maintenance mode: "
         "vim-cmd returned error (%d).  Please see esxupdate logs for "
         "more details.  To be safe, installation will not continue."
         % rc)

   for line in out.splitlines():
      tokens = line.split('=')
      if tokens[0].strip() == 'inMaintenanceMode':
         return tokens[1].strip().startswith('true')

   raise errors.MaintenanceModeError("Unable to determine if the system "
                                     "is in maintenance mode because vim-"
                                     "cmd did not return inMaintenanceMode "
                                     "value. File a bug against hostd.")


VSISH_CMD = '/usr/sbin/vsish'

def GetBiosVendorModel():
   """ Returns the BIOS vendor name and model strings if they can be
       obtained using vsish.  If they cannot or vsish does not exist,
       returns '', ''.
   """
   #
   # Note: COS doesn't ship with vsish (debug-tools RPM) by default
   #
   if os.path.exists(VSISH_CMD):
      cmd = [VSISH_CMD, '-e', '-p', 'cat', '/hardware/bios/dmiInfo']
      rc, out = runcommand.runcommand(cmd)
      if rc == 0:
         try:
            info = eval(out)
         except Exception, e:
            log.warn('%s parsing vsish output:\n%s' % (e, out))
         else:
            return info.get('vendorName', ''), info.get('productName', '')
      else:
         log.warn('%s returned nonzero status %d\nOutput:\n%s' % (
            cmd, rc, out))
   else:
      log.info('vsish command %s cannot be found' % VSISH_CMD)

   log.info('No hardware vendor/model information obtained')
   return '', ''
