#######################################################################
# Copyright (C) 2008 VMWare, Inc.
# All Rights Reserved
########################################################################
#
# COS-specific platform functions
#
from vmware.esx45update.const import *
from vmware.esx45update import errors
import os
import errno
import fcntl
import select
import threading
import rpm
import logging
import shutil
from vmware.esx45update.util import runcommand
import smart.channel
import smart.backends.rpm
from vmware.esx45update import config

log = logging.getLogger('cos')

class RpmLogFile(object):
   '''Emulates a file object to provide a way to capture RPM log messages and
      pass them to a logger.  Upon instantiation, it sets up a pipe, and starts
      a thread with a select loop to read from the pipe.  Data read from the
      pipe is logged using the logger.  This enables passing the file
      descriptor associated with the write end of the pipe to the RPM library
      so that we can capture log messages written to it.
   '''

   def __init__(self, log = log, loglevel = logging.DEBUG):
      '''Parameters:
            * log -      (Optional) The logging object to which messages should
                         be written.  Must implement, at a minimum, a log()
                         method that takes two parameters: a logging level
                         (integer) and a log message (string). Note that this
                         object must also be thread-safe.
            * loglevel - (Optional) The loglevel at which messages will be
                         logged.
      '''
      self.log = log
      self.loglevel = loglevel
      readfd, writefd = os.pipe()
      # This is for convenience, as the file class' read() method is easier to
      # use than os.read().
      self._readf = os.fdopen(readfd, 'r')
      self._writef = os.fdopen(writefd, 'w')
      self._stop_readloop = threading.Event()
      # We record _mainThread as well, so that we can test to see if it's alive
      # in our child thread.  It is OS-dependent as to whether or not child
      # threads exit with the parent, and we don't want the child thread to
      # keep running indefinitely.
      self._mainThread = threading.currentThread()
      self._childThread = threading.Thread(target = self._readloop)
      self._childThread.start()

   def fileno(self):
     '''Returns a low-level file descriptor (integer) that is suitable for
        passing to os.write() or assigning to rpm.ts.scriptFd.
     '''
     return self._writef.fileno()

   def file(self):
     '''Returns a file object that is suitable for calling write() on, or for
        passing to rpm.setLogFile().
     '''
     # RPM is rather strict on what it accepts to rpm.setLogFile.  Simply
     # having a file-like object that provides a write() method is not
     # sufficient, hence this method.
     return self._writef

   def write(self, data):
      '''Logs the specified data.  The 'data' parameter must be a string.
         Multiple lines are logged as separate log messages.
      '''
      for line in data.splitlines():
         self.log.log(self.loglevel, line)

   def close(self):
      '''Stops the thread which reads from the pipe, and closes open file
         descriptors.
      '''
      self._stop_readloop.set()
      # This timeout is a small eternity, but it gaurantees that we never hang.
      self._childThread.join(5.0)

   def _readloop(self):
      # Set O_NONBLOCK on _readfd to prevent blocking during reads.
      flags = fcntl.fcntl(self._readf, fcntl.F_GETFL) | os.O_NONBLOCK
      fcntl.fcntl(self._readf, fcntl.F_SETFL, flags)

      data = ''
      while self._mainThread.isAlive() and not self._stop_readloop.isSet():
         try:
            ready = select.select([self._readf], [], [], 1.0)[0]
         except select.error, (n, msg):
            if n == errno.EINTR:
               continue
            else:
               # If we have data, try to log it before we bail.
               if data:
                  self.log.log(self.loglevel, data)
               self.log.log(logging.ERROR, 'Error during select: %s' % msg)
               return
         if not ready:
            # select timed out.  Loop again so we can check _stop_readloop.
            continue
         try:
            data += self._readf.read()
         except IOError, (n, msg):
            if n == errno.EAGAIN:
               # Nothing to read yet; let's go back to our loop.
               continue
            else:
               # Something bad happened.  Log any pending data and get out.
               if data:
                  self.log.log(self.loglevel, data)
               self.log.log(logging.ERROR, 'Error reading from pipe: %s' % msg)
               return

         lines = data.splitlines()
         if data[-1] in ('\r', '\n'):
            data = ''
         else:
            # If the last line is a partial line, save it for the next iteration.
            data = lines[-1]
            del lines[-1]

         # Do the actual logging.
         for line in lines:
            self.log.log(self.loglevel, line)

      # We've exited the while loop, presumably because the _stop_readloop
      # event was set. Log any remaining data and return.
      try:
         data += self._readf.read()
      except IOError, (n, msg):
         if n != errno.EAGAIN:
            self.log.log(logging.ERROR, 'Error reading from pipe: %s' % msg)
      if data:
         self.log.log(self.loglevel, data)
      self._writef.close()
      self._readf.close()
      return

# Set up a default RPM log level and file. We do this during an import of this
# module (platform), so that other users of the rpm Python module (i.e. smart)
# do not cause RPM to print things to stderr.
rpmlogfile = RpmLogFile(logging.getLogger('cos.rpm'))
rpm.setVerbosity(rpm.RPMLOG_INFO)
rpm.setLogFile(rpmlogfile.file())

class RpmTransactionCb(object):
   def __init__(self, progress=None, logfile=None):
      self.fd = None
      self.progress = progress
      self.logfile = logfile
      self.updates = set()
      self.errors = dict()

   def _log(self, msg):
      '''Write message to logfile with newline'''
      #
      # We don't use the standard log mechanism because
      # our handler is a RotatingFileHandler, and we
      # don't want the file rotated, and the file object
      # closed silently, while RPM is still set to log
      # to it.
      #
      if self.logfile:
         self.logfile.write('%s\n' % msg)

   def __call__(self, reason, amount, total, key, client_data):
      if reason == rpm.RPMCALLBACK_INST_OPEN_FILE:
         if self.fd != None:
            os.close(self.fd)
         self.fd = os.open(key.packagelocation, os.O_RDONLY)
         return self.fd
      elif reason == rpm.RPMCALLBACK_INST_START:
         self.updates.add(key.name)
         msg = 'Installing %s' % key.name
         self._log('%s %s' % (msg, key.version))
         if self.progress:
            self.progress.start()
            self.progress.setTopic(msg)
            self.progress.set(amount, total)
            self.progress.show()
      elif reason == rpm.RPMCALLBACK_INST_PROGRESS:
         if self.progress:
            self.progress.set(amount, total)
            self.progress.show()
      elif reason == rpm.RPMCALLBACK_INST_CLOSE_FILE:
         if key not in self.errors:
            key.state.add(key.state.INSTALLED)
            if key.rebootrequired:
               key.state.add(key.state.PENDING)
            key.updatetimestamp()
         if self.fd != None:
            os.close(self.fd)
            self.fd = None
         if self.progress:
            self.progress.stop()
      elif reason == rpm.RPMCALLBACK_UNINST_START:
         if key in self.updates:
            msg = 'Cleaning up %s' % key
         else:
            msg = 'Removing %s' % key
         self._log(msg)
         if self.progress:
            self.progress.start()
            self.progress.setTopic(msg)
            self.progress.set(amount, total)
            self.progress.show()
      elif reason == rpm.RPMCALLBACK_UNINST_PROGRESS:
         if self.progress:
            self.progress.set(amount, total)
            self.progress.show()
      elif reason == rpm.RPMCALLBACK_UNINST_STOP:
         if self.progress:
            self.progress.stop()
      elif reason in (rpm.RPMCALLBACK_UNKNOWN, rpm.RPMCALLBACK_UNPACK_ERROR,
                      rpm.RPMCALLBACK_CPIO_ERROR):
         reasons = {rpm.RPMCALLBACK_UNKNOWN: 'Unknown error',
                    rpm.RPMCALLBACK_UNPACK_ERROR: 'Unpacking error',
                    rpm.RPMCALLBACK_CPIO_ERROR: 'CPIO error'}
         self._log('ERROR in RPM Transaction processing %s: %s' %
                   (key, reasons[reason]))
         self.errors[key] = reasons[reason]

def ProblemTuplestoStrings(problems):
   '''Process tuples returned by ts.check() method into friendlier strings.
      Parameters:
         problems - as returned by ts.check().
      Returns: A list of strings, one for each problem found.
   '''
   messages = list()
   for (N, V, R), (reqN, reqV), needsFlags, suggestedPackage, sense in problems:
      msg = '%s-%s-%s' % (N, V, R)
      # Conflict or requires?
      if sense == rpm.RPMDEP_SENSE_CONFLICTS:
         msg += ' conflcts with'
      elif sense == rpm.RPMDEP_SENSE_REQUIRES:
         msg += ' requires'
      else:
         msg += ' has unknown relationship to'
      msg += ' %s' % reqN
      # <, <=, =, >=, >, or none of the above?
      op = ''
      if needsFlags & rpm.RPMSENSE_GREATER:
         op += '>'
      elif needsFlags & rpm.RPMSENSE_LESS:
         op += '<'
      if needsFlags & rpm.RPMSENSE_EQUAL:
         op += '='
      if op and reqV:
         msg += ' %s %s' % (op, reqV)
      messages.append(msg)
   return messages

def PackageTransaction(vibs, removes=(), progress=None, test=False,
                       nodeps=False, olderversion=False):
   '''Install RPMs from a VibCache or other collection of VIBs.  Accepts the
      following parameters:
      * vibs     - A VIBCollection instance.
      * removes  - A list of package names to remove from the system.
      * progress - If specified, must be an object that supports the following
                   methods:
                   * start()    - Called at the beginning of download with no
                                  arguments.
                   * set()      - Takes two arguments, the first is the amount
                                  completed and the second is the total.
                   * show()     - Takes no arguments, and causes the progress
                                  to be shown on screen.
                   * setTopic() - Takes one argument, the name of the file
                                  being installed/removed.
                   * stop()     - Takes no arguments, called at finish.
      * test     - If True, just run a test transaction to determine if it
                   would fail or not (unsatisfied deps, disk space, etc.)
      * nodeps   - If True, skip RPM dependency checks.
      * olderversion - If True, allow installation of older packages.
      May raise rpm.error or OSError.  These should probably later be
      caught and raised as a higher-level exception.
   '''
   ts = rpm.ts()
   ts.setVSFlags(rpm._RPMVSF_NOSIGNATURES)
   if olderversion:
      flags = ts.setProbFilter(0)
      flags |= rpm.RPMPROB_FILTER_OLDPACKAGE
      ts.setProbFilter(flags)
   #
   # PR 321752: all RPM output, including pre and post scripts, must be logged
   # so that stdout XML output is not affected for --HA mode.  Make sure
   # to unset the log file so RPM won't hang on to a bad file pointer later.
   #
   ts.scriptFd = rpmlogfile.fileno()
   for vib in vibs.values():
      if vib.packagelocation:
         fd = os.open(vib.packagelocation, os.O_RDONLY)
         hdr = ts.hdrFromFdno(fd)
         os.close(fd)
         ts.addInstall(hdr, vib, 'u')
   for pkg in removes:
      try:
         ts.addErase(pkg)
      except rpm.error, e:
         log.info('Cannot erase package %s: %s' % (pkg, e))
   # We must always call ts.check() before ts.order(), at least in RPM versions
   # before 4.2.2.3.  Failing to do so means the transaction's
   # "available packages" index is not properly created, and the order() method
   # will not properly order items based on dependencies.
   problems = ts.check()
   if problems and not nodeps:
      problemstr = '; '.join(ProblemTuplestoStrings(problems))
      raise errors.PackageManagerError(vibs.values(),
                                       'Problems found in RPM transaction: %s'
                                       % problemstr)
   ts.order()
   try:
      # Note re-use of 'problems' variable here.  ts.check() only catches
      # dependency problems.  All other issues are given in return value of
      # ts.run().
      cb = RpmTransactionCb(progress, rpmlogfile)
      problems = ts.run(cb, 1)
   except rpm.error, e:
      raise errors.PackageManagerError(vibs.values(),
                                       'Error running RPM transaction: %s' %
                                       str(e))
   if problems:
      # 'problems' is a list of complex tuples.  Each tuple in the list looks
      # something like:
      # (prob_string, (prob_type, type_specific_str, type_specific_ulong))
      # The meaning of type_specific_str and type_specific_ulong vary depending
      # on what prob_type is.  We don't want to handle specific problems
      # differently (at least for now), so we just aggregate the user-friendly
      # strings and put them in the exception.
      probstr = '; '.join([problem[0] for problem in problems])
      raise errors.PackageManagerError(vibs.values(),
                                       'Error running RPM transaction: %s' %
                                       probstr)
   if cb.errors:
      # Callback received one or more error reasons:  Let's report them.
      errorstr = 'Errors reported by RPM transaction callback: '
      errorstr += ', '.join(['(%s: %s)' % e for e in cb.errors.items()])
      raise errors.PackageManagerError(vibs.values(), errorstr)
   # Validate that the things we expected to be installed were.
   failed_installs = list()
   failed_removes = list()
   tenevra = lambda t: (t.N(), str(t.E()), t.V(), t.R(), t.A())
   hdrnevra = lambda h: (h[rpm.RPMTAG_NAME], str(h[rpm.RPMTAG_EPOCH]),
                         h[rpm.RPMTAG_VERSION], h[rpm.RPMTAG_RELEASE],
                         h[rpm.RPMTAG_ARCH])
   for te in ts:
      mi = ts.dbMatch(rpm.RPMTAG_NAME, te.N())
      found = tenevra(te) in (hdrnevra(h) for h in mi)
      if not found and te.Type() == rpm.TR_ADDED:
         failed_installs.append(te.NEVR())
      elif found and te.Type() == rpm.TR_REMOVED:
         failed_removes.append(te.NEVR())

   if failed_installs or failed_removes:
      details = ''
      if failed_installs:
         details = ' Installation failed for %s.' % ', '.join(failed_installs)
      if failed_removes:
         details += ' Removal failed for %s.' % ', '.join(failed_removes)
      raise errors.PackageManagerError(vibs.values(),
                                       'An unknown RPM error prevented '
                                       'installation or removal of one or more '
                                       'packages.%s' % details)

def CreateESXChannel(name, metabase, metatype, paths=None):
   """ Create Smart channel for metadata or rpm db """
   channeldata = dict()
   channeldata['name'] = name
   channeldata['type'] = metatype
   channeldata['baseurl'] = metabase
   channeldata['path'] = metabase
   channeldata['recursive'] = True
   return smart.channel.createChannel(channeldata['name'], channeldata)

COS_VIBDB_DIR = '/etc/vmware/esxupdate'

def GetDbPath():
   return COS_VIBDB_DIR

def RestartHostd():
   """ Restart hostd """
   # Apparently, we don't care if this succeeds or not?
   rc, out = runcommand.runcommand(['/sbin/service', 'mgmt-vmware', 'restart'])

def PlatformInit():
   pass

def CheckCache(cachesize=None, cleancache=False, doumount=False):
   cachedir = config.GetConf().default.cachedir
   if cleancache:
      if os.path.isdir(cachedir):
         shutil.rmtree(cachedir)

   if not os.path.exists(cachedir):
      log.info('Creating cache directory...')
      os.makedirs(cachedir)

   s = os.statvfs(cachedir)
   actualsize = s.f_bfree * s.f_bsize / 1024 / 1024
   return {'size':  actualsize,
           'location': cachedir}

def GetRunningMode():
   return PLATFORMS.cos

def getsmartclasses():
   return (smart.backends.rpm.base.RPMNameProvides,
           smart.backends.rpm.base.RPMObsoletes,
           smart.backends.rpm.base.RPMPackage)

def smartnv(vib):
   """ Return a smart-compatible name, version tuple given a VIB """
   name = vib.name
   version = vib.version.versionstring
   #
   # A cross-platform VIB assumes an arch, but the arch is not added to the
   # version string.  For smart package creation on COS, we need to append
   # the arch.
   #
   if version and '@' not in version:
      version += '@' + vib.arch
   return name, version
