#######################################################################
# Copyright (C) 2008 VMWare, Inc.
# All Rights Reserved
########################################################################
#
# Visor-specific platform functions
#
import os
import tarfile
from vmware.esx45update.util import runcommand
import shutil
import logging

from vmware.esx45update import errors
from vmware.esx45update import config
from vmware.esx45update.const import *

import smart
import smart.interface
from smart.sysconfig import SysConfig

log = logging.getLogger('_visor')

#
# smart.sysconf must be intialized. Otherwise we cannot import backends.deb
# TODO: deb-arch is overrided here, we should have our own detection.
#
smart.sysconf.object = SysConfig()
ESXI_ARCH = 'x86_64'
smart.sysconf.set('deb-arch', ESXI_ARCH)

import smart.channel
import smart.cache
from smart.channel import PackageChannel
from smart.backends.deb.loader import DebTagFileLoader, DebTagLoader, DebFileChannel
from smart.util.tagfile import TagFile


#
# Visor environment variable and installation destination setup
#
BOOTBANK = config.GetConf().visor.bootbankdir
ALTBOOTBANK = config.GetConf().visor.altbootbankdir
LOCKER = config.GetConf().visor.lockerdir

BOOTBANKSCRIPT = config.GetConf().visor.helperscript

# this will also be the installation order
#
# 'root' goes into /, yes that's right.  It's intended
# for the (hopefully rare) cases when 'stage' and the
# host scripts do not work, or we just need to do
# something drastically different.
# Don't use it unless you know what you're doing.
#
IPKG_DESTINATIONS = ('root', 'stage', 'oem', 'custom', 'locker')
FIRMWARE_DESTINATIONS = ('stage', 'oem', 'custom')
IPKG_DEST_PATH_MAP = {'stage':  config.GetConf().visor.stagedir,
                      'oem' :  config.GetConf().visor.oemstgdir,
                      'custom': config.GetConf().visor.customstgdir,
                      'locker' : os.path.join(LOCKER, 'packages'),
                      'root' : '/'}

PKG_DB_DIR = 'var/db/esxupdate/'
BOOTBANK_DB_TAR = 'pkgdb.tgz'
BOOTBANK_DB_TMP = 'pkgdb.tmp'
BANK_DIR = 'bank'

CACHENAME = 'updatecache'
STAGENAME = 'updatestg'

# Staging directory is mounted as a ramdisk to support its size and symbolic
# link. It is created with a upper limit of 750MB. 
# Assuming 250MB bootbank, 50% of OME/custom vibs, and compressed/expanded ratio
# of 1:4
# oem/custom staging = 250MB * 50% * 4 = 500 MB
# bootbank image = 250 MB
# total = 500 + 250 = 750 MB
STAGE_MINMB = 0
STAGE_MAXMB = 750

# Change mod.tgz to m.z to shorten the module names.
STAGE2TARGET = {'oem': 'oem.tgz',
                'custom': 'm.z'}

#
# Set of packages which cannot be removed
IMMUTABLE_PKGS = set(('vmware-esx-firmware',))

RESTRICTED_MODE = config.GetConf().platform.restricted

# Blacklist of vibs not allowed in restricted mode
RESTRICTED_PKGS = set(('vmware-esx-firmware',
                       'vmware-esx-tools-light',
                       'vmware-esx-viclient'))

IPKG_DB_PATH = 'usr/lib/ipkg'
IPKG_VERBOSITY = 2
IPKG_CONF_ARCH = 'arch all 1\narch %s 10' % (ESXI_ARCH)

BOOTSTATE_SUCCESS = '0'
BOOTSTATE_UPDATED = '1'
BOOTSTATE_BOOT = '2'
BOOTSTATE_EMPTY = '3'
#
# Visor platform public functions
#
def PackageTransaction(vibs, removes=(), progress=None, test=False,
                       nodeps=False, olderversion=False):
   '''Installs Vibs from a VibCache using ipkg. Accepts the
      following parameters:
      * vibs     - A VIBCollection instance.
      * removes  - A list of package names to remove from the system.
      * progress - If specified, must be an object that supports the following
                   methods:
                   * start()    - Called at the beginning of download with no
                                  arguments.
                   * set()      - Takes two arguments, the first is the amount
                                  completed and the second is the total.
                   * show()     - Takes no arguments, and causes the progress
                                  to be shown on screen.
                   * setTopic() - Takes one argument, the name of the file
                                  being installed/removed.
                   * stop()     - Takes no arguments, called at finish.
      * test     - If True, just run a test transaction to determine if it
                   would fail or not (unsatisfied deps, disk space, etc.)
      * nodeps   - If True, skip RPM dependency checks.
      * olderversion - If True, allow installation of older packages.
   '''

   #
   # TODO: ipkg doesn't install the latest package, we need to select the latest
   # one in esxupdate
   #
   def _installPkg(vibs, ipkgconf, test=False):
      '''Install vibs for a single destination.'''
      cmd = []
      cmd.extend(['-f', ipkgconf])
      cmd.extend(['-d', vibs[0].visordestination])
      if test:
         cmd.append('-test')
      if olderversion:
         cmd.append('-force-downgrade')

      cmd.append('install')
      for vib in vibs:
         cmd.append(vib.packagelocation)

      details = ', '.join(vib.id for vib in vibs)
      return _invokeIpkg(cmd, topic='Installing packages',
                         details=details)

   def _invokeIpkg(args, topic, details=''):
      '''Invoke ipkg and return returncode + any error messages'''
      cmd = ['/bin/ipkg']
      cmd.extend(['-V', '%s' % (IPKG_VERBOSITY)])

      # NOTE: disable ipkg dependency checking since visor has several
      # destinations and packages are grouped and installed by destination.
      # ipkg dep checking fails when firmware has dependencies on locker,
      # since ipkg installs locker vibs after firmware destinations.
      cmd.append('-nodeps')
      cmd.append('-force-depends')
      cmd.extend(args)
      if progress:
         progress.start()
         progress.setTopic('%s :%s' % (topic, details))
         progress.set(0, 1)
         progress.show()

      log.info('%s with :%s' % (topic, ' '.join(cmd)))
      errlogger = ErrorLogger(log)
      rc, nada = runcommand.runcommand(cmd, outfile = errlogger,
                                       returnoutput = False)
      errs = errlogger.close()

      if progress:
         progress.set(1, 1)
         progress.show()
         progress.stop()

      return rc, errs

   def _removePkg(pkgnames, ipkgconf, test=False):
      '''Call ipkg remove on one or more packages.
       * pkgnames - iterable of package names
       * ipkgconf - full path to ipkg.conf
      '''
      packages = list(pkgnames)
      args = []
      args.extend(['-f', ipkgconf])
      if test:
         args.append('-test')
      args.append('remove')
      args.extend(packages)
      return _invokeIpkg(args, topic='Removing packages',
                         details=' '.join(packages))

   def _getPackageNVsFromDB(debdb):
      # Use base interface class, because we don't want anything printed on
      # stdout. Otherwise --HA XML message will be messed up.
      # TODO: As dest for a vib is also in vib DB, we should consider to use the
      # information directly rather than looking around.
      if smart.initialized():
         _control = smart.iface.getControl()
      else:
         _control = smart.init()
      smart.iface.object = smart.interface.Interface(_control)

      cache = smart.cache.Cache()
      loader = VisorTagFileLoader(debdb)
      cache.addLoader(loader)
      cache.load()
      pkgnvs = [(pkg.name, pkg.version) for pkg in cache.getPackages()]
      loader.file.close()
      return set(pkgnvs)

   # grouping vibs based on dest
   pkggrps = {}
   for dest in IPKG_DESTINATIONS:
      pkggrps[dest] = []

   firmwareUpdate = False
   for vib in vibs.values():
      if RESTRICTED_MODE and vib.name in RESTRICTED_PKGS:
         raise errors.RestrictedVibError(vibs,
               "Unable to install vib %s." %
               (vib.name))
      if vib.visordestination in IPKG_DESTINATIONS:
         pkggrps[vib.visordestination].append(vib)
         if vib.visordestination in FIRMWARE_DESTINATIONS:
            firmwareUpdate = True
      else:
         raise errors.PackageManagerError(vibs,
               "Visor destination %s for vib %s is not supported." %
               (vib.visordestination, vib.id))
   numOems = len(pkggrps['oem'])
   numCustom = len(pkggrps['custom'])

   # are there any packages to remove or upgrade?
   # We have to remove the older package for upgrades due to
   # async OEM/third party install/update requirements:
   #  - ipkg does not do conditional Replaces: (PR 333426), its a big change
   #  - ipkg can't correctly handle upgrading from one dest to another
   # Base image has a faked package status file to indicate the default
   # installed drivers. A faked package doesn't need to be removed if it
   # is replaced by some packages. But if a faked package is really obsoleted 
   # by a new package, we do need to remove it.
   pkgsToRemove = set(removes)

   hostObsoletedNVs = set([(n, v) for vib in vibs.values() for n, v in \
                           vib.hostobsoleted])

   # Older version of the same package
   obsoletedNVs = set([(n, v) for vib in vibs.values() for n, v in \
                       vib.hostobsoleted if n == vib.name])

   # Specified by Replaces field
   replacedNVs = hostObsoletedNVs - obsoletedNVs

   fakedNVs = _getPackageNVsFromDB(conf['rootDB'])
   baseReplaced = replacedNVs & fakedNVs
   if baseReplaced:
      log.info("The following packages are from base image and replaced by "
            "other vibs, but don't need to be removed: %s" % (baseReplaced))
      replacedNVs -= baseReplaced
   pkgsToRemove |= set([n for n, v in replacedNVs])

   obsnames = set([n for n, vi in obsoletedNVs])
   if obsnames:
      log.info('The following packages are obsoleted by new version: %s\n'
               'They will be added to the remove list.' % (obsoletedNVs))
      pkgsToRemove |= obsnames

   cannotRemove = pkgsToRemove & IMMUTABLE_PKGS
   if cannotRemove:
      log.info('The following packages cannot be removed: %s. '
               'They will be pruned from the remove list.' %
               (', '.join(cannotRemove)) )
      pkgsToRemove -= cannotRemove
   log.debug('removes=%s\nobsnames=%s\npkgsToRemove=%s' %
             (removes, obsnames, pkgsToRemove))

   oemDBPath = conf['oem'].GetDBFile()
   oemPkgs = set([n for n, v in _getPackageNVsFromDB(oemDBPath)])
   customDBPath = conf['custom'].GetDBFile()
   customPkgs = set([n for n, v in _getPackageNVsFromDB(customDBPath)])

   # Need to repackage oem or mod.tgz when removing OEM or custom Vib.
   if oemPkgs & pkgsToRemove:
      numOems += 1

   if customPkgs & pkgsToRemove:
      numCustom += 1

   log.debug('Number of oem/custom vibs: %d/%d' % (numOems, numCustom))
   if numOems or numCustom:
      firmwareUpdate = True

   # Update vib.conf for install transaction
   vibconfFile = VIBCONF_FILE
   if not os.path.isfile(vibconfFile):
      raise errors.VisorSetupError('vib.conf %s not found' % (vibconfFile))

   vibconf = config.EsxupdateConfig(vibconfFile)
   vibconf.set('visor', 'oemvibs', str(numOems))
   vibconf.set('visor', 'customvibs', str(numCustom))
   vibconf.set('visor', 'bootbankdir', conf['sourceBank'])
   try:
      conffp = open(vibconfFile, 'w')
      vibconf.write(conffp)
      conffp.close()
   except EnvironmentError, e:
      raise errors.VisorSetupError('Error writing vib.conf %s: %s' %
            (vibconfFile, e))

   if firmwareUpdate:
      conf.BeginBootbankUpdate()

   #
   # Remove packages first
   # Note: prerm scripts in any stage/oem/custom vibs will automatically
   # cause the right tgz's to be unpacked.
   # NOTE: should we have 'root' run first?
   if pkgsToRemove:
      ret, errs = _removePkg(pkgsToRemove, conf['ipkgconf'], test)
      if ret != 0 or errs:
         raise errors.PackageManagerError(pkgsToRemove,
            'Error (%d) running ipkg to remove packages %s:\n%s' %
            (ret, ', '.join(pkgsToRemove), errs))

   # install vibs in root, stage, and locker order
   for dest in IPKG_DESTINATIONS:
      if pkggrps[dest]:
         ret, errs = _installPkg(pkggrps[dest], conf['ipkgconf'], test)
         vibNames =  ' '.join([vib.id for vib in pkggrps[dest]])
         #
         # ipkg sometimes returns 0 for ERROR's, such as postinstall
         # script failing.  This case is important since without
         # scripts, oem.tgz/m.z (previously mod.tgz) as well as the ipkg db will not
         # be updated.
         #
         if ret != 0 or errs:
            raise errors.PackageManagerError(vibs,
                  'Error (%d) running ipkg to install vibs %s:\n%s' %
                  (ret, vibNames, errs))
         else:
            log.info('%s have been successfully installed to %s.' % (
               vibNames, dest))
            for vib in pkggrps[dest]:
               if vib.rebootrequired:
                  vib.state.add(vib.state.PENDING)
               vib.state.add(vib.state.INSTALLED)
               vib.updatetimestamp()

def CreateESXChannel(name, metabase, metatype, paths=None):
   '''Create Visor specific channel.'''
   channeldata = dict()
   channeldata['name'] = name
   channeldata['type'] = metatype
   channeldata['baseurl'] = metabase
   channeldata['path'] = metabase
   channeldata['manual'] = False
   channeldata['disabled'] = False
   channeldata['removable'] = False
   channeldata['priority'] = 0

   if metatype == 'esxi-sys':
      if isinstance(paths, list):
         channeldata['paths'] = paths
      else:
         raise ValueError('paths must a list of path for esxi-sys channel')
      return ESXiSysChannel(channeldata['type'],
            'ESXi system channel',
            channeldata["name"],
            channeldata["manual"],
            channeldata["removable"],
            channeldata["priority"],
            paths=channeldata['paths'])
   elif metatype == 'apt-deb':
      channeldata['components'] = ''
      channeldata['distribution'] = config.GetConf().visor.distribution
   elif metatype == 'deb-file':
      return DebFileChannel(metabase)
   else:
      raise ValueError('Unsupported ESXi channel :%s' % (metatype))
   return smart.channel.createChannel(name, channeldata)

def CheckCache(cachesize=None, cleancache=None, doumount=False):
   cachedir = config.GetConf().default.cachedir
   mountpath = os.path.realpath(cachedir)
   if cleancache :
      log.info('Cleaning up cache directory...')
      if _isPathMounted(mountpath):
         try:
            if doumount:
               _umount(mountpath)
         except errors.VisorSetupError, e:
            log.warning('Failed to umount cache directory: %s' % (e))
      elif os.path.isdir(cachedir):
         shutil.rmtree(cachedir)

   if not os.path.exists(cachedir):
      log.info('Creating cache directory...')
      os.makedirs(cachedir)

   #
   # If cachesize is specified as a positive int, esxupdate will try to mount a
   # ramdisk with the given size. If cachesize is 0, esxupdate will try to mount
   # ramdisk with min/max of 0/defaultcachemb. defaultcachemb is from configuration
   # NOTE: max is not guaranteed to be available.
   #
   try:
      defaultcache = int(config.GetConf().default.defaultcachemb)
      mincache = int(config.GetConf().default.mincachemb)
   except ValueError, e:
      raise errors.VisorSetupError("Failed to convert either defaultcachemb "
            "or mincachemb in config file to integer: %s" % (e))

   requestedMin = -1
   if cachesize is not None \
         and _isScratchOnTmp() and not config.GetConf().platform.offline \
         and not _isPathMounted(mountpath):
      freemem = _getFreeMemory()
      log.info('%d MB is available to mount cache ramdisk' % (freemem))

      if cachesize > 0 and freemem > cachesize:
         requestedMin = cachesize
         requestedMax = cachesize
      elif cachesize == 0:
         requestedMin = 0
         requestedMax = defaultcache

      if requestedMin >= 0:
         _mountVisorFS(CACHENAME, mountpath, requestedMin, requestedMax)

   actualsize = 0
   if _isScratchOnTmp() and not config.GetConf().platform.offline:
      if _isPathMounted(cachedir):
         actualsize = _getVisorfsFreeSize(CACHENAME)
   else:
      s = os.statvfs(cachedir)
      actualsize = s.f_bfree * s.f_bsize / 1024 / 1024

   return {'size': actualsize,
           'location': cachedir}

#
# share the same directory as ipkg DB
#
def GetDbPath():
   return os.path.dirname(conf['workingDB'])

def RestartHostd():
   pass

#
# XXX There is a race condition where two processes importing Transaction / _visor.py
#     could potentially be mounting the stage dir and unpacking the databases
#     at the same time.  Solving this problem would involve some kind of lock
#     outside of the staging area, always located in VisorFS.
#     This is not urgent today as esxupdate is called during init, so most likely
#     the setup will be done way before a race condition comes up.
#
def PlatformInit():
   conf.Setup()

def GetRunningMode():
   VISOR_CODE2BOOTTYPE = {0: 'visor-usb',
         1: 'visor-thin',
         2: 'visor-pxe' }

   if config.GetConf().platform.offline:
      return 'visor-offline'
   elif os.uname()[0].lower() == 'vmkernel':
      try:
         rc, out = runcommand.runcommand(['/sbin/esxcfg-info', '-e'])
         if rc in VISOR_CODE2BOOTTYPE:
            return VISOR_CODE2BOOTTYPE[rc]
         else:
            msg = 'Unknow return code from "esxcfg-info -e" :%s, %s' % (rc, out)
            log.warn(msg)
            raise errors.VisorSetupError(msg)
      except Exception, e:
         raise errors.VisorSetupError('Failed to run command :%s' % (e))

   raise errors.VisorSetupError('Unkown visor running mode')

#
# Visor platform classes
#
class ComponentStage(object):
   ''' Represents a staging area for assembling a bootbank component.

       <stagedir>/usr/lib/ipkg   is the ipkg database;

       Post install scripts pack up directories in <stagedir> and place
       the resulting tgzs in bank/ of <stagedir>.

       ComponentStage instances may also be created just to unpack
       and examine the ipkg database for base firmware, oem and m (mod previously)
       Useful attributes:
          stagedir  - <stagedir>
          libdir    - <stagedir>/usr/lib/ipkg
   '''
   def __init__(self, stagedir, name):
      ''' Creates a new ComponentStage instance; At the end,
          * <stagedir>/ will exist
          * <stagedir>/usr/lib/ipkg/* will be preserved (but not info/)
          * Everything else under <stagedir> will be erased
          * <stagedir>/usr/lib/ipkg/status will exist
      '''
      self.name = name
      if stagedir in [BOOTBANK, ALTBOOTBANK]:
         raise errors.VisorSetupError('%s cannot be used as a stagedir'
                                      % (stagedir))

      self.stagedir = stagedir
      self.libdir = os.path.join(stagedir, IPKG_DB_PATH)
      self.dbfile = os.path.join(self.libdir, 'status')

      if not os.path.exists(self.stagedir):
         _createEmptyDir(self.stagedir, 'Unable to initialize staging dir',
               safedirlist = (IPKG_DB_PATH,))
      # Preserve status file if it exists, o/w create an empty one
      if not os.path.exists(self.dbfile):
         _createEmptyFile(self.dbfile, 'Could not create ipkg status')

   def GetDBFile(self):
      ''' Return path to ipkg DB file if file exists'''
      if self.dbfile and os.path.isfile(self.dbfile):
         return self.dbfile
      else:
         raise errors.VisorSetupError("DB file doesn't exist for %s" %
                  (self.stagedir))

   def ValidIpkgDB(self):
      ''' Returns true if a valid ipkg database exists '''
      return os.path.exists(self.dbfile) and os.path.getsize(self.dbfile) > 0

   def SetupPkgDb(self, pkgDbTgz):
      ''' Unpack pkgDbTgz to <stagedir>, creating <stagedir>/usr/lib/ipkg.
          If pkgDbTgz is None, create an empty ipkg db.
      '''
      _createEmptyDir(self.libdir)
      if pkgDbTgz:
         try:
            pkgdbtar = tarfile.open(pkgDbTgz, 'r')
            for member in pkgdbtar:
               pkgdbtar.extract(member, path=self.libdir)
            pkgdbtar.close()
         except Exception, e:
            raise errors.VisorSetupError("Error extracting bootbank DB: %s" % e)

      # Populate database with default status if necessary.
      if self.dbfile and not os.path.exists(self.dbfile):
         log.info('No package database passed, creating an empty one')
         f = file(self.dbfile, 'w+')
         f.close()

   def GetTarDBPath(self):
      '''Return relative pkgdb tar file within a tar container

         For firmware stage destination, return pkgdb name directly. For custom
         and oem destination, the path is
         var/db/esxupdate/{STAGE_NAME}-pkgdb.tgz
      '''
      if self.name == 'stage':
         return BOOTBANK_DB_TAR
      else:
         return os.path.join(PKG_DB_DIR, self.name + '-' + BOOTBANK_DB_TAR)

   def ExtractDBTar(self, container):
      '''Try to extract pkg DB tar file from the container tar file (m.z or oem.tgz)

         The default pkg DB tar is var/db/esxupdate/(oem|custom)-pkgdb.tgz. If
         no such file in the container, try ./pkgdb.tgz, which is kl default
      '''
      dbfile = self.GetTarDBPath()
      dbtar = self._extractDb(container, self.stagedir, dbfile)
      # To be backward compatible, also search DB tar at root
      if dbtar is None:
         dbfile = BOOTBANK_DB_TAR
         dbtar = self._extractDb(container, self.stagedir, dbfile)
      return dbtar

   def UnpackPkgDb(self, dbtar):
      '''Unpack source pkg DB tar file to ipkg DB dir if there is no valid DB

         If dbtar is None, empty ipkg DB will be created.
      '''
      if not self.ValidIpkgDB():
         log.debug('Updating pkg DB for %s with %s' % (self.name, dbtar))
         self.SetupPkgDb(dbtar)
      else:
         log.debug("Staged package DB is unchanged.")

   def RepackPkgDb(self, targetDir=None):
      ''' Packs <stagedir>/usr/lib/ipkg to <stagedir>/pkgdb.tgz '''
      if not targetDir:
         targetDir = self.stagedir
      dst = os.path.join(targetDir, BOOTBANK_DB_TAR)
      dst_tmp = os.path.join(targetDir, BOOTBANK_DB_TMP)
      try:
         pkgdbtar = tarfile.open(dst_tmp, 'w:gz')
         # tarfile's gzopen and bz2open methods stupidly override the 'name'
         # parameter passed to TarFile. This causes a bug when the add() method
         # does a check to make sure it is not trying to add the tar archive to
         # itself. It calls os.path.abspath() on its own name attribute, which
         # raises an exception because the path in self.name doesn't exist.
         pkgdbtar.name = dst_tmp
         for name in os.listdir(self.libdir):
            srcname = os.path.join(self.libdir, name)
            # By default, add() method works recursively.
            pkgdbtar.add(srcname, name)
         pkgdbtar.close()
      except Exception, e:
         msg = 'Unable to tar bootbank DB to alt boot bank: %s' % e
         raise errors.DatabaseIOError(dst_tmp, msg)
      #
      # Atomic update to new pkgdb.tgz
      try:
         os.rename(dst_tmp, dst)
      except OSError, e:
         raise errors.DatabaseIOError(dst, 'Unable to rename %s to %s: %s'
                                      % (dst_tmp, dst, e))
   def _extractDb(self, compTar, dest, dbname):
      try:
         t = tarfile.open(compTar, 'r')
         t.extract(dbname, dest)
         t.close()
         return os.path.join(dest, dbname)
      except:
         return None



class VisorConf(object):
   ''' Represent visor configuration values as a dictionary

   This class uses singleton pattern and only one object can be created.
   '''

   class __impl(dict):
      def __init__(self):
         self._altbootcfgdgst = ''
         self._altbootbankstatus = BOOTSTATE_EMPTY
         self['bootbankUpdate'] = False
         self['latestDBs'] = []
         self['rootDB'] = config.GetConf().visor.rootdb
         self['version'] = config.GetConf().platform.version
         self._packagingdir = '/tmp/stage'

      def Setup(self):
         ''' Setup visor platform update environment'''
         self._packagingdir = config.GetConf().visor.packagingdir
         self.SetupStagingRamdisk(STAGE_MINMB, STAGE_MAXMB)
         self._setupIpkgDests()
         self._createIpkgConf()

      def IsAltbootbankUpdated(self):
         ''' Check whether altbootbank has been or is being updated.

         Return True if updatestate in /altbootbank/boot.cfg equals to 1 or
         altbootbank has been duplicated from current-bootbank. Updating
         latestDBs and locker ipkg conf if bootbank just changed to updated.
         '''
         return self.GetAltbootbankStatus() == BOOTSTATE_UPDATED

      def BeginBootbankUpdate(self, iface=None):
         ''' Prepare boot bank for update - format, etc '''

         # Make sure staging dirs are clean
         self.CleanupStagingdir()
         _createEmptyDir(os.path.join(self['stage'].stagedir, BANK_DIR))

         # Make sure we have enough space to stage firmware
         if not config.GetConf().platform.offline:
            freespace = _getFreeMemory()
            fwsize = _estimateFWStagingSize(self['sourceBank'])
            if freespace < fwsize:
               log.critical('Not enough free memory available to copy '
                     'bootbank image.\nAt least %sMB is required, but only '
                     '%sMB is available' % (fwsize, freespace))
               raise errors.VisorSetupError('Not enough free memory available '
                     'to stage firmware image.')

         #
         # Invoke script to prep/format target boot bank
         # In case of 3i -> 4i upgrade, altbootbank might have
         # to be repartitioned since it doubled in size.
         #
         cmd = [BOOTBANKSCRIPT, '-v', 'prep']
         log.info('Running [%s]...' % (cmd))
         errlogger = ErrorLogger(log, iface)
         rc, nada = runcommand.runcommand(cmd, outfile = errlogger,
                                          returnoutput = False)
         errlogger.close()
         if rc != 0:
            raise errors.VisorSetupError('Bootbank preparation failed: %s '
                                         'returned (%d)' % (cmd, rc))

         self['bootbankUpdate'] = True

      def UpdateBootbank(self, iface=None):
         '''Write pkgdb changes to new boot bank, then
            call a script to copy remaining files, edit
            boot config, finish preparing new boot image.
         '''
         if self['bootbankUpdate']:
            self['stage'].RepackPkgDb(os.path.join(self['stage'].stagedir, BANK_DIR))
            cmd = [BOOTBANKSCRIPT, '-v', 'install']

            # TODO: add progress bar here.  This might take a really
            # long time, esp writing to USB.
            log.info('Running [%s]...' % (cmd))
            errlogger = ErrorLogger(log, iface)
            rc, nada = runcommand.runcommand(cmd, outfile = errlogger,
                                             returnoutput = False)
            errlogger.close()
            if rc != 0:
               log.error('Error writing to boot bank, see log for details')
               raise errors.PostScriptError(cmd, rc)
         else:
            log.info('Boot bank does not need to be updated.')
            self.UpdateStagedDb()

      def UpdateStagedDb(self):
         '''Repacks pkgdb.tgz to source bank so next run of
            esxupdate can benefit from staging of Vibs
         '''
         self['stage'].RepackPkgDb(targetDir=self['sourceBank'])

      def GetAltbootbankStatus(self):
         '''Return bootstate value in /altbootbank/boot.cfg'''
         altcfgfile = os.path.join(ALTBOOTBANK, 'boot.cfg')
         #
         # boot.cfg is not guaranteed to exist.
         if not os.path.exists(altcfgfile):
            self._altbootbankstatus = BOOTSTATE_EMPTY
            return self._altbootbankstatus

         # package install script might change the boot.cfg
         dgst = os.path.getmtime(altcfgfile)
         if dgst != self._altbootcfgdgst:
            self._altbootcfgdgst = dgst
            fp = open(altcfgfile)
            try:
               for line in fp:
                  if line.startswith('bootstate='):
                     self._altbootbankstatus = line[len('bootstate='):].strip()
                     break
            finally:
               fp.close()

            if self._altbootbankstatus not in (BOOTSTATE_SUCCESS,
                  BOOTSTATE_UPDATED, BOOTSTATE_BOOT, BOOTSTATE_EMPTY):
               self._altbootbankstatus = BOOTSTATE_EMPTY

         return self._altbootbankstatus


      def SetupStagingRamdisk(self, minMB, maxMB):
         mountpath = os.path.realpath(self._packagingdir)
         if not config.GetConf().platform.offline \
               and not _isPathMounted(mountpath):
            if not os.path.exists(self._packagingdir):
               os.makedirs(self._packagingdir)
            _mountVisorFS(STAGENAME, mountpath, minMB, maxMB)

      def CleanupStagingdir(self, preservedb = True):
         # XXX Due to PR 352051, don't try to umount stage dir for 4.0 host
         # Umount looks successful, but the handle is still there and remount
         # will fail.
         #
         # PR 454685: Do not umount the stage dir at all, we need to leave
         #            ipkg status file & esxupdate database file intact.
         #            Just delete all other files within each stage dir.
         #
         safedirs = ()
         if preservedb:
            safedirs = [self[dest].libdir for dest in FIRMWARE_DESTINATIONS if
                  dest in self]
            safedirs += [ os.path.join(self[dest].libdir, 'info') for dest 
                  in FIRMWARE_DESTINATIONS if dest in self]

         _createEmptyDir(self._packagingdir, 'Could not clean up staging dir',
                         safedirlist = safedirs)

      def Revertdb(self):
         ''' Revert the databases from sourcebank.  The existing in-memory
             ipkg and esxupdate databases will be overwritten.
             Please make sure database is read-locked when calling this method.
         '''
         log.info('Reverting all databases from %s' % self['sourceBank'])
         self.CleanupStagingdir(preservedb=False)

         for dest in FIRMWARE_DESTINATIONS:
            if dest in self:
               del self[dest]

         self._setupIpkgDests()

      def _setupIpkgDests(self):
         ''' Set up ipkg db for both bootbank and locker, and esxupdate db.
             Populates sourceBank, workingDB, and latestDBs.
         '''
         #
         # workingDB in VisorConf will always be updated by ipkg and esxupdate.
         # workingDB is based on the lasted pkgdb.tgz in the two bootbanks.
         #
         if self.IsAltbootbankUpdated():
            self['sourceBank'] = ALTBOOTBANK
         else:
            self['sourceBank'] = BOOTBANK

         log.info('Set up staging area for stage')
         self['stage'] = ComponentStage(IPKG_DEST_PATH_MAP['stage'],
               'stage')

         sourcePkgDb = os.path.join(self['sourceBank'], self['stage'].GetTarDBPath())
         if not os.path.isfile(sourcePkgDb):
            # Default firmware DB tar in s.z (sys.vgz)
            sourcePkgDb = os.path.join('/', PKG_DB_DIR, self['stage'].GetTarDBPath())
            log.info('Bootbank DB not found, trying %s' % (sourcePkgDb))
         self['stage'].UnpackPkgDb(sourcePkgDb)

         self['workingDB'] = self['stage'].GetDBFile()
         self['latestDBs'] = [ self['workingDB'] ]

         # don't need to setup oem and custom dest for visor-pxe
         mode = GetRunningMode()
         for firmwaredest in ('oem', 'custom'):
            log.info('Setup staging area for %s' % (firmwaredest))
            self[firmwaredest] = ComponentStage(
                  IPKG_DEST_PATH_MAP[firmwaredest], firmwaredest)
            if mode in ('visor-offline', 'visor-usb', 'visor-thin'):
               srcfile = os.path.join(self['sourceBank'], STAGE2TARGET[firmwaredest])

               # Since we are updating from 4.0 to 4.5 and 4.0 will not have
               # m.z. Try mod.tgz as well, if m.z is not found.
               if firmwaredest == 'custom':
                  if not os.path.exists(srcfile):
                     log.info('m.z not found, using mod.tgz instead.')
                     srcfile = os.path.join(self['sourceBank'], 'mod.tgz')

               dbtar = self[firmwaredest].ExtractDBTar(srcfile)
            elif mode == 'visor-pxe':
               # Try default pkgdb tar in root file system
               dbtar = os.path.join('/', self[firmwaredest].GetTarDBPath())
               if not os.path.isfile(dbtar):
                  dbtar = None

            self[firmwaredest].UnpackPkgDb(dbtar)
            self['latestDBs'].append(self[firmwaredest].GetDBFile())


         # Default DB don't need to update
         rootDB = config.GetConf().visor.rootdb
         if not os.path.isfile(rootDB):
            _createEmptyFile(rootDB)
         self['latestDBs'].append(rootDB)


         #
         # The ipkg database in the locker is not used by esxupdate.
         #
         lockerDB = os.path.join(IPKG_DEST_PATH_MAP['locker'],
                                 IPKG_DB_PATH, 'status')
         if not os.path.isfile(lockerDB) and \
               mode in ('visor-offline', 'visor-usb', 'visor-thin'):
            _createEmptyFile(lockerDB)
         if os.path.isfile(lockerDB):
            self['latestDBs'].append(lockerDB)
         log.debug('Latest DBs are :%s' %( ','.join(self['latestDBs'])))

      def _createIpkgConf(self):
         '''Create ipkg conf file for dest destname, which contains dests'''
         ipkgconf = '\n'.join('dest %s %s' % (dest, IPKG_DEST_PATH_MAP[dest])
               for dest in IPKG_DEST_PATH_MAP)
         destfile = config.GetConf().visor.ipkgconf

         try:
            fp = open(destfile, mode='w+')
            fp.write(ipkgconf)
            fp.write('\n' + IPKG_CONF_ARCH)
            fp.close()
         except EnvironmentError, err:
            raise errors.VisorSetupError('Unable to create ipkg configuration file at'
               ' %s :%s' % (destfile, err))

         self['ipkgconf'] = destfile

   __instance = None
   def __init__(self):
      if VisorConf.__instance is None:
         VisorConf.__instance = VisorConf.__impl()
         self.__dict__['_Singleton__instance'] = VisorConf.__instance

   def __getattr__(self, attr):
      ''' Delegate access to implementation '''
      return getattr(self.__instance, attr)

   def __setattr__(self, attr, value):
      ''' Delegate access to implementation '''
      return setattr(self.__instance, attr, value)

   def __getitem__(self, key):
      return self.__instance[key]


class VisorTagFileLoader(DebTagFileLoader):
   ''' Override DebTagFileLoader __init__ method so we can
       close files we open.  This is important for
       working with ipkg, who will update the status file
       we read.  On some file systems (VFAT), opening a file
       locks it from writes by others.
   '''
   def __init__(self, filename, baseurl=None):
      DebTagLoader.__init__(self, baseurl)
      self._filename = filename
      self.file = open(filename, 'r')
      self._tagfile = TagFile(self.file)


class ESXiSysChannel(PackageChannel):
   ''' Represent visor ipkg database

   Basically the same as deb_sys channel, except it might have multiple loaders
   '''
   def __init__(self, *args, **kw):
      super(ESXiSysChannel, self).__init__(*args)
      self._paths = kw['paths']

   def fetch(self, fetcher, progress):
      mtimes = []
      for path in self._paths:
         mtimes.append(str(os.path.getmtime(path)))

      digest = ''.join(mtimes)
      if digest == self._digest:
         return True

      self.removeLoaders()
      for path in self._paths:
         loader = VisorTagFileLoader(path)
         loader.setInstalled(True)
         loader.setChannel(self)
         self._loaders.append(loader)

      self._digest = digest
      return True

   def removeLoaders(self):
      ###
      # Close open status file objects first so
      # use of status file can be freed up
      # NB: This should really be done in smart
      #
      for loader in self._loaders:
         log.debug("Closing tag file " + loader._filename)
         loader.file.close()

      PackageChannel.removeLoaders(self)


#
# Visor platform helper functions
#
def _createEmptyDir(dst, errmsg='Could not create empty dir', safedirlist=[]):
   '''If dst already exists, just remove everything inside, except
      for files in safedirlist.  dst itself is not removed.
      If dst does not exist, create it.
      * safedirlist - a list of paths relative to dst, or full paths. Subdirs
                      under each path are not protected, so list
                      them separately.
   '''
   try:
      if os.path.isdir(dst):
         log.debug('Cleaning up existing dir %s' % (dst))
         safedirfullpaths = [os.path.join(dst, p) for p in safedirlist]
         for path, dirs, files in os.walk(dst, topdown=False):
            for name in dirs:
               try:
                  os.rmdir(os.path.join(path, name))
               except OSError:
                  # ignore non-empty errors - files may have been excepted
                  if not safedirlist:
                     raise
            if path in safedirlist or path in safedirfullpaths:
               continue
            for file in files:
               os.unlink(os.path.join(path, file))
      else:
         os.makedirs(dst)
   except EnvironmentError, e:
      raise errors.VisorSetupError('%s %s: %s' % (errmsg, dst, e))

class ErrorLogger:
   def __init__(self, log, iface=None):
      self.log = log
      self.iface = iface
      self.partialline = ''
      self.errs = list()

   def _handleline(self, line):
      if 'ERROR' in line:
         self.errs.append(line)
      log.info(line)
      if self.iface:
         self.iface.ShowMessage(line)

   def write(self, msg):
      if not msg:
         return
      msg = self.partialline + msg
      lines = msg.splitlines()
      if not msg.endswith('\n'):
         self.partialline = lines[-1]
         del lines[-1]
      else:
         self.partialline = ''
      for line in lines:
         self._handleline(line)

   def close(self):
      self._handleline(self.partialline)
      return '\n'.join(self.errs)

def _createEmptyFile(dst, errmsg='Could not create empty file'):
   try:
      if not os.path.exists(os.path.dirname(dst)):
         _createEmptyDir(os.path.dirname(dst))
      f = file(dst, 'w+')
      f.close()
   except EnvironmentError, e:
      raise errors.VisorSetupError('%s %s: %s' % (errmsg, dst, e))


def _isScratchOnTmp():
   #
   # Only scratch is not configured, scratch will link to /tmp
   #
   return os.path.realpath('/scratch') == os.path.realpath('/tmp/scratch')

def _isPathMounted(mountpoint):
   rc, out = runcommand.runcommand('mount')
   if rc == 0:
      return mountpoint in out
   else:
      raise errors.VisorSetupError('Unable to check mount point % '
            'with "mount" command :%s, return code :%s' %
            (mountpoint, out, rc))

def _getFreeMemory():
   cmd = ['/sbin/vsish', '-e', '-b', '-r', 'cat',
          '/sched/groups/0/stats/capacity']
   rc, out = runcommand.runcommand(cmd)
   if rc != 0:
      log.error('vsish returned nonzero status %d\nOutput:\n%s' % (rc, out))
      raise errors.VisorSetupError(
            "Unable to get unreserved memory size and can't go ahead to "
            "mount cachedir")

   for line in out.splitlines():
      tokens = line.split(':')
      if tokens[0].strip() == 'memUnreserved':
         try:
            unreserved = int(tokens[1].strip(), 16)
            log.info('Unreserved memory is %d [MB]' % (unreserved))
            return unreserved
         except ValueError, e:
            raise errors.VisorSetupError(
                  "Unable to convert memUnreserved value '%s' to integer"
                  % (tokens[1].strip()))

   raise errors.VisorSetupError("Unable to determine the unreserved memory "
         "because (vsish -e -r /sched/groups/0/stats/capacity) didn't return "
         "memUnreserved value.")

def _getPythonobjFromVsish(cmd):
   log.debug('Running vsish command : %s' % (cmd))
   rc, out = runcommand.runcommand(cmd)
   if rc != 0:
      log.error('vsish returned nozero status %d\nCommand:\n%s\nOutput:\n%s'
                % (rc, cmd, out))
      raise errors.VisorSetupError('%s exited with non-zero code :%d'
                                   % (cmd, rc))
   try:
      pyobj = eval(out)
   except Exception, e:
      raise errors.VisorSetupError('%s parsing vsish ouput:\nOutput:\n%s' % (
         e, out))
   return pyobj

def _estimateFWStagingSize(bootbankpath):
   '''Estimate the space required in MB to stage firmware for update

      - all the files from source bootbank
      - space to expand oem.tgz, m.z, cimstg.tgz; assuming 1:4 ratio here
      - 5 MB padding for config migration
   '''
   compressratio = 4
   vfs = os.statvfs(bootbankpath)
   fwsize = (vfs.f_blocks - vfs.f_bfree) * vfs.f_bsize

   # cimstg.tgz is used for CIM compilation.
   for name in STAGE2TARGET.values() + ['cimstg.tgz',]:
      fn = os.path.join(bootbankpath, name)
      if os.path.isfile(fn):
         fwsize += (compressratio * os.path.getsize(fn))

   return fwsize / 1024 / 1024 + 5

def _getVisorfsFreeSize(name):
   cmd = ['/sbin/vsish', '-e', 'set', '/sched/groupPathNameToID', 'host',
          'system', 'kernel', 'kmanaged', 'visorfs', name]
   log.debug('Running %s to determine group ID for ramdisk %s' % (cmd, name))
   rc, out = runcommand.runcommand(cmd)
   if rc != 0:
      log.error('vsish returned nozero status %d\nCommand:\n%s\nOutput:\n%s'
                % (rc, cmd, out))
      raise errors.VisorSetupError('Unable to get the groupID for visorf %s'
                                   % name)
   cachegid = out.split()[0]

   memstatCmd = ['/sbin/vsish', '-e', '-p', 'cat',
                 '/sched/groups/%s/stats/memoryStats' % cachegid]
   memstat = _getPythonobjFromVsish(memstatCmd)
   capacityCmd = ['/sbin/vsish', '-e', '-p', 'cat',
                  '/sched/groups/%s/stats/capacity' % cachegid]
   capacity = _getPythonobjFromVsish(capacityCmd)

   consumed = 0
   totalMem = 0
   if 'consumed' in memstat and 'memReserved' in capacity \
         and 'memUnreserved' in capacity:
      try:
         totalMem = int(capacity['memReserved']) + int(capacity['memUnreserved'])
         consumed = (int(memstat['consumed']) + 1023) / 1024  #Round up to MB
      except ValueError, e:
         log.error('Failed to convert visorfs stats to integer %s' % (e))
         raise errors.VisorSetupError('Unable to convert either consumed, '
               'memReserved, or memUnreserved to integer')
   else:
      log.error('Unable to determine either consumed , memReserved, or '
            'memUnreserved from vsish ouput:\n%s\n%s' % (str(memstat),
               str(capacity)))
      raise errors.VisorSetupError("Unable to determine visorfs free space "
            "with vsish commands")
   log.debug('ramdisk %s total/consumed (MB) :%d/%d' % (name, totalMem, consumed))
   return totalMem - consumed

def _mountVisorFS(name, mountpoint, minMB, maxMB=None):
   if maxMB is None:
      maxMB = minMB
   log.info('Mounting visorfs %d/%d MB on %s' % (minMB, maxMB, mountpoint))
   # Apparently, mount is a built-in to the busybox shell, so we use a string
   # here to ensure it is passed to a shell.
   cmd = 'mount -t visorfs -o %d,%d,01777,%s %s %s' % (
         minMB, maxMB, name, name, mountpoint)
   rc, out = runcommand.runcommand(cmd)
   if rc != 0:
      log.error('mount returned nonzero status %d\nCommand:\n%s\nOutput:\n%s' %
            (rc, cmd, out))
      raise errors.VisorSetupError(
            'Unable to mount visorfs. Mount returned error (%d). Please see '
            'syslog for more details.' % (rc))

def _umount(mountpoint):
   log.info('Umounting %s...' % (mountpoint))
   # umount is a built-in to the busybox shell, so do this via shell (i.e. pass
   # as string, not list).
   rc, out = runcommand.runcommand('umount %s' % mountpoint)
   if rc != 0:
      log.warn('Failed to umount %s\numount returned nozero status (%d)\n'
               'Output: %s' % (mountpoint, rc, out))
      raise errors.VisorSetupError('Failed to umount: %s, returned :%s' %
            (mountpoint, rc))
   log.info('Umount done.')

def getsmartclasses():
   return (smart.backends.deb.base.DebNameProvides,
           smart.backends.deb.base.DebUpgrades,
           smart.backends.deb.base.DebPackage)

def smartnv(vib):
   """ Return a smart-compatible name, version tuple given a VIB """
   name = vib.name
   version = vib.version.versionstring
   return name, version

#
# Initialize visor system configuration
#
conf = VisorConf()
