#!/usr/bin/python
#######################################################################
# Copyright (C) 2008 VMWare, Inc.
# All Rights Reserved
########################################################################
#
# scan.py
#   PkgScanner, scan functions
#

import smart, smart.cache, smart.channel, smart.interface
import platform
from const import *
import config
import logging

log = logging.getLogger('scan')


class Constraint(object):
   '''A class representing the constraints, such as a dependency or
      conflict, of a package.  It defines the following attributes:
      * key          - a string describing the constraint, in either
                       "name" or "name relation version" form.  The former
                       is used if the constraint is not versioned.
      * name         - a file name or package name
      * relation     - comparison operator, one of '<', '<=', '=', '>=', '>'
      * version      - if specified, version of package to compare
      * packages     - A PkgList instance of the packages matching the
                       constraint.
   '''

   def __init__(self, name, relation='', version=''):
      if not name:
         raise TypeError("Name cannot be '%s'" % name)
      self.name = name
      self.relation = relation
      self.version = version
      self.packages = PkgList()
      self.pkgset = set()

   def _key(self):
      if self.version:
         return "%s %s %s" % (self.name, self.relation, self.version)
      else:
         return self.name

   key = property(_key)
   __str__ = _key

class ConstraintList(list):
   pass

class Pkg(object):
   '''A class representing a package, with the following attributes:
      * obsoletes - A ConstraintList instance representing packages obsoleted
                    by this one.
      * requires  - A ConstraintList instance representing packages satisfying
                    requirements of this one.
      * conflicts - A ConstraintList instance representing packages conflicting
                    with this one.
      * installed - A boolean indicating whether or not the packages is
                    installed.
   '''
   def __init__(self, key, name='', version=''):
      if not key:
         raise ValueError("Key cannot be '%s'" % key)
      self._key = key
      self.name = name
      self.version = version
      self.obsoletes = ConstraintList()
      self.requires = ConstraintList()
      self.conflicts = ConstraintList()
      self.newerpkgs = set()
      self.installed = True

   _getkey = lambda self: self._key

   key = property(_getkey)
   __str__ = _getkey
   __repr__ = lambda self: 'Pkg <%s>' % (self._key)
   __hash__ = lambda self: hash(self._key)

class PkgList(list):
   pass

class PkgScanner(object):
   '''A cross-platform package scanner.
      Useful instance attributes:
      * reqcache     - A dict of the Constraint instance for each dependency,
                       useful for quickly indexing the providing packages
      * requiredby   - A dict of the set of installed packages needing each
                       dependency
      * replaceddeps - A dict of the set of potentially incompatible replacements
                       for installed providers of a dependency. A replacement
                       is incompatible if any of the installed package provides
                       are missing from the replacement.  The keys are the set
                       of deps on the host that could be potentially broken by
                       any replacement package that has been scanned.
   '''
   # This needs to be a class variable.  There can be only one.
   _control = smart.init()

   def __init__(self, name=None, metatype=None, metabase=None,
                usesyschannel=True):
      '''Constructer for PkgScanner.  Accepts following arguments:
         * name          - The name of an initial channel to add to this
                           instance.  If specified, metatype is also
                           required.
         * metatype      - The channel type for the initial channel.  If
                           specified, name parameter must also be specified.
                           Must be a valid smart channel type.
         * metabase      - The base URL for channel metadata.  Not required
                           for system package databases.
         * usesyschannel - If True, default system package database channel
                           is automatically added.
      '''
      # Use base interface class, because we don't want anything printed on
      # stdout.  Anything that requires interactive input will die.
      smart.iface.object = smart.interface.Interface(self._control)
      # Remove channels created from config files.
      for channel in self._control.getChannels():
         self._control.removeChannel(channel.getAlias())
      self._channels = list()
      if usesyschannel:
         if config.GetPlatform() == PLATFORMS.cos:
            self.AddChannel('RPM Database', None, 'rpm-sys')
         elif config.GetPlatform() == PLATFORMS.visor:
            dbpaths = platform.conf['latestDBs']
            self.AddChannel('ipkg Database', None, 'esxi-sys', dbpaths=dbpaths)

      if name and metatype:
         self.AddChannel(name, metabase, metatype)

      self.reqcache = {}
      self.requiredby = {}    # only maps reqs from installed pkgs
      self.replaceddeps = {}

   def AddChannel(self, name, metabase, metatype, dbpaths=None):
      '''Add appropriate channel for this PkgScanner.
         Arguments:
         * name     - The channel name.
         * metabase - The URL base of the metadata (may be None for sys* types.
         * metatype - The type of channel.  Must be a valid smart channel type.
      '''
      channel = platform.CreateESXChannel(name, metabase, metatype,
                                          paths=dbpaths)
      self._channels.append(channel)

   def AddVibCollectionChannel(self, viblist):
      '''Add a channel from a VIBCollection'''
      channel = VibCollectionChannel('viblist', '', 'viblist', vibs=viblist)
      self._channels.append(channel)

   # channels is specifically a read-only property
   channels = property(lambda self: self._channels,
                       doc='The list of channels associated with this scanner.')

   def __add__(self, other):
      '''Returns a new scanner, concatenating the list of channels from
         the two sides of the operand.'''
      if not isinstance(other, PkgScanner):
         raise TypeError("Type '%s' not supported for '+' operand" %
                         other.__class__)
      newobj = PkgScanner(usesyschannel=False)
      for channel in self.channels + other.channels:
         if not channel.getName() in [ch.getName() for ch in newobj.channels]:
            newobj._channels.append(channel)
      return newobj

   def Load(self):
      '''Load metadata from associated channels.  If any channel fails,
         raises exception.'''
      # We must specify channels parameter, otherwise smart defaults to
      # re-loading channels from config files.  Also must specify
      # caching=NEVER, otherwise validation fails if local file is not cached.
      result = self._control.reloadChannels(channels=self._channels,
                                            caching=smart.const.NEVER)
      if not result:
         # Todo:  Raise more intelligent error here.
         raise Exception('Downloading metadata failed')

   @staticmethod
   def _IterSmartProvides(provideslist):
      for provide in provideslist:
         for package in provide.packages:
            yield package

   def _MkConstraint(self, smartdep):
      constraint = Constraint(smartdep.name,
                              smartdep.relation,
                              smartdep.version)
      constraint.packages = [self._SmartPkgToPkg(pkg) for pkg in
                             self._IterSmartProvides(smartdep.providedby)]
      return constraint

   def _MkRequires(self, smartdep, pkg):
      key = str(smartdep)
      #
      # Cache the dependencies in reqcache for efficiency
      if key not in self.reqcache:
         constraint = self._MkConstraint(smartdep)
         constraint.pkgset = set(constraint.packages)
         self.reqcache[key] = constraint
      else:
         constraint = self.reqcache[key]

      if pkg.installed:
         if key in self.requiredby:
            self.requiredby[key].add(pkg)
         else:
            self.requiredby[key] = set((pkg,))

      return constraint

   def _MkObs(self, smartdep, pkg):
      c = self._MkConstraint(smartdep)
      for oldpkg in c.packages:
         oldpkg.newerpkgs.add(pkg)
      return c

   def _SmartPkgToPkg(self, smartpkg):
      pkgkey = str(smartpkg)
      if pkgkey in self._pkglistcache:
         return self._pkglistcache[pkgkey]
      newpkg = Pkg(pkgkey, smartpkg.name, smartpkg.version)
      self._pkglistcache[pkgkey] = newpkg
      self._pkglist.append(newpkg)
      newpkg.installed = smartpkg.installed
      newpkg.requires = [self._MkRequires(r, newpkg) for r in smartpkg.requires]
      newpkg.obsoletes = [self._MkObs(o, newpkg) for o in smartpkg.upgrades]
      #
      # PR 318869: smart returns a "conflict" for every
      # explicit Obsoletes:/Replaces:, so we have to filter
      # them out.
      #
      newpkg.conflicts = [self._MkConstraint(c) for c in smartpkg.conflicts \
                          if c not in smartpkg.upgrades]
      return newpkg

   def RemoveAllChannels(self):
      '''Remove all the channels associated with PkgScanner'''
      for channel in self._channels:
         if isinstance(channel, smart.channel.PackageChannel):
            channel.removeLoaders()

   def ComputeBreakages(self):
      '''Computes all the potentially incompatible upgrade/downgrades
      '''
      log.debug('# of elems in reqcache = %d' % (len(self.reqcache)))
      log.debug('# of elems in requiredby = %d' % (len(self.requiredby)))

      upgradeddeps = {}
      for hostreq, pkgs in self.requiredby.items():
         # if the packages which require this hostreq have newerpkgs
         # then the hostreq is not broken
         # if all([True for x in pkgs if x.newerpkgs]): continue
         newpkgs = 0
         for pkg in pkgs:
             if pkg.newerpkgs:
                 newpkgs += 1
         if newpkgs == len(pkgs):
             continue
         #
         # List any upgrade/downgrades that do not satisfy an existing
         # host dependency.
         #
         for provider in self.reqcache[hostreq].packages:
            if not provider.installed:
               continue
            replacements = set(provider.newerpkgs)
            for obs in provider.obsoletes:
               replacements.update(set(obs.packages))
            #
            # Of the packages this provider could be upgraded/downgraded to,
            # which ones do not provide the host requirement?
            #
            incompatibles = replacements - self.reqcache[hostreq].pkgset
            if incompatibles:
               upgradeddeps.setdefault(hostreq, set()).update(incompatibles)

      self.replaceddeps = upgradeddeps

   def Scan(self):
      '''Returns a list of Pkg objects, complete with their requires,
         obsoletedby, conflicts, etc.'''
      cache = self._control.getCache()
      smartpkgs = cache.getPackages()
      self._pkglist = PkgList()

      # This populates _pkglist.  _pkglistcache is a dict that is used
      # temporarily so that we can do lookups efficiently.  It is explicitly
      # deleted once we're finished populating _pkglist.
      self._pkglistcache = dict()
      self.reqcache.clear()
      self.requiredby.clear()    # only maps reqs from installed pkgs
      self.replaceddeps.clear()
      for smartpkg in smartpkgs:
         self._SmartPkgToPkg(smartpkg)
      del self._pkglistcache
      # We no longer need the smart package list.  Deleting it frees up some
      # more memory.
      cache.reset()

      self.ComputeBreakages()

      return self._pkglist


class VibCollectionLoader(smart.cache.Loader):
   """ Populates smart cache/package dict from a dictionary of Vib objects """
   #
   # Known Issues:
   # 3. Scan is slow when installed Vibs are in vibdb, two instances of a smart Package are created
   #    (One from RPM and one created by this class)
   #    Does smart merge them or return as two separate things?
   #      - 9/15: smart cache considers them as two separate pkgs during linking
   #              but only returns the Package data from sys channel
   #              Thus the Package created from vibdb does not satisfy as many
   #              reqs b/c we don't expose provides etc
   #              That's fine as long as the sys channel is returned and the (name,ver)
   #              is identical -> one reference
   #              *** As long as (n,v) tuples in VIB class are sets, dups will disappear
   #

   __stateversion__ = smart.cache.Loader.__stateversion__+3

   def __init__(self, vibs):
      """Constructor for VibCollectionLoader.  Parameters:
         * vibs    - a dictionary of VIB instances, such as a VIBCollection.
      """
      assert hasattr(vibs, 'values')
      smart.cache.Loader.__init__(self)
      self.vibs = vibs

   def reset(self):
      smart.cache.Loader.reset(self)
      self.vibs = dict()

   def load(self):
      ProvidesNV, Obs, Pkg = platform.getsmartclasses()
      for vib in self.vibs.values():
         info = {}
         name, version = platform.smartnv(vib)
         if not name or not version:
            log.debug('Cannot obtain (name, version) from Vib %s in DB, '
                      'skipping.' % (vib.id))
            continue

         #
         # Both of these are needed for smart cache.linkdeps():
         # A package provides its own name and version, and obsoletes
         # older versions of itself.
         #
         obsolder = (Obs, name, '<', version)
         upgargs = []
         prvargs = [(ProvidesNV, name, version)]

         for repname, repop, repver in vib.replaces:
            upgargs.append((Obs, repname, repop, repver))

         # On COS smart auto adds (obsolder) into its scan results
         if not obsolder in upgargs:
            upgargs.append(obsolder)

         pkg = self.buildPackage((Pkg, name, version),
                                 prvargs, [], upgargs, [])
         pkg.loaders[self] = info


class VibCollectionChannel(smart.channel.PackageChannel):
   ''' Load smart cache from a VIBCollection or other dict of VIBs
       >> channel = VibCollectionChannel(vibs=vibDB.vibs)
   '''
   def __init__(self, *args, **kw):
      super(VibCollectionChannel, self).__init__(*args)
      self._vibs = kw['vibs']

   def fetch(self, fetcher, progress):
      self.removeLoaders()
      loader = VibCollectionLoader(self._vibs)
      loader.setInstalled(False)
      loader.setChannel(self)
      self._loaders.append(loader)
      return True


if __name__ == '__main__':
   # Test me!
   import sys

   if len(sys.argv) < 3:
      sys.stderr.write('Usage: %s <repo-type> <url> [<repo-type> <url>, ...]\n'
                       '  <repo-type> - "rpm-md", "apt-deb", "vibdb"\n'
                       '  <url>       - location of repo metadata or vibs.xml\n' %
                       sys.argv[0])
      sys.exit(1)

   platform.PlatformInit()
   repotype, url = sys.argv[1:3]
   if repotype == 'vibdb':
      # url is path to vibs.xml
      # NOTE: We cannot import Database b/c Database imports scan.py
      # and when Python parses 'class PkgScanner' it hangs
      #
      import Vib
      try:
         import elementtree.ElementTree as etree
      except ImportError:
         import xml.etree.ElementTree as etree

      # Open vibs.xml and get the Vibs ourselves
      vibs = dict()
      root = etree.ElementTree(file=url).getroot()
      for child in root:
         if child.tag == 'vib':
            vib = Vib.VIB.fromxml(child)
            vibs[vib.id] = vib

      scanner = PkgScanner(usesyschannel=True)
      scanner.AddVibCollectionChannel(vibs)
   else:
      scanner = PkgScanner('test-repo', repotype, url, True)
   scanner.Load()
   pkgs = scanner.Scan()
   for pkg in pkgs:
      sys.stdout.write('Package - %s\n' % pkg)
      sys.stdout.write('  Installed: %s\n' % pkg.installed)
      for req in pkg.requires:
         sys.stdout.write('  Requires: %s\n' % req)
         for p in req.packages:
            sys.stdout.write('    Provided by: %s\n' % p)
      for obs in pkg.obsoletes:
         sys.stdout.write('  Obsoletes: %s\n' % obs)
         for p in obs.packages:
            sys.stdout.write('    %s\n' % p)
      for con in pkg.conflicts:
         sys.stdout.write('  Conflicts: %s\n' % con)
         for p in con.packages:
            sys.stdout.write('    %s\n' % p)
      sys.stdout.write('  Newer packages:\n')
      for upg in pkg.newerpkgs:
         sys.stdout.write('    %s\n' % upg)

   #
   # Dump out indexes
   #
   sys.stdout.write('\nrequiredby: (%d entries)\n' % (len(scanner.requiredby)))
   for req in scanner.requiredby:
      sys.stdout.write('  Req: %s\n' % (req))
      for pkg in scanner.requiredby[req]:
         sys.stdout.write('    Required by: %s\n' % (pkg))

   sys.stdout.write('\nreqcache: (%d entries)\n' % (len(scanner.reqcache)))
   for req in scanner.reqcache:
      sys.stdout.write('  Req: %s\n' % (req))
      for pkg in scanner.reqcache[req].packages:
         sys.stdout.write('    Provided by: %s\n' % (pkg))

   sys.stdout.write('\nreplaceddeps: (%d entries)\n' % (len(scanner.replaceddeps)))
   for req in scanner.replaceddeps:
      sys.stdout.write('  Req: %s\n' % (req))
      for pkg in scanner.replaceddeps[req]:
         sys.stdout.write('    Broken by: %s\n' % (pkg))
