__all__ = ["xmlUtils", "arfile", "enum"]

import sys, os, select, fcntl, logging, re, datetime, calendar

import xmlUtils

log = logging.getLogger('util')

def readstdin(timeout = 3.0):
   '''Read input from stdin, until:
         * An EOF occurs (i.e. parent's pipe to stdin is closed).
         * The timeout (in seconds) expires with no input being read.
      Returns a string with the data read (or empty string if no data read).
   '''
   # Make stdin non-blocking to avoid hanging on read().
   try:
      flags = fcntl.fcntl(sys.stdin, fcntl.F_GETFL) | os.O_NONBLOCK
      fcntl.fcntl(sys.stdin, fcntl.F_SETFL, flags)
   except Exception, e:
      log.warn('Unable to set stderr non-blocking: %s' % e)
      return ''
   data = ''
   while True:
      try:
         ready = select.select([sys.stdin], [], [], timeout)[0]
      except select.error, (n, msg):
         if n == errno.EINTR:
            # If interrupted by a signal, just continue.
            continue
         else:
            log.error('select error waiting for input on stdin: %s' % msg)
            break
      if ready:
         newdata = sys.stdin.readline()
         data += newdata
      else:
         log.warn('Timeout waiting for input on stdin.')
         break
      if newdata in ('', '\n'):
         break
   return data

class UTC(datetime.tzinfo):
   ZERO = datetime.timedelta(0)
   utcoffset = lambda self, dt: self.ZERO
   dst       = lambda self, dt: self.ZERO
   tzname    = lambda self, dt: 'UTC'

# We define this outside of the function, only so that we don't have to
# compile the regex each time. Marginal efficiency enhancement, I know... -RS
tsregex = re.compile(r'(?P<year>\d{4})-(?P<month>\d\d)-(?P<day>\d\d)'
                     r'(.'
                     r'(?P<hour>\d\d):(?P<minute>\d\d):(?P<second>\d\d)'
                     r'(?:\.(?P<microsecond>\d{6}))?'
                     r'(?:(?P<offsetop>[+-])'
                     r'(?P<offsethours>\d\d):(?P<offsetminutes>\d\d))?'
                     r')?')
def isotofloat(isostring):
   """Parse format returned by datetime.isoformat() back into a
      float value, as returned by time.time().
         Parameters:
            * isostring - A string with a date and time, in the format given
                          by the isoformat() method in the datetime module.
         Returns: A float value representing the POSIX timestamp in UTC, as
                  returned by time.time().
         Raises:
            * ValueError - On failure to parse the timestamp string.
   """
   # The isoformat() method in the datetime module, despite it's name, doesn't
   # seem to return a timestamp that is compliant with ISO 8601:
   #    1. It includes microseconds, but I could find nothing in the standard
   #       that specifies a way to express increments of time less than 1
   #       second.
   #    2. When expressing a UTC time, the format specifies the use of 'Z' or
   #       no +/-hh:mm offset designation at all. isoformat(), however, uses
   #       +00:00.
   # That being said, it's really, really difficult to write a parser anyway
   # that can interpret the wide array of possible notations specified by ISO
   # 8601 (which may be why I couldn't find one for Python). So, here, we just
   # attempt to parse the format returned by the isoformat() function. -RS
   match = tsregex.match(str(isostring))
   if match is None:
      raise ValueError("Can not parse timestamp from '%s'." % isostring)

   dtargs = dict(tzinfo=UTC())
   for key in ('year', 'month', 'day', 'hour', 'minute', 'second',
               'microsecond'):
      if match.group(key) is None:
         continue
      try:
         dtargs[key] = int(match.group(key))
      except:
         raise ValueError("Can not convert '%s' field to integer in '%s'." %
                          (key, isostring))

   dt = datetime.datetime(**dtargs)

   if (match.group('offsetop') and match.group('offsethours') and 
       match.group('offsetminutes')):
      try:
         offset = datetime.timedelta(hours=int(match.group('offsethours')),
                                     minutes=int(match.group('offsetminutes')))
      except:
         raise ValueError("Invalid offset specification in '%s'." % isostring)
      if match.group('offsetop') == '+':
         dt -= offset
      elif match.group('offsetop') == '-':
         dt += offset

   ts = calendar.timegm(dt.utctimetuple())
   ts += dt.microsecond / 1000000.0
   return ts

def floattoiso(ts):
   localtz = xmlUtils.LocalTimezone()
   dt = datetime.datetime.fromtimestamp(ts, localtz)
   return dt.isoformat()
