#!/usr/bin/python

'''This module provides a class for manipulating archives created with GNU ar.
It provides support for the GNU extension for long file names, but it currently
only supports extraction.'''

import os
from stat import *
import struct
import time

STRUCT_AR_HEADER = '16s12s6s6s8s10s2s'
AR_FILEMAGIC = '\140\012'


class ArError(Exception):
   def __init__(self, filename, msg):
      Exception.__init__(self, filename, msg)
      self.filename = filename
      self.msg = msg

class ArInfo(object):
   '''A class representing a member of the archive.  It contains the following
      members:
      * filename  - the original name of the member (string)
      * timestamp - the mtime of the member (int)
      * uid       - UID of file (int)
      * gid       - GID of file (int)
      * mode      - permissions mask of file (int)
      * size      - length of file data
      * offset    - offset of archive where file data begins
   '''
   def __init__(self, filename, timestamp=0, uid=0, gid=0, mode=0, size=0,
                offset=0):
      self.filename = filename
      self.timestamp = timestamp
      self.uid = uid
      self.gid = gid
      self.mode = mode
      self.size = size
      self.offset = offset


class _FileInFile(object):
    """A thin wrapper around an existing file object that
       provides a part of its data as an individual file
       object.  Taken from python 2.5's tarfile module.
    """

    def __init__(self, fileobj, offset, size):
        self.fileobj = fileobj
        self.offset = offset
        self.size = size
        self.position = 0

    def tell(self):
        """Return the current file position.
        """
        return self.position

    def seek(self, position):
        """Seek to a position in the file.
        """
        if 0 <= position <= self.size:
           self.position = position
        else:
           raise ValueError("Seek(%d) exceeds bounds [0, %d]" %
                            (position, self.size))

    def read(self, size=None):
        """Read data from the file.
        """
        if size is None:
            size = self.size - self.position
        else:
            size = min(size, self.size - self.position)

        self.fileobj.seek(self.offset + self.position)
        bytesread = self.fileobj.read(size)
        self.position += len(bytesread)
        return bytesread


class ArFile(object):
   '''A class representing a GNU ar archive.  Supports the GNU extension to
      handle long file names when reading, but can only write short filenames.'''
   def __init__(self, name=None, mode='rb', fileobj=None):
      '''If name is specified, that file is opened in the mode specified.
         If name is omitted, fileobj may be specified, in which case data will
         be read from that file object.  However, the object must support the
         seek method.   This class does not currently support reading from a
         non-seekable stream, although it could be made to do so.
         Once object is instantiated, a list of ArInfo objects are available
         in the filelist attribute.
         To create a new ar file, one of the 'w' modes must be used.
         To append to an existing file, the 'a+b' mode is recommended, other
         wise with 'ab' the existing headers will not be read.
         '''
      if name:
         self._fileobj = open(name, mode)
         self._filename = name
      elif fileobj:
         if mode.startswith('w'):
            raise ValueError('Must initialize with name in w, wb modes')
         self._fileobj = fileobj
         if hasattr(fileobj, 'name'):
            self._filename = fileobj.name
         else:
            self._filename = str(fileobj)
      else:
         raise ValueError('Must initialize with either name or file object')
      self._longfns = ''
      self.filelist = list()
      if mode.startswith('w'):
         self._fileobj.write('!<arch>\n')
      elif mode[0] == 'r' or '+' in mode:
         self._getFilelist()

   def close(self):
      '''Close the archive.  Further attempts to read member data will fail.'''
      self._fileobj.close()

   def _getFilelist(self):
      self._fileobj.seek(0, 0)
      armagic = self._fileobj.read(8)
      if armagic != '!<arch>\n':
         raise ArError(self._filename, 'Not a valid ar file')
      for info in self._readHeaders():
         if info.filename == '/': continue # We don't deal with symbol tables
         if info.filename == '//': # Long filename list
            pos = self._fileobj.tell()
            self._longfns = self.read(info)
            self._fileobj.seek(pos, 0)
         else:
            if info.filename.endswith('/'):
               info.filename = info.filename.rstrip('/')
            elif self._longfns: # Filename is an index into long file name list
               try:
                  start = int(info.filename.lstrip('/'))
                  end = self._longfns.find('\n', start)
                  info.filename = self._longfns[start:end]
                  info.filename = info.filename.rstrip('/')
               except ValueError, IndexError:
                  raise ArError(self._filename, 'Error getting long filename')
            self.filelist.append(info)

   def _readHeaders(self):
      while True:
         header = self._fileobj.read(1)
         #
         # Some versions of ar insert '\n' at end of last data section so
         # entire header is shifted down by one.  We check file magic
         # to align the header string. (from libbb/unarchive.c)
         #
         if header == '\n':
            header = self._fileobj.read(60)
         else:
            header += self._fileobj.read(59)
         if not header:
            break
         elif len(header) != 60:
            raise ArError(self._filename, 'Unexpected EOF')

         (filename, timestamp, uid, gid,
          mode, size, filemagic) = struct.unpack(STRUCT_AR_HEADER, header)

         if filemagic != AR_FILEMAGIC:
            raise ArError(self._filename, 'Unexpected file magic')

         filename = filename.strip(' \00')
         offset = self._fileobj.tell()

         def intfromstr(x, radix=10):
            x = x.strip(' \00')
            if x: return int(x, radix)
            else: return 0

         try:
            timestamp = intfromstr(timestamp)
            uid = intfromstr(uid)
            gid = intfromstr(gid)
            mode = intfromstr(mode, 8)
            size = intfromstr(size)
         except ValueError:
            raise ArError(self._filename, 'Invalid format: timestamp=%s '
                          'uid=%s gid=%s mode=%s size=%s header=[%s]' %
                          (timestamp, uid, gid, mode, size, header))

         self._fileobj.seek(size, 1)
         yield ArInfo(filename, timestamp, uid, gid, mode, size, offset)

   def nameToInfo(self, filename):
      '''Returns ArInfo object corresponding to filename, or None
         if no archive memeber matches filename.
      '''
      for i in self.filelist:
         if filename == i.filename:
            return i
      return None

   def read(self, member):
      '''Read file data of archive member.  Member may be either a file name
         or an instance of ArInfo.  Entire data is returned as a string. '''
      if isinstance(member, ArInfo):
         info = member
      else:
         info = self.nameToInfo(member)
         if not info:
            raise KeyError('Archive member not found: %s' % member)
      offset = info.offset
      self._fileobj.seek(offset, 0)
      return self._fileobj.read(info.size)

   def extractfile(self, member):
      '''Returns a file object for a member of the archive.
         member may be either a file name or an ArInfo instance.
         The file object is read-only and provides these methods:
         read(), seek(), tell()
      '''
      if isinstance(member, ArInfo):
         info = member
      else:
         info = self.nameToInfo(member)
         if not info:
            raise KeyError('Archive member not found: %s' % member)
      return _FileInFile(self._fileobj, info.offset, info.size)

   def writestr(self, info, bytes):
      '''Write ar header and data bytes at current file pos.
         info can be ArInfo object or archive member name.
         If a name, default values will be filled in. '''
      if not isinstance(info, ArInfo):
         info = ArInfo(filename=info, timestamp=int(time.time()),
                       mode=0644)
      if len(info.filename) > 15:
         raise ArError(self._filename, "Long filename writes are "
                       "not currently supported.")
      
      header = '%-16.16s%-12.12s%-6.6s%-6.6s%-8.8s%-10.10s%-2.2s' % (
         info.filename, info.timestamp, info.uid, info.gid, oct(info.mode),
         len(bytes), AR_FILEMAGIC)
      self._fileobj.write(header)
      self._fileobj.write(bytes)
      #
      # GNU ar apparently expects short word (2-byte)
      # alignment of headers, and \n is used to pad if neeeded
      #
      if self._fileobj.tell() & 0x01:
         self._fileobj.write('\n')
      self.filelist.append(info)

   def write(self, filename):
      '''Append filename to this archive'''
      stats = os.stat(filename)
      info = ArInfo(os.path.basename(filename), stats[ST_MTIME],
                    stats[ST_UID], stats[ST_GID], stats[ST_MODE])
      fp = open(filename, 'rb')
      self.writestr(info, fp.read())
      fp.close()


if __name__ == '__main__':
   import sys

   def usage():
      sys.stderr.write('Usage: %s [rxt] archive [member]\n'
                       '  r - replace or add member to archive\n'
                       '  x - extract member from archive (or all if omitted)\n'
                       '  t - list members in archive\n'
                       % os.path.basename(sys.argv[0]))
      raise SystemExit()

   if len(sys.argv) < 3:
      usage()
   (op, archive) = sys.argv[1:3]
   if len(sys.argv) < 4:
      member = None
   else:
      member = sys.argv[3]

   if op == 'r':
      if not member:
         usage()
      if os.path.exists(archive):
         a = ArFile(archive, 'ab')
      else:
         a = ArFile(archive, 'wb')
      a.write(member)
      a.close()

   elif op == 'x':
      a = ArFile(archive, 'rb')
      if member:
         f = open(member, 'wb')
         f.write(a.read(member))
         f.close()
      else:
         for info in a.filelist:
            f = open(info.filename, 'wb')
            f.write(a.read(info))
            f.close()
      a.close()

   elif op == 't':
      a = ArFile(archive, 'rb')
      for info in a.filelist:
         print info.filename
      a.close()
