#!/usr/bin/python
########################################################################
# Copyright (C) 2008 VMWare, Inc.
# All Rights Reserved
########################################################################

import os, stat

class CpioFile(object):
   '''A class supporting creating a CPIO archive with the CRC format.
      Read support is currently not implemented, but should not be difficult.
   '''
   def __init__(self, filename, mode='wb'):
      '''filename is either a file name (string) or file-like object.
         If filename is a string, a file with that name is opened with the
         specified mode.  Otherwise, filename must be an object that supports
         read() and write() methods, and mode is ignored.
      '''
      if isinstance(filename, basestring):
         self._file = open(filename, mode)
      else:
         self._file = filename
      self._length = 0

   @staticmethod
   def crcfile(filename):
      f = open(filename, 'rb')
      data = f.read(512)
      crc = 0
      while data:
         crc = (crc + sum(ord(c) for c in data)) & 0xffffffff
         data = f.read(512)
      f.close()
      return crc

   def addfile(self, filename, archivename=None):
      '''Add filename to the archive.  If archivename is specified, it is used
         as the name in the archive.
      '''
      if archivename is None:
         archivename = filename
      crc = 0
      size = 0
      s = os.lstat(filename)
      if stat.S_ISLNK(s.st_mode):
         data = os.readlink(filename)
         size = len(data)
      elif stat.S_ISREG(s.st_mode):
         # You may note that we read the entire contents of a file twice:
         # once here, and again when adding it to the archive.  We could avoid
         # doing so by calculating the CRC at the same time we read the data,
         # but because the checksum appears in the archive header before the
         # actual data, it would mean that our self._file object would have to
         # gaurantee support for seek() and tell() methods.  This would rule
         # out writing to stdout.
         crc = self.crcfile(filename)
         size = s.st_size
      archivename += '\0' # Must always be null-terminated.
      hdr = '070702%08X%08X%08X%08X%08X%08X%08X%08X%08X%08X%08X%08X%08X%s' % \
            (s.st_ino, s.st_mode, s.st_uid, s.st_gid, s.st_nlink, s.st_mtime,
             size, os.major(s.st_dev), os.minor(s.st_dev), os.major(s.st_rdev),
             os.minor(s.st_rdev), len(archivename), crc, archivename)
      if len(hdr) % 4:
         # Pad hdr with NULL characters to make length a multiple of 4.
         hdr += '\0' * (4 - len(hdr) % 4)
      self._file.write(hdr)
      self._length += len(hdr)
      if stat.S_ISLNK(s.st_mode):
         if len(data) % 4:
            # data is also padded to a multiple of 4.
            data += '\0' * (4 - len(data) % 4)
         self._file.write(data)
         self._length += len(data)
      elif stat.S_ISREG(s.st_mode):
         f = open(filename, 'rb')
         data = f.read(512)
         datalen = len(data)
         while data:
            self._file.write(data)
            self._length += len(data)
            data = f.read(512)
            datalen += len(data)
         f.close()
         if datalen % 4:
            padding = '\0' * (4 - datalen % 4)
            self._file.write(padding)
            self._length += len(padding)

   def close(self):
      '''Write trailer and close file.'''
      trailer = ('070702000000000000000000000000000000000000000100000000'
                 '00000000000000000000000000000000000000000000000B000000'
                 '00TRAILER!!!\0')
      self._file.write(trailer)
      self._length += len(trailer)
      if self._length % 512:
         # GNU cpio pads archive to make total length a multiple of 512 bytes,
         # so we do the same.
         self._file.write('\0' * (512 - self._length % 512))
      self._file.close()

def mkrpmcpio(srcspec, dest, recurse=True, root=''):
   '''A convenience function that makes a CPIO archive that can be used as an
      RPM payload.
      Parameters:
         * srcspec - Either a string specifying a single file or directory, or
                     a sequence with a list of files and/or directories.
         * dest    - Either a string specifying a path name to write the
                     archive to, or a file-like object.
         * recurse - If True, recurses through directories, adding all sub-
                     directories and their contents to the archive.  If False,
                     only those paths explicitly specified in srcspec are
                     added.
         * root    - If specified, represents a directory name to strip from
                     file names when adding them to the archive.  For example,
                     if srcspec is '/tmp/cpio.stage/mydir/myfile', and root is
                     '/tmp/cpio.stage', then the file name recorded in the
                     archive is './mydir/myfile'.
   '''
   def mkrpmname(p):
      # strip 'root' from beginning of path.
      if root and p.startswith(root):
         p = p[len(root):]
      # In RPM CPIO archives, directory names don't use a trailing /.
      p = p.rstrip('/')
      # All RPM CPIO archive paths start with './'.
      if p.startswith('./'): return p
      if p.startswith('/'): return '.' + p
      return './' + p

   if isinstance(srcspec, basestring):
      sources = [srcspec]
   else:
      sources = list(srcspec)

   cpio = CpioFile(dest)

   for path in sources:
      isdir = os.path.isdir(path) and not os.path.islink(path)
      if not recurse or not isdir:
         cpio.addfile(path, mkrpmname(path))
         continue
      for dirpath, dirnames, filenames in os.walk(path):
         if os.path.islink(dirpath):
            continue
         cpio.addfile(dirpath, mkrpmname(dirpath))
         for subdir in dirnames:
            subdirpath = os.path.join(dirpath, subdir)
            if os.path.islink(subdirpath):
               cpio.addfile(subdirpath, mkrpmname(subdirpath))
         for fn in filenames:
            filepath = os.path.join(dirpath, fn)
            cpio.addfile(filepath, mkrpmname(filepath))

   cpio.close()

if __name__ == '__main__':
   import sys

   if len(sys.argv) < 3:
      sys.exit(2)

   if sys.argv[1] == '-':
      mkrpmcpio(sys.argv[2:], sys.stdout)
   else:
      mkrpmcpio(sys.argv[2:], sys.argv[1])
