#!/usr/bin/python
#
# This module implements some work-arounds for broken HTTPS proxy support in
# the httplib and urllib2 modules included with the Python distribution.
# See http://bugs.python.org/issue1424152 for more information.
#
# This code has a lot of dependencies on implementation-specific things in
# httplib and urllib2, and is likely to break if/when they are updated.  It
# seems to work, however, with versions 2.4 and 2.5.

import socket, httplib, urllib, urllib2

class ConnectProxyError(httplib.error):
   pass

class HTTPSProxyConnection(httplib.HTTPSConnection):
   '''Like httplib.HTTPSConnection, but we establish a tunnel through a proxy
      first.
   '''
   def __init__(self, host, port=None, key_file=None, cert_file=None,
                strict=None, proxy=None):
      httplib.HTTPSConnection.__init__(self, host, port, key_file, cert_file,
                                       strict)
      if proxy and proxy.count('://'):
         proto, proxy = proxy.split('://', 1)
         if proto != 'http':
            raise ValueError("Unknown proxy protocol '%s'" % proto)
      self.proxy = proxy

   def connect(self):
      if not self.proxy:
         return httplib.HTTPSConnection.connect(self)
      # Let's talk to our proxy server:
      h = httplib.HTTPConnection(self.proxy)
      h.request('CONNECT', '%s:%d' % (self.host, self.port))
      resp = h.getresponse()
      if resp.status != 200:
         raise ConnectProxyError("Error connecting to proxy: %s %s" %
                                 (resp.status, resp.reason))
      # If our proxy server gave us a green light, anything we send/receive
      # on this socket from here on out should be tunneled to the remote
      # server.  Re-use the socket (resp.fp) from our connection to the proxy,
      # and set up our SSL connection now.
      ssl = socket.ssl(resp.fp, self.key_file, self.cert_file)
      self.sock = httplib.FakeSocket(resp.fp, ssl)

class ProxyHandler(urllib2.ProxyHandler, urllib2.AbstractHTTPHandler):
   '''Sub-classes urllib2.ProxyHandler in order to handle https:// connections
      using the HTTPSProxyConnection class.
   '''
   def __init__(self, *args, **kwargs):
      urllib2.AbstractHTTPHandler.__init__(self)
      urllib2.ProxyHandler.__init__(self, *args, **kwargs)
      # See the call to setattr in urllib2.ProxyHandler.__init__ to understand
      # why this is (unfortunately) necessary.
      if hasattr(self, 'https_open') and hasattr(httplib, 'HTTPS'):
         self.https_open = self._https_open

   https_request = urllib2.AbstractHTTPHandler.do_request_

   def _https_open(self, req):
      # This is essentially duplicated from
      # urllib2.AbstractHTTPHandler.do_open.  :-(
      host = req.get_host()
      if not host:
         raise URLError('no host given')

      # Here is our change to use HTTPSProxyConnection if we have an https
      # proxy.
      if self.proxies.has_key('https'):
         h = HTTPSProxyConnection(host, proxy=self.proxies['https'])
      else:
         h = httplib.HTTPSConnection(host)
      # End our changes.

      h.set_debuglevel(self._debuglevel)

      headers = dict(req.headers)
      headers.update(req.unredirected_hdrs)
      headers["Connection"] = "close"
      headers = dict(
         (name.title(), val) for name, val in headers.items())
      try:
         h.request(req.get_method(), req.get_selector(), req.data, headers)
         r = h.getresponse()
      except socket.error, err:
         raise urllib2.URLError(err)

      r.recv = r.read
      fp = socket._fileobject(r)

      resp = urllib.addinfourl(fp, r.msg, req.get_full_url())
      resp.code = r.status
      resp.msg = r.reason
      return resp

if __name__ == '__main__':
   import sys, optparse

   parser = optparse.OptionParser()
   parser.set_usage('%prog [options] URL')
   parser.add_option('-p', '--proxy', dest='proxy',
                     metavar='proto://host:port')
   options, args = parser.parse_args()
   if len(args) < 1:
      parser.error('URL required')
   if len(args) > 1:
      parser.error('Too many arguments')

   proxies = dict()
   if options.proxy:
      proxies['ftp'] = options.proxy
      proxies['http'] = options.proxy
      proxies['https'] = options.proxy
   opener = urllib2.build_opener(ProxyHandler(proxies))
   f = opener.open(args[0])
   sys.stdout.write(f.read())
