try:
   from lxml.etree import ElementTree, Element, iselement
except ImportError:
   try:
      from xml.etree.cElementTree import ElementTree, Element, iselement
   except ImportError:
      from cElementTree import ElementTree, Element, iselement


#
# Globals and Constants
#
SPACING = "  "

#
# Returns a string  \n + SPACING * indent
#
def Indentstr(indent):
    return "\n" + SPACING * indent

#
# Recursively indents an Element tree for readable XML output.
# NOTE: Mixed-mode elements (text + child Elements) are not
# supported properly;  the assumption is that an Element
# contains only text or only child Elements.
#
# Side effects: Modifies text and tail members of Element instances
#
def IndentElementTree(elem, indent=0):
    """ Recursively indents Element elem and all children,
    starting at indentation level indent.
    """
    assert iselement(elem), "Expecting Element type but got %s" % (type(elem))
    elem.tail = Indentstr(indent)
    for child in elem:
        IndentElementTree(child, indent + 1)

    if len(elem):
        #
        # Fix tail string of last child element so it returns us to our indent level
        #
        elem.text = Indentstr(indent + 1)
        elem[-1].tail = Indentstr(indent)

#
# Adds <tag **attrs>newVal</tag> element to the end of root.
# **attrs are keyword arguments that become element attributes.
#
def AppendElement(root, tag, val, **attrs):
    assert iselement(root), "root is not an Element instance"
    elem = Element(tag, **attrs)
    if val:
        elem.text = val
    root.append(elem)

#
# Adds element only if val is not None or empty string
#
def SetElement(root, tag, val, **attrs):
   """ Adds new sub element to root if val is not empty or None"""
   if val:
      AppendElement(root, tag, val, **attrs)

#
# Adds element <tag>true/false</tag>
# (per xsd:bool standard)
#
def AppendBoolElement(root, tag, val, **attrs):
   if val:
      AppendElement(root, tag, 'true', **attrs)
   else:
      AppendElement(root, tag, 'false', **attrs)

#
# False if empty, or '0', or 'false'
#
def StrToBool(val):
   if val and val != '0' and val.lower() != 'false':
      return True
   else:
      return False

#
# TODO: move to a proper place
# From python document http://docs.python.org/lib/datetime-tzinfo.html
#
import time as _time
from datetime import tzinfo, timedelta, datetime

ZERO = timedelta(0)

STDOFFSET = timedelta(seconds = -_time.timezone)
if _time.daylight:
   DSTOFFSET = timedelta(seconds = -_time.altzone)
else:
   DSTOFFSET = STDOFFSET

DSTDIFF = DSTOFFSET - STDOFFSET

class LocalTimezone(tzinfo):

   def utcoffset(self, dt):
      if self._isdst(dt):
         return DSTOFFSET
      else:
         return STDOFFSET

   def dst(self, dt):
      if self._isdst(dt):
         return DSTDIFF
      else:
         return ZERO

   def tzname(self, dt):
      return _time.tzname[self._isdst(dt)]

   def _isdst(self, dt):
      tt = (dt.year, dt.month, dt.day,
            dt.hour, dt.minute, dt.second,
            dt.weekday(), 0, -1)
      stamp = _time.mktime(tt)
      tt = _time.localtime(stamp)
      return tt.tm_isdst > 0

def GetCurrentDateTime():
   return datetime.now(LocalTimezone())
