#######################################################################
# Copyright (C) 2008 VMWare, Inc.
# All Rights Reserved
########################################################################

"""This module contains classes to express and compare version information."""

__all__ = ['Version', 'RPMVersion', 'DebVersion', 'ESXVersion']

class Version(object):
   """A generic class for expressing a version.
         Attributes:
            * versionstring - A string containing the version.
   """
   def __init__(self, versionstring=''):
      """Class constructor.
            Parameters:
               * versionstring - The string containing version information.
            Returns: A new Version object.
      """
      self.versionstring = versionstring

   __str__ = lambda self: self.versionstring
   __nonzero__ = lambda self: bool(self.versionstring)

   @classmethod
   def fromstring(cls, versionstring):
      """Parses versionstring into appropriate fields, returning a Version
         object. (This method is more useful in child classes."""
      return cls(versionstring)

   def __cmp__(self, other):
      """Provides a comparison function between two Version objects.
         Comparison is based solely on the 'versionstring' attribute, but the
         string is divided into fields by splitting at '.' characters. Each
         field is then further divided into a leading numeric portion and a
         trailing portion. Fields from each versionstring are compared left-
         to-right, by first doing a comparison between the numeric portions,
         then comparing the string portions of the numeric parts are equal.
         Comparison ends when one of the fields is greater or less than the
         corresponding field in the other version, or when we run out of
         fields to compare. This algorithm provides the least-surprising
         results.  For example:
             * 4.2 is less than 4.10. (a stringwise comparision would provide
               the opposite result.)
             * 4.0 is less than 4.0a. (a numeric comparison could not account
               for the 'a'.)
             * 4 is less than 4.0.
             * 4.a is less than 4.0a.
             * 4.0 is less than 4a.
         Returns: -1, 0, or 1, if other is less than, equal to or greater
                  than self.
      """
      def splitfield(field):
         i = 0
         while i < len(field):
            if field[i] not in '0123456789':
               break
            i += 1
         numeric = (i or None) and int(field[:i])
         trailing = field[i:] or None
         return (numeric, trailing)
      mine = tuple(splitfield(f) for f in self.versionstring.split('.'))
      theirs = tuple(splitfield(f) for f in str(other).split('.'))
      return cmp(mine, theirs)

class FWVersion(Version):
   """A class for representing a firmware version.
         Attributes:
            * majorversion   - A positive integer expressing the major
                               firmware version. May be None if all
                               other version fields are None.
            * minorversion   - A positive integer expressing the minor
                               firmware version. May be None if
                               revisionnumber and buildnumber are None.
            * revisionnumber - A positive integer expressing the
                               firmware revision number. May be None if
                               buildnumber is None.
            * buildnumber    - A positive integer expressing the
                               firmware build number. May be None.
         Properties:
            * versionstring - A read-only property expressing the
                              version fields as a string.
   """
   def __init__(self, majorversion=None, minorversion=None,
                revisionnumber=None, buildnumber=None):
      """Class constructor.
            Parameters:
               * majorversion   - Sets attribute of same name.
               * minorversion   - Sets attribute of same name.
               * revisionnumber - Sets attribute of same name.
               * buildnumber    - Sets attribute of same name.
      """
      self.majorversion    = majorversion
      self.minorversion    = minorversion
      self.revisionnumber  = revisionnumber
      self.buildnumber     = buildnumber

   def __cmp__(self, other):
      if isinstance(other, FWVersion):
         mine   = (self.majorversion, self.minorversion,
                   self.revisionnumber, self.buildnumber)
         theirs = (other.majorversion, other.minorversion,
                   other.revisionnumber, other.buildnumber)
         return cmp(mine, theirs)
      else:
         return Version.__cmp__(self, other)

   @property
   def versionstring(self):
      """Return firmware version as a string. All numeric fields are
         concatenated with a '.', and None values are not included.
         Returns the empty string if all values are None.
      """
      attrs = (self.majorversion, self.minorversion, self.revisionnumber,
               self.buildnumber)
      fields = tuple(str(f) for f in attrs if f is not None)
      if fields:
         return '.'.join(fields)
      return ''

   @classmethod
   def fromstring(cls, versionstring):
      """Create a new FWVersion object from a string.
            Parameters:
               * versionstring - A properly formatted string consisting of
                                 zero to four decimal integers separated by
                                 '.' characters.
            Returns: A new FWVersion object.
            Raises:
               * ValueError - If string is not in the proper format.
      """
      try:
         parameters = tuple(int(field) for field in versionstring.split('.'))
         return cls(*parameters)
      except:
         raise ValueError("Could not parse '%s' into version fields." %
                          versionstring)

class RPMVersion(Version):
   """A class to express an RPM version.
        Attributes:
           * epoch   - A positive integer giving the RPM epoch, or None.
           * version - A string giving the RPM version.
           * release - A string giving the RPM release.
           * arch    - A string giving the RPM arch, or None if unknown.
        Properties:
           * versionstring - Represents RPMVersion as a string.
   """
   def __init__(self, epoch, version, release, arch=None):
      """Class constructor.
            Parameters:
               * epoch   - A positive integer value or None to assign
                           to the epoch attribute.
               * version - A string to assign to the version attribute.
               * release - A string to assign to the release attribute.
               * arch    - A string to assign to the arch attribute, or None.
            Returns: A new RPMVersion object.
      """
      self.epoch = epoch
      self.version = Version(version)
      self.release = Version(release)
      self.arch = arch

   @property
   def versionstring(self):
      """Returns RPMVersion information as a string."""
      vrfields = (self.version.versionstring, self.release.versionstring)
      rc = '-'.join(x for x in vrfields if x)
      if self.epoch:
         rc = ':'.join((str(self.epoch), rc))
      if self.arch:
         rc = '@'.join((rc, self.arch))
      return rc

   @classmethod
   def fromstring(cls, versionstring):
      """Create an RPMVersion object from a string.
            Parameters:
               * versionstring - Must be a string in either version-release
                                 or epoch:version-release format.
            Returns: A new RPMVersion object.
            Raises:
               * ValueError - If string cannot be parsed into an RPM version.
      """
      epoch = None
      if versionstring.count(':'):
         try:
            (epoch, vra) = versionstring.split(':')
            epoch = int(epoch)
         except:
            msg = "Could not parse '%s' to an RPM version." % versionstring
            raise ValueError(msg)
      else:
         vra = versionstring

      arch = None
      if vra.count('@'):
         try:
            (versionrelease, arch) = vra.split('@')
         except:
            msg = "Could not parse '%s' to an RPM version." % versionstring
            raise ValueError(msg)
      else:
         versionrelease = vra

      try:
         (version, release) = versionrelease.split('-')
      except:
         msg = "Could not parse '%s' to an RPM version." % versionstring
         raise ValueError(msg)

      if not version or not release:
         msg = "Could not parse '%s' to an RPM version." % versionstring
         raise ValueError(msg)

      return cls(epoch, version, release, arch)

   def __cmp__(self, other):
      if isinstance(other, RPMVersion):
         mine   = (self.epoch, self.version, self.release)
         theirs = (other.epoch, other.version, other.release)
         return cmp(mine, theirs)
      else:
         mine = '-'.join((self.version.versionstring,
                          self.release.versionstring))
         return cmp(mine, str(other))

class DebVersion(Version):
    pass

class ESXVersion(Version):
   """A class representing the ESX version.
         Attributes:
            * major       - A non-negative integer giving the ESX major
                            version, or None. Cannot be None if minor,
                            maintenance, update or patch attributes are
                            integers.
            * minor       - A non-negative integer giving the ESX minor
                            version, or None. Cannot be None if
                            maintenance, update or patch attributes are
                            integers.
            * maintenance - A non-negative integer giving the ESX
                            maintenance release, or None. Cannot be None
                            if update or patch are integers.
            * update      - A non-negative integer giving the update
                            version, or None. Cannot be None if patch is
                            an integer.
            * patch       - A non-negative integer giving the patch
                            release, or None.
         Properties:
            * versionstring - Represents ESX version as a string.
   """
   def __init__(self, major=None, minor=None, maintenance=None,
                 update=None, patch=None):
      """Class constructor.
            Parameters:
               * major       - Optional parameter giving the major version.
               * minor       - Optional parameter giving the minor version.
               * maintenance - Optional parameter giving the maintenance
                               release.
               * update      - Optional parameter giving the update version.
               * patch       - Optional parameter giving the patch version.
            Returns: New ESXVersion object.
      """
      self.__major = major
      self.__minor = minor
      self.__maintenance = maintenance
      self.__update = update
      self.__patch = patch

   major =       property(lambda self: self.__major)
   minor =       property(lambda self: self.__minor)
   maintenance = property(lambda self: self.__maintenance)
   update =      property(lambda self: self.__update)
   patch =       property(lambda self: self.__patch)

   def __hash__(self):
      return hash((self.__major, self.__minor, self.__maintenance, self.__update,
                   self.__patch))

   def __cmp__(self, other):
      if isinstance(other, ESXVersion):
         # Note that major, minor, maintenance may be 'e', 'x', 'p'. In that
         # case, strings are always greater than integers (in Python), so this
         # has the effect that 'e.x.p' is always greater than '4.0.0' (etc.).
         mine = (self.__major, self.__minor, self.__maintenance, self.__update,
                 self.__patch)
         theirs = (other.major, other.minor, other.maintenance, other.update,
                   other.patch)
         return cmp(mine, theirs)
      else:
         return Version.__cmp__(self, other)

   @property
   def versionstring(self):
      """Return ESX version as a string. None values are not included. Returns
         the empty string if all values are None.
      """
      versionfields = (self.__major, self.__minor, self.__maintenance)
      releasefields = (self.__update, self.__patch)
      version = '.'.join(str(f) for f in versionfields if f is not None)
      release = '.'.join(str(f) for f in releasefields if f is not None)
      if release:
         return '%s-%s' % (version, release)
      elif version:
         return version
      else:
         return ''

   @classmethod
   def fromstring(cls, versionstring):
      """Create ESXVersion object from string.
            Parameters:
               * versionstring - String specifiying version number in
                                 major.minor.maintenance-update.patch format.
            Returns: New ESXVersion object.
            Raises:
               * ValueError - If versionstring cannot be parsed.
      """
      if versionstring.count('-'):
         try:
            version, release = versionstring.split('-')
         except:
            msg = "Could not parse ESX version string '%s'." % versionstring
            raise ValueError(msg)
         if not version or not release:
            msg = "Could not parse ESX version string '%s'." % versionstring
            raise ValueError(msg)
      else:
         version = versionstring
         release = ''
      versionfields = list()
      releasefields = list()
      for x in version.split('.'):
         if not x:
            raise ValueError("Empty field in '%s'." % versionstring)
         # e.x.p version still exists in places. Boo. -RS
         try:
            versionfields.append(int(x))
         except:
            versionfields.append(x)
      if release:
         for x in release.split('.'):
            if not x:
               raise ValueError("Empty field in '%s'." % versionstring)
            try:
               releasefields.append(int(x))
            except:
               releasefields.append(x)
      if len(versionfields) > 3 or len(releasefields) > 2:
         raise ValueError("Too many fields in '%s'." % versionstring)
      versionfields.extend([None] * (3 - len(versionfields)))
      releasefields.extend([None] * (2 - len(releasefields)))
      return cls(*(versionfields + releasefields))

   def matches(self, host):
      """Determine if this ESXVersion object matches another. Each attribute
         in self is compared to see if it matches the same attribute in
         'host'. However, if the attribute in 'self' is None, the attribute
         is considered to match the one in 'host'.  (I.e. a None value in
         'self' acts as a wild card.)  This means that if 'self' corresponds
         to version 4.0, self.matches(host) will return true if host is
         4.0.1, 4.0.0-1, 4.0.0-1.2, etc.
            Parameters:
               * other - Must be an instance of ESXVersion, or a string
                         which can be used to create a valid ESXVersion
                         object using ESXVersion.fromstring().
            Returns: True if matches, False if not.
            Raises:
               * ValueError - If 'other' is not an ESXVersion object, but
                              cannot be parsed by ESXVersion.fromstring().
      """
      if not isinstance(host, ESXVersion):
         host = ESXVersion.fromstring(str(host))
      for attr in ('major', 'minor', 'maintenance', 'update', 'patch'):
         myval = getattr(self, attr)
         if myval in (None, '*'):
            return True
         elif getattr(host, attr) != myval:
            return False
      return True
